// ==============================================================
//						ORBITER MODULE:
//                  Part of the ORBITER SDK
//               Copyright (C) 2013 Greg Hlynka
//                    All rights reserved
//
// AAPO_Common.h
// Common values and functions for the Apollo Aplication Project for Orbiter
//
// Notes: This file exists
// ==============================================================

#include "orbitersdk.h"
#include "AAPO_Common.h"

// ==============================================================
// Math Functions
// ==============================================================

// --------------------------------------------------------------
// Convert axis/angle to quaternion form, angle is in radians.
// --------------------------------------------------------------
void to_quaternion (VECTOR3 axis, float angle, float result[4])
{
	normalise (axis);

	result[0] = axis.x * sin (angle / 2.0);
	result[1] = axis.y * sin (angle / 2.0);
	result[2] = axis.z * sin (angle / 2.0);
	result[3] = cos(angle / 2.0);
}

// --------------------------------------------------------------
// Convert unit-length quaternion to axis/angle pair.
// --------------------------------------------------------------
void from_quaternion (float q[4], VECTOR3 axis, float *angle)
{
	*angle = acos (q[3]) * 2.0;

	axis.x = q[0] / sin (*angle / 2.0);
	axis.y = q[1] / sin (*angle / 2.0);
	axis.z = q[2] / sin (*angle / 2.0);
}

// --------------------------------------------------------------
// Multiply two quaternions together.
// --------------------------------------------------------------
void multiply_quaternions (float q1[4], float q2[4], float result[4])
{
	result[0] = (q2[3] * q1[0]) + (q2[0] * q1[3]) + (q2[1] * q1[2]) - (q2[2] * q1[1]);
	result[1] = (q2[3] * q1[1]) + (q2[1] * q1[3]) + (q2[2] * q1[0]) - (q2[0] * q1[2]);
	result[2] = (q2[3] * q1[2]) + (q2[2] * q1[3]) + (q2[0] * q1[1]) - (q2[1] * q1[0]);
	
	result[3] = (q2[3] * q1[3]) - (q2[0] * q1[0]) - (q2[1] * q1[1]) - (q2[2] * q1[2]);
}

// --------------------------------------------------------------
// To rotate a vector in 3d space, angle is in radians
// --------------------------------------------------------------
VECTOR3 RotateVector (const VECTOR3& input, double angle, const VECTOR3& rotationaxis) 
{
	// To rotate a vector in 3D space we'll need to build a matrix, these are the variables required to do so.
	MATRIX3 rMatrix;
	double c = cos(angle);
	double s = sin(angle);
	double t = 1.0 - c;
	double x = rotationaxis.x;
	double y = rotationaxis.y;
	double z = rotationaxis.z;

	// Build rotation matrix
	rMatrix.m11 = (t * x * x + c);
	rMatrix.m12 = (t * x * y - s * z);
	rMatrix.m13 = (t * x * z + s * y);
	rMatrix.m21 = (t * x * y + s * z);
	rMatrix.m22 = (t * y * y + c);
	rMatrix.m23 = (t * y * z - s * x);
	rMatrix.m31 = (t * x * z - s * y);
	rMatrix.m32 = (t * y * z + s * x);
	rMatrix.m33 = (t * z * z + c);

	// Perform Rotation
	VECTOR3 output = mul (rMatrix, input); // multiply the input vector by our rotation matrix to get our output vector
	return output; // Return rotated vector
} // End "RotateVector"

// --------------------------------------------------------------
// Convert time to hours minutes and seconds. Time is in seconds 
// --------------------------------------------------------------
VECTOR3 to_HrsMinSec (double time)
{
	float neg;
	if (time < 0) neg = -1;
	else neg = 1;

	time = fabs(time);

	float hrs = (int)floor(time/3600.0);
	float min = (int)floor((time-3600.0*hrs)/60);
	float sec = (int)floor(time-(3600.0*hrs)-(60*min));\
	
	VECTOR3 output = {hrs, min, sec};
	return output * neg;
}

// --------------------------------------------------------------
// Convert time from hours minutes and seconds to total seconds. 
// --------------------------------------------------------------
double from_HrsMinSec (VECTOR3 time)
{
	double output = 0;

	output += time.x * 3600.0;
	output += time.y * 60;
	output += time.z;
	
	return output;
}

// --------------------------------------------------------------
// Convert MJD to Day, Month, and Year  
// --------------------------------------------------------------
VECTOR3 MJD_to_date (double MJD)
{
	int JD = MJD + 2400001;
	int j = JD + 32044;
	int g = j / 146097; 
	int dg = j % 146097;
	int c = (dg / 36524 + 1) * 3 / 4; 
	int dc = dg - c * 36524;
	int b = dc / 1461; 
	int db = dc % 1461;
	int a = (db / 365 + 1) * 3 / 4; 
	int da = db - a * 365;
	int y = g * 400 + c * 100 + b * 4 + a;	//note: this is the integer number of full years elapsed since March 1, 4801 BC at 00:00 UTC);
	int m = (da * 5 + 308) / 153 - 2;		//note: this is the integer number of full months elapsed since the last March 1 at 00:00 UTC);
	int d = da - (m + 4) * 153 / 5 + 122;	//note: this is the number of days elapsed since day 1 of the month at 00:00 UTC, including fractions of one day);

	int Year = y - 4800 + (m + 2) / 12; 
	int Month = (m + 2) % 12 + 1; 
	int Day = d + 1;

	VECTOR3 output = {Day, Month, Year};
	return output;
}

// ==============================================================
// Orbiter Generic Functions
// ==============================================================

// --------------------------------------------------------------
// Rotate an entire mesh around an arbitrary axis. Orginally written by Face as part of the "Ascension Ultra" addon.
// --------------------------------------------------------------
void RotateMesh (MESHHANDLE mesh, float angle, VECTOR3 rotationaxis, VECTOR3 ref)
{
	// Build rotation matrix (I modified this portion of Face's code for readability)
	double c = cos (angle);
	double s = sin (angle);
	double t = 1.0 - c;
	double x = rotationaxis.x;
	double y = rotationaxis.y;
	double z = rotationaxis.z;

	MATRIX3 rMatrix;
	rMatrix.m11 = (t * x * x + c);
	rMatrix.m12 = (t * x * y - s * z);
	rMatrix.m13 = (t * x * z + s * y);
	rMatrix.m21 = (t * x * y + s * z);
	rMatrix.m22 = (t * y * y + c);
	rMatrix.m23 = (t * y * z - s * x);
	rMatrix.m31 = (t * x * z - s * y);
	rMatrix.m32 = (t * y * z + s * x);
	rMatrix.m33 = (t * z * z + c);

	// Perform Rotation 
	int k = oapiMeshGroupCount (mesh);									// Get number of meshgroups in mesh and save it as "k"
	for (int i = 0; i < k; i++)											// Go through each mesh group in turn
	{
		MESHGROUP *m = oapiMeshGroup (mesh, i);							// Get group data and store it in "m".
		DWORD l = m->nVtx;												// Get number of vertices in "m" in and store it as "l".
		for (DWORD j = 0; j < l; j++)									// For each vertex...
		{
			VECTOR3 p = _V( m->Vtx[j].x, m->Vtx[j].y, m->Vtx[j].z);		// Store the original position of the vertex in p.
			VECTOR3 n = _V( m->Vtx[j].nx, m->Vtx[j].ny, m->Vtx[j].nz);	// Store the original normal of the vertex in n.
			p = mul ( rMatrix, (p-ref)) + ref;							// Shift p by ref, multiply it by rMatrix, and then shift the result back.
			n = mul ( rMatrix, (n-ref)) + ref;							// Do the same to n.

			// Write new value of p into the live structure.
			m->Vtx[j].x = p.x;
			m->Vtx[j].y = p.y;
			m->Vtx[j].z = p.z;

			// Write new value of n into the live structure.
			m->Vtx[j].nx = n.x;
			m->Vtx[j].ny = n.y;
			m->Vtx[j].nz = n.z;
		} // End vertext operations
	} // End meshgroup operations       
} // End void

// --------------------------------------------------------------
// Spawn a new vessel object (adapted from the Atlantis' ET seperation routine)
// --------------------------------------------------------------
OBJHANDLE SpawnObject (VESSEL *parent, char* name, char* classname, VECTOR3 ofs, double rvel) 
{
	VESSELSTATUS vs;
	parent->GetStatus (vs);

	parent->Local2Rel (ofs, vs.rpos);
	vs.eng_main = vs.eng_hovr = 0.0;
	vs.status = 0;

	VECTOR3 vel = { ofs.x, ofs.y, ofs.z};	// seperation vector
	normalise (vel);
	parent->Local2Rel (ofs, vs.rpos);		

	VECTOR3 rofs;
	parent->GlobalRot (vel, rofs);
	vs.rvel += rofs * rvel;

	oapiCreateVessel (name, classname, vs);
	return oapiGetObjectByName (name);
}

// --------------------------------------------------------------
// Detach vessel from parent
// --------------------------------------------------------------
bool DetachParent (VESSEL *child, double vel)
{
	OBJHANDLE oh_parent, oh_temp;
	oh_parent = NULL;
	int i;

	// Cycle through parent attachment points
	for (i = 0; i < child->AttachmentCount (true); i++)
	{
		oh_temp = child->GetAttachmentStatus (child->GetAttachmentHandle (true, i));
		if (oh_temp != NULL) oh_parent = oh_temp;
	}

	// If valid parent found...
	if (oh_parent != NULL)	
	{
		VESSEL *parent = oapiGetVesselInterface (oh_parent);		
		int attcount = parent->AttachmentCount (false);
		int idx = 0;
		
		// find the point we are attached to...
		for (i = 0; i < attcount; i++) if (parent->GetAttachmentStatus (parent->GetAttachmentHandle (false, i)) == child->GetHandle ()) idx = i;
		
		// ...and dettach us.
		return parent->DetachChild (parent->GetAttachmentHandle (false, idx), vel);
	}
	else return false;
}

// ==============================================================
// Calculate lift, moment, and zero-lift drag coefficients as a
// function of angle of attack (alpha or beta)
// Copied from Francisdrake's CEV addon (http://www.orbiter-forum.com/showthread.php?p=143325&postcount=4)
// ==============================================================

// Vertical component
void VLiftCoeff (double aoa, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*30.0;
	static const double istep = 1.0/step;
	static const int nabsc = 13;
	// Angle of attack -180 -150 -120 -90 -60 -30 0 30 60 90 120 150 180
	// static const double CL[nabsc] = { 0, -0.12, -0.1, 0, 0, 0, 0, 0, 0, 0, 0.1, 0.12, 0};
	// static const double CM[nabsc] = { 0, 0.0002, 0.0004, 0.0004, 0.0003, 0.0002, 0, -0.0002,-0.0003,-0.0004,-0.0004,-0.0002, 0};
	static const double CL[nabsc] = { 0, -0.12, -0.1, 0, 0, 0, 0, 0, 0, 0, 0.1, 0.12, 0};
	static const double CM[nabsc] = { 0, -0.0002,-0.0004, -0.0004,-0.0003,-0.0002, 0, 0.0002, 0.0003, 0.0004, 0.0004,0.0002, 0};
	// lift and moment coefficients from -180 to 180 in 30 degree steps.

	aoa += PI;
	int idx = max (0, min (11, (int)(aoa*istep)));
	double d = aoa*istep - idx;
	*cl = CL[idx] + (CL[idx+1]-CL[idx])*d;
	*cm = ( CM[idx] + (CM[idx+1]-CM[idx])*d ) / 10;
	*cd = 0.25 + oapiGetInducedDrag (*cl, 1.27, 0.3);
}

// Horizontal component (vertical stabiliser and body)
void HLiftCoeff (double beta, double M, double Re, double *cl, double *cm, double *cd)
{
	static const double step = RAD*30;
	static const double istep = 1.0/step;
	static const int nabsc = 13;
	static const double CL[nabsc] = { 0, -0.1, -0.1, 0, 0, 0, 0, 0, 0, 0, 0.1, 0.1, 0};

	beta += PI;
	int idx = max (0, min (11, (int)(beta * istep)));
	double d = beta * istep - idx;
	*cl = CL[idx] + (CL[idx + 1] - CL[idx])*d;
	*cm = 0.0;
	*cd = 0.25 + oapiGetInducedDrag (*cl, 1.27, 0.3);
}

// Alternate coefficient function
void CoeffFunc (double aoa, double M, double Re, double *cl, double *cm, double *cd)
{
	const int nlift = 11;
	
	double frac, drag, lift;
	double factor = 0.0;
	double dfact = 0.9;
	double lfact = 1.0;

	static const double AOA[nlift]	= {-180.*RAD, -160.*RAD, -150.*RAD, -120.*RAD, -90.*RAD, 0*RAD, 90.*RAD, 120.*RAD, 150.*RAD, 160.*RAD, 180.*RAD};
	static const double Mach[17]	= {0.0, 0.7, 0.9, 1.1, 1.2, 1.35, 1.65, 2.0, 3.0, 5.0, 8.0, 10.5, 13.5, 18.2, 21.5, 31.0, 50.0};
	static const double LFactor[17] = {0.3, 0.392, 0.466, 0.607, 0.641, 0.488, 0.446, 0.435, 0.416, 0.415, 0.405, 0.400, 0.385, 0.385, 0.375, 0.35, 0.33};
	static const double DFactor[17] = {0.9, 0.944, 0.991, 1.068, 1.044, 1.270, 1.28, 1.267, 1.213, 1.134, 1.15, 1.158, 1.8, 1.8, 1.193, 1.224, 1.25};
	static const double CL[nlift]	= {0.0, -0.9, -1.1, -0.5, 0.0, 0.0, 0.0, 0.5, 1.1, 0.9, 0.0};
	static const double CM[nlift]	= {0.0, 0.004, 0.006, 0.012, 0.015, 0.0, -0.015, -0.012, -0.006, -0.004, 0.};
	static const double CD[nlift]	= {1.143, 1.0, 1.0, 0.8, 0.8, 0.8, 0.8, 0.8, 1.0, 1.0, 1.143};
	
	int j;

	for (j = 0; (j < 16) && (Mach[j+1] < M); j++);
	frac = (M - Mach[j]) / (Mach[j+1] - Mach[j]);
	drag = dfact * (frac * DFactor[j+1] + (1.0 - frac) * DFactor[j]);
	lift = drag * lfact * (frac * LFactor[j+1] + (1.0-frac) * LFactor[j]);

	for (j = 0; (j < nlift-1) && (AOA[j+1] < aoa); j++);
	frac = (aoa - AOA[j]) / (AOA[j+1] - AOA[j]);
	*cd = drag * (frac * CD[j+1] + (1.0-frac) * CD[j]);
	*cl = lift * (frac * CL[j+1] + (1.0-frac) * CL[j]);
	*cm = factor * (frac * CM[j+1] + (1.0-frac) * CM[j]);
}

// ==============================================================
// AAPO Specific Functions
// ==============================================================

// --------------------------------------------------------------
// Switch interface logic
// --------------------------------------------------------------
bool ActivateSwitch (int event, enum switchstate* toggle, int npos, int sound, int OrbiterSoundID)
{
	// Get current switch position
	int state = *toggle;
	int newstate = 0;

	// If switch has 2 or fewer defined positions, process it as a binary (on/off) switch...
	if (npos <= 2)
	{
		npos = 2;												// Sanity check, a switch must have at least two positions in order to be valid
		if (event & PANEL_MOUSE_LBDOWN) newstate = UP;			// If left mouse button set switch to "UP" position
		else if (event & PANEL_MOUSE_RBDOWN) newstate = DOWN;	// Else if right mouse button set switch to "DOWN" position
	}

	// ...if not, process it as a multi-position switche/dial
	else
	{
		// If left mouse button increase state by 1.
		if (event & PANEL_MOUSE_LBDOWN) newstate = state + 1;

		// Else if right mouse button and "state" greater than 0, decrease it.
		if (event & PANEL_MOUSE_RBDOWN) newstate = state - 1;

		// Sanity check, restrict value of "state" to valid switch positions.
		if (newstate < 0) newstate = 0;
		else if (newstate > npos - 1) newstate = npos - 1; 
	}

	// Update switch position
	if (SetSwitchState (toggle, newstate))
	{
		PlayVesselWave (OrbiterSoundID, sound);
		return true;
	}
	else return false;
}

// --------------------------------------------------------------
// Set switch state
// --------------------------------------------------------------
bool SetSwitchState (enum switchstate* toggle, int newstate)
{
	int state = *toggle;
	if (state != newstate)
	{
		*toggle = (switchstate)newstate;
		return true;
	}
	else return false;
}

// --------------------------------------------------------------
// Load Sound files  
// --------------------------------------------------------------
void LoadSoundFiles (int id)
{
	RequestLoadVesselWave (id, SILENCE, "Sound\\_CustomVesselsSounds\\AAPO\\silence.wav", DEFAULT);
	RequestLoadVesselWave (id, SOUND_BTN, "Sound\\_CustomVesselsSounds\\AAPO\\switch_button.wav", INTERNAL_ONLY);
	RequestLoadVesselWave (id, SOUND_ROTARY, "Sound\\_CustomVesselsSounds\\AAPO\\switch_rotary.wav", INTERNAL_ONLY);
	RequestLoadVesselWave (id, SOUND_SWITCH, "Sound\\_CustomVesselsSounds\\AAPO\\switch_click.wav", INTERNAL_ONLY);
	RequestLoadVesselWave (id, SOUND_SWITCHGUARD, "Sound\\_CustomVesselsSounds\\AAPO\\switch_guard.wav", INTERNAL_ONLY);
	RequestLoadVesselWave (id, SOUND_HATCHOPEN, "Sound\\_CustomVesselsSounds\\AAPO\\hatch_open.wav", BOTHVIEW_FADED_MEDIUM);
	RequestLoadVesselWave (id, SOUND_HATCHCLOSE, "Sound\\_CustomVesselsSounds\\AAPO\\hatch_close.wav", BOTHVIEW_FADED_MEDIUM);
	RequestLoadVesselWave (id, SOUND_EXPLSVDEVICE, "Sound\\_CustomVesselsSounds\\AAPO\\cad.wav", BOTHVIEW_FADED_MEDIUM);
	RequestLoadVesselWave (id, SOUND_CAUTION, "Sound\\_CustomVesselsSounds\\AAPO\\alarm_caution.wav", RADIO_SOUND);
	RequestLoadVesselWave (id, SOUND_WARNING, "Sound\\_CustomVesselsSounds\\AAPO\\alarm_warning.wav", RADIO_SOUND);
	RequestLoadVesselWave (id, SOUND_SPLASH, "Sound\\_CustomVesselsSounds\\AAPO\\splash.wav", BOTHVIEW_FADED_FAR);
	RequestLoadVesselWave (id, SOUND_SEPERATION, "Sound\\_CustomVesselsSounds\\AAPO\\seperation.wav", BOTHVIEW_FADED_MEDIUM);
	RequestLoadVesselWave (id, SOUND_STAGESEP, "Sound\\_CustomVesselsSounds\\AAPO\\stage_sep.wav", BOTHVIEW_FADED_MEDIUM);
	RequestLoadVesselWave (id, SOUND_SMSEP, "Sound\\_CustomVesselsSounds\\AAPO\\sm_sep.wav", BOTHVIEW_FADED_MEDIUM);
	RequestLoadVesselWave (id, SOUND_TOWERJET, "Sound\\_CustomVesselsSounds\\AAPO\\les_sep.wav", BOTHVIEW_FADED_MEDIUM);
}