// ==============================================================
//						ORBITER MODULE:
//                  Part of the ORBITER SDK
//               Copyright (C) 2013 Greg Hlynka
//                    All rights reserved
//
// AAPO_Common.h
// Common variables and functions used in the Apollo Aplication 
// Project for Orbiter (AAPO).
//
// ==============================================================

#pragma once
#ifndef AAPO_COMMON
#define AAPO_COMMON

#define STRICT
#define ORBITER_MODULE

#include <math.h>
#include <vector>

// SDK files
#include "orbitersdk.h"
#include "OrbiterSoundSDK40.h"
#include "UmmuSDK.h"
#include "UCGOCargoSDK.h"

// ==============================================================
// Sound Definitions
// ==============================================================

#define SILENCE				00 // "silence.wav" The sound of...
#define SOUND_BTN			01 // Button press "button.wav"
#define SOUND_ROTARY		02 // Dial click "rotary.wav"
#define SOUND_SWITCH		03 // Switch sound "switch_throw.wav"
#define SOUND_SWITCHGUARD	04 // Alternate Switch sound "switch_guard.wav"
#define SOUND_HATCHOPEN		05 // EVA Hatch opening "hatch_open.wav"
#define SOUND_HATCHCLOSE	06 // EVA Hatch closing "hatch_close.wav"
#define SOUND_EXPLSVDEVICE	07 // Explosive device firing "cad.wav"
#define SOUND_CAUTION		10 // Basic alert tone "alarm_caution.wav"
#define SOUND_WARNING		11 // OH SHIT! alarm tone "alarm_warning.wav"
#define SOUND_SPLASH		12 // Splashdown! "splash.wav"
#define SOUND_SEPERATION	13 // Booster / SLA Seperation "sla_sep.wav"
#define SOUND_STAGESEP		14 // Stage Seperation "stage_sep.wav"
#define SOUND_SMSEP			15 // Module Seperation "SM_sep.wav"
#define SOUND_TOWERJET		16 // Abort Tower jettison "LES_sep.wav"

// ==============================================================
// Physical Constants
// ==============================================================

const double	C					= 299792458.0;		// Speed of light in vacuum [m/s]
const double	GRAV				= 6.67259e-11;		// Gravitational constant [m^3 kg^-1 s^-2]
const double	STEFANBOLTZMANN		= 5.67e-8;			// Stephan-Boltzman constant for radiance of black bodies
const double	REGNAULT			= 8.314462175;		// Regnault (Ideal) gas constant 

const double	PA					= 6894.75729;		// Convert PSI to Pascals
const double	PSI					= 1 / PA;			// Convert Pascals to PSI
const double	METERS				= 0.3048;			// Convert meters to feet
const double	FEET				= 1 / METERS;		// Convert of feet to meters
const double	KILOS				= 0.45359237;		// Convert lbs to kg
const double	LBS					= 1 / KILOS;		// Convert kg to lbs
const double	NEWTONS				= 4.4482216153;		// Convert lbf to n
const double	LBFORCE				= 1 / NEWTONS;		// Convert n to lbf

const double	H2_MOLAR_MASS		= 2.0158 / 1000;	// H2 molar mass [kg] 2 grams
const double	H2_HEAT				= 14.20;			// Specific heat of H2 [kj/kg K]

const double	N2_MOLAR_MASS		= 28.01344 / 1000;	// N2 molar mass [kg] 28 grams
const double	N2_HEAT				= 1.039;			// Specific heat of N2 [kj/kg K]

const double	O2_MOLAR_MASS		= 31.99886 / 1000;	// O2 molar mass [kg] 32 grams
const double	O2_HEAT				= 0.919;			// Specific heat of O2 [kj/kg K]

const double	CO2_MOLAR_MASS		= 44.00964 / 1000;	// CO2 Molar mass [kg] 44 grams
const double	CO2_HEAT			= 0.037;			// Specific heat of CO2 [kj/kg K]

const double	H2O_MOLAR_MASS		= 18.01528 / 1000;	// Water Molar mass [kg] 18 grams
const double	H2O_HEAT			= 4.187;			// Specific heat of Water [kj/kg K]

const VECTOR3	NULL_VECTOR			= { 0.0, 0.0, 0.0};	// Null vector

// ==============================================================
// Enumerators
// ==============================================================
enum chutestate		{STOWED, DROGUE, REEFED, UENREEFING, DEPLOYED, JETTISONED};
enum switchstate	{DOWN, CENTER, UP};
enum doorstate		{CLOSED, OPEN, CLOSING, OPENING};

// ==============================================================
// Shared Functions
// ==============================================================

// Math Functions
void		to_quaternion (VECTOR3 axis, float angle, float result[4]);								// Convert axis/angle to quaternion form, angle is in radians.
void		from_quaternion (float q[4], VECTOR3 axis, float *angle);								// Convert unit-length quaternion to axis / angle pair.
void		multiply_quaternions (float q1[4], float q2[4], float result[4]);						// Multiply two quaternions together.
VECTOR3		RotateVector (const VECTOR3& input, double angle, const VECTOR3& rotationaxis);			// Rotate a vector in 3d space, angle is in radians
VECTOR3		to_HrsMinSec (double time);																// Convert time to hours minutes and seconds. Time is in seconds 
double		from_HrsMinSec (VECTOR3 time);															// Convert time from hours minutes and seconds to total seconds. 
VECTOR3		MJD_to_date (double MJD);																// Convert MJD to Day, Month, and Year  

// Orbiter Generic Functions
void		RotateMesh (MESHHANDLE mesh, float angle, VECTOR3 rotationaxis, VECTOR3 ref);			// Rotate an entire mesh around an arbitrary axis.
OBJHANDLE	SpawnObject (VESSEL *parent, char* name, char* classname, VECTOR3 ofs, double rvel);	// Spawn a new vessel object (adapted from the Atlantis' ET seperation routine)
bool		DetachParent (VESSEL *child, double vel);												// Detach a vessel from it's parent
void		VLiftCoeff (double aoa, double M, double Re, double *cl, double *cm, double *cd);		// Calculate vertical lift coefficient
void		HLiftCoeff (double beta, double M, double Re, double *cl, double *cm, double *cd);		// Calculate horizontal lift coefficient (for fins and vertical stabilisers)
void		CoeffFunc (double aoa, double M, double Re, double *cl, double *cm, double *cd);		// Alternate lift coefficient function copied from NASSP

// AAPO Specific functions
bool		ActivateSwitch (int event, enum switchstate* toggle, int npos, int sound, int OrbiterSoundID);	// Switch control logic			
bool		SetSwitchState (enum switchstate* toggle, int newstate);								// Set switchstate
void		LoadSoundFiles (int id);																// Load sound files

#endif