// ==============================================================
//		ORBITER MODULE: 'SPIDER' LUNAR EXCURSION MODULE
//			A custom Vessel For Orbiter 2010/2011
//  
// FDAI.cpp
// Implimentation of FDAI Orbiter vessel sub-system class
// ==============================================================

#include "AAPO_FDAI.h"

// Constructor
FDAI::FDAI (const VESSEL *vessel, const VECTOR3 pos, double biasangle, VECTOR3 biasaxis)
{
	// Set starting values
	v = vessel;
	ball_pos = pos;
	SetBias (biasangle, biasaxis);
	Projection = 0;
	NavChan = 0;

	SetRefMode (0);
	SetTgtMode (0);
	SetErrorScale (RAD*15);
	SetRateScale (RAD*5);

	PowerOn = true;

	curr_rho = 0; 
	curr_theta = 0; 
	curr_phi = 0;

	balloffset = NULL_VECTOR;
	tgtoffset = NULL_VECTOR;

	// Generate visual instance
	MESHHANDLE mh_orig = oapiLoadMeshGlobal ("AAPO/FDAI");
	mgh_ADI_orig = oapiMeshGroupEx (mh_orig, 0); 
	mesh_FDAI = v->AddMesh (mh_orig, &ball_pos);
	v->SetMeshVisibilityMode (mesh_FDAI, MESHVIS_VC);

	// Define transformations
	static UINT	meshgroups_FDAI[11] = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11};
	static MGROUP_ROTATE mgt_bias (mesh_FDAI, meshgroups_FDAI, 11, _V( 0,0,0), biasaxis, biasangle); 
	
	anim_bias = v->CreateAnimation (0.0);
	ANIMATIONCOMPONENT_HANDLE ach_bias = v->AddAnimationComponent (anim_bias, 0.0f, 1.0f, &mgt_bias);
	v->SetAnimation (anim_bias, 1.0);

	// Rate needles
	static UINT	meshgroup_pitchrate = {FDAI_GRP_RateNeedle_Pitch};
	static UINT	meshgroup_rollrate = {FDAI_GRP_RateNeedle_Roll};
	static UINT	meshgroup_yawrate = {FDAI_GRP_RateNeedle_Yaw};

	static MGROUP_TRANSLATE  mgt_pitchrate (mesh_FDAI, &meshgroup_pitchrate, 1, mul(bias, _V( 0.0, FDAI_NEEDLERANGE, 0.0)));
	anim_PitchRate = v->CreateAnimation (0.5);
	v->AddAnimationComponent (anim_PitchRate, 0, 1, &mgt_pitchrate);
	
	static MGROUP_TRANSLATE mgt_rollrate (mesh_FDAI, &meshgroup_rollrate, 1, mul(bias, _V( FDAI_NEEDLERANGE, 0.0, 0.0)));
	anim_RollRate = v->CreateAnimation (0.5);
	v->AddAnimationComponent (anim_RollRate, 0, 1, &mgt_rollrate);
	
	static MGROUP_TRANSLATE mgt_yawrate (mesh_FDAI, &meshgroup_yawrate, 1, mul(bias, _V(-FDAI_NEEDLERANGE, 0.0, 0.0)));
	anim_YawRate = v->CreateAnimation (0.5);
	v->AddAnimationComponent (anim_YawRate, 0, 1, &mgt_yawrate);

	// Error needles
	static UINT	meshgroup_pitcherror = {FDAI_GRP_ErrorNeedle_Pitch};
	static UINT	meshgroup_rollerror = {FDAI_GRP_ErrorNeedle_Roll};
	static UINT	meshgroup_yawerror = {FDAI_GRP_ErrorNeedle_Yaw};

	static MGROUP_TRANSLATE  mgt_pitcherror (mesh_FDAI, &meshgroup_pitcherror, 1, mul(bias, _V( 0.0, FDAI_NEEDLERANGE, 0.0)));
	anim_PitchError = v->CreateAnimation (0.5);
	v->AddAnimationComponent (anim_PitchError, 0, 1, &mgt_pitcherror);
	
	static MGROUP_TRANSLATE mgt_rollerror (mesh_FDAI, &meshgroup_rollerror, 1, mul(bias, _V( FDAI_NEEDLERANGE, 0.0, 0.0)));
	anim_RollError = v->CreateAnimation (0.5);
	v->AddAnimationComponent (anim_RollError, 0, 1, &mgt_rollerror);
	
	static MGROUP_TRANSLATE mgt_yawerror (mesh_FDAI, &meshgroup_yawerror, 1, mul(bias, _V(-FDAI_NEEDLERANGE, 0.0, 0.0)));
	anim_YawError = v->CreateAnimation (0.5);
	v->AddAnimationComponent (anim_YawError, 0, 1, &mgt_yawerror);
}

// ==============================================================
// Public functions
// ==============================================================

// --------------------------------------------------------------
// Build a rotation matrix to compensate for panel bias. 
// NOTE: angle should be in radians and rotationaxis defined in vessel's local coordinates.
// --------------------------------------------------------------
bool FDAI::SetBias (double angle, VECTOR3 rotationaxis)
{
	MATRIX3 rMatrix;
	double c = cos(angle);
	double s = sin(angle);
	double t = 1.0 - c;
	double x = rotationaxis.x;
	double y = rotationaxis.y;
	double z = rotationaxis.z;

	// Build rotation matrix
	rMatrix.m11 = (t * x * x + c);
	rMatrix.m12 = (t * x * y - s * z);
	rMatrix.m13 = (t * x * z + s * y);
	rMatrix.m21 = (t * x * y + s * z);
	rMatrix.m22 = (t * y * y + c);
	rMatrix.m23 = (t * y * z - s * x);
	rMatrix.m31 = (t * x * z - s * y);
	rMatrix.m32 = (t * y * z + s * x);
	rMatrix.m33 = (t * z * z + c);

	// Assign bias
	bias = rMatrix;
	return true;
}

// Set Error Scale 
bool FDAI::SetErrorScale (double scale)
{
	// Sanity check
	if (scale <= 0) scale = RAD;

	// Apply change if required
	if (ErrorScale != scale)
	{
		ErrorScale = scale;
		return true;
	}
	else return false;
}

// Set projection mode
bool FDAI::SetProjMode (int newmode)
{
	// Sanity check
	if (newmode < 0) newmode = 0;
	else if (newmode > 1) newmode = 1;

	// Apply change if required
	if (Projection != newmode)
	{
		Projection = newmode;
		return true;
	}
	else return false;
}

// Set reference frame 
bool FDAI::SetRefMode (int newmode)
{
	// Sanity check
	if (newmode < 0) newmode = 0;
	else if (newmode > 4) newmode = 4;

	// Apply change if required
	if( RefMode != (Reference)newmode)
	{
		RefMode = (Reference)newmode;
		return true;
	}
	else return false;
}

// Set Rate Scale 
bool FDAI::SetRateScale (double scale)
{
	// Sanity check
	if (scale <= 0) scale = RAD;

	// Apply change if required
	if (RateScale != scale)
	{
		RateScale = scale;
		return true;
	}
	else return false;
}

// Set target mode
bool FDAI::SetTgtMode (int newmode)
{
	// Sanity check
	if (newmode < 0) newmode = 0;
	else if (newmode > 3) newmode = 3;

	// Apply change if required
	if (TgtMode != (Mode)newmode)
	{
		TgtMode = (Mode)newmode;
		return true;
	}
	else return false;
}

// Set target Nav Channel
bool FDAI::SetNavChan (int newnavchan)
{
	// Sanity check
	if (newnavchan < 0) newnavchan = 0;

	// Apply change if required
	if(NavChan != newnavchan)
	{
		NavChan = newnavchan;
		return true;
	}
	else return false;
}

// Set ADI ball offset
bool FDAI::SetBallOffset (VECTOR3 newoffset)
{
	// Convert input from degrees to radians 
	VECTOR3 output;
	output.x = newoffset.x*RAD;
	output.y = newoffset.y*RAD;
	output.z = newoffset.z*RAD;

	// Normalise angles
	for (int i = 0; i < 3; i++) output.data[i] = posangle (output.data[i]);

	// Apply change
	balloffset = output;
	return true;
}

// Set target (error-needle) offset
bool FDAI::SetTargetOffset (VECTOR3 newoffset)
{
	// Convert input from degrees to radians 
	VECTOR3 output;
	output.x = newoffset.x*RAD;
	output.y = newoffset.y*RAD;
	output.z = newoffset.z*RAD;

	// Normalise angles
	for (int i = 0; i < 3; i++) output.data[i] = posangle(output.data[i]);

	// Apply change
	tgtoffset = output;
	return true;

}

// ==============================================================
// Callback functions 
// NOTE:	These functions will need to be called manually from within the Orbiter default callback functions
//			of the vessel to which this ADI ball belongs
// ==============================================================

// Create Visuals
void FDAI::VisualCreated (VISHANDLE vis, int refcount)
{
	dmh_FDAI = v->GetDevMesh (vis, mesh_FDAI);			
	mgh_ADI_current = oapiMeshGroupEx (dmh_FDAI, 0); 
}

// Clear visuals
void FDAI::VisualDestroyed (VISHANDLE vis, int refcount)
{
	mgh_ADI_current = NULL; // Nothing to rotate without visuals...
}

// Post Step Processes
void FDAI::PostStep (double simt, double simdt, double mjd)
{
	// Get handle for the celestial body currently being orbited.
	oh_rBody = v->GetGravityRef ();

	// Get handle of Nav Source
	nh_NavTarget = v->GetNavSource (NavChan);
	if (nh_NavTarget == NULL) TgtMode = NO_TARGET;

	if (PowerOn)
	{
		// Rotate ADI ball mesh
		AnimateADI (GetEulerAngles ());

		// Animate Roll Indicator
		//**Nothing here yet**

		// Animate Error Needles
		AnimateErrorNeedles (GetRotationRates ());
	
		// Animate Rate Needles
		AnimateRateNeedles (GetRotationRates ());
	}
}

// ==============================================================
// Private functions
// ==============================================================

// Calculate rotations
//void FDAI::RotateVector (VECTOR3 &Initial, VECTOR3 &Angle, VECTOR3 &Result)
//{
//	// Temporary variables
//	MATRIX3 RotMatrixX, RotMatrixY, RotMatrixZ;
//	VECTOR3 AfterZ, AfterZY;
//
//	// X axis
//	RotMatrixX.m11 = 1;
//	RotMatrixX.m12 = 0;
//	RotMatrixX.m13 = 0;
//	RotMatrixX.m21 = 0;
//	RotMatrixX.m22 = cos(Angle.x);
//	RotMatrixX.m23 = sin(Angle.x);
//	RotMatrixX.m31 = 0;
//	RotMatrixX.m32 = -sin(Angle.x);
//	RotMatrixX.m33 = cos(Angle.x);
//
//	// Y Axis
//	RotMatrixY.m11 = cos(Angle.y);
//	RotMatrixY.m12 = 0;
//	RotMatrixY.m13 = -sin(Angle.y);
//	RotMatrixY.m21 = 0;
//	RotMatrixY.m22 = 1;
//	RotMatrixY.m23 = 0;
//	RotMatrixY.m31 = sin(Angle.y);
//	RotMatrixY.m32 = 0;
//	RotMatrixY.m33 = cos(Angle.y);
//
//	// Z Axis
//	RotMatrixZ.m11 = cos(Angle.z);
//	RotMatrixZ.m12 = sin(Angle.z);
//	RotMatrixZ.m13 = 0;
//	RotMatrixZ.m21 = -sin(Angle.z);
//	RotMatrixZ.m22 = cos(Angle.z);
//	RotMatrixZ.m23 = 0;
//	RotMatrixZ.m31 = 0;
//	RotMatrixZ.m32 = 0;
//	RotMatrixZ.m33 = 1;
//
//	// Apply transformation	
//	AfterZ	= mul(RotMatrixZ, Initial);
//	AfterZY	= mul(RotMatrixY, AfterZ); 
//	Result	= mul(RotMatrixX, AfterZY);
//}

// Return a rotation matrix for translation from reference frame to global frame
MATRIX3	FDAI::GetFrameRotMatrix ()
{
	MATRIX3	R;
	VECTOR3 axis1, axis2, axis3;
	VECTOR3 localVert, localHorizon;
	VECTOR3 vv, vm;
	NAVDATA	ndata;

	switch (RefMode) 
	{
	case ECLIPTIC: // Inertial ecliptic (Orbiter Global) Reference frame 
		axis3 = _V( 1, 0, 0);
		axis2 = _V( 0, 1, 0);
		break;

	case EQUATORIAL: // Inertial equatorial (Geocentric) Reference frame   
		oapiGetPlanetObliquityMatrix (v->GetGravityRef (), &R);
		axis3 = _V(R.m11, R.m21, R.m31);
		axis2 = _V(R.m12, R.m22, R.m32);
		break;

	case ORBIT: // Orientation relative to orbital momentum (Prograde) vector 
		v->GetRelativeVel (oh_rBody, axis3);	// Get velocity of vessel relative to reference body.
		v->GetRelativePos (oh_rBody, vv);		// local vertical
		vm = crossp (axis3, vv);				// direction of orbital momentum
		axis2 = unit (crossp (vm, axis3));
		break;

	case SURFACE: // Surface relative (local horizon)
		{
		OBJHANDLE hRef = v->GetSurfaceRef ();
		v->GetRelativePos (hRef, axis2);
		axis2 = unit (axis2);
		MATRIX3 prot;
		oapiGetRotationMatrix (hRef, &prot);
		VECTOR3 paxis = {prot.m12, prot.m22, prot.m32};	// planet rotation axis in global frame
		VECTOR3 yaxis = unit (crossp (paxis,axis2));	// direction of yaw=+90 pole in global frame
		axis3 = crossp (axis2,yaxis);
		}
		break;

	case NAV_SOURCE: // Synced to NAV source (type-specific)
		axis3 = _V( 0, 0, 1);
		axis2 = _V( 0, 1, 0);
		if (nh_NavTarget) 
		{
			oapiGetNavData (nh_NavTarget, &ndata);
			switch (ndata.type) 
			{
			case TRANSMITTER_IDS: 
				{
				VECTOR3 pos, dir, rot;
				MATRIX3 R;
				VESSEL *vtgt = oapiGetVesselInterface (ndata.ids.hVessel);
				vtgt->GetRotationMatrix (R);
				vtgt->GetDockParams (ndata.ids.hDock, pos, dir, rot);
				
				axis3 = -mul (R,dir);
				axis2 = mul (R,rot);
				} 
				break;
			
			case TRANSMITTER_VTOL:
			case TRANSMITTER_VOR: 
				{
				OBJHANDLE hRef = v->GetSurfaceRef ();
				VECTOR3 spos, npos;
				v->GetRelativePos (hRef, axis2);
				v->GetGlobalPos (spos);
				axis2 = unit (axis2);
				oapiGetNavPos (nh_NavTarget, &npos);
				npos -= spos;
				axis3 = unit (crossp (crossp (axis2, npos), axis2));
				} 
				break;
			} // End "switch (ndata.type)'
		} // End "if (nh_NavTarget)"
		break;
	} //End "switch (RefMode)"

	// axis1 = cross product of axis2 and axis3
	axis1 = crossp (axis2, axis3);

	// Normalise vectors
	normalise (axis1);
	normalise (axis2);
	normalise (axis3);

	// Construct matrix
	MATRIX3 output = _M( axis1.x, axis2.x, axis3.x,  axis1.y, axis2.y, axis3.y,  axis1.z, axis2.z, axis3.z);
	return output;
}

// Calculate Euler Rotations (in radians) from current orientation to reference frame
VECTOR3 FDAI::GetEulerAngles ()
{
	VECTOR3 euler;

	// Update reference frame
	MATRIX3 Rref = GetFrameRotMatrix();

	// Get vessels current rotation matrix
	MATRIX3 srot;
	v->GetRotationMatrix(srot);

	// map ship's local axes to reference frame
	VECTOR3 shipx = {srot.m11, srot.m21, srot.m31};
	VECTOR3 shipy = {srot.m12, srot.m22, srot.m32};
	VECTOR3 shipz = {srot.m13, srot.m23, srot.m33};
	shipx = tmul (Rref, shipx);
	shipy = tmul (Rref, shipy);
	shipz = tmul (Rref, shipz);

	// Martins values
	if (Projection == 0)
	{
		euler.x = atan2(shipx.y, shipy.y);  // roll angle
		euler.y = asin(shipz.y);            // pitch angle
		euler.z = atan2(shipz.x, shipz.z);  // yaw angle
	}
	else
	{
		euler.x = -atan2(shipy.x, shipx.x); // roll angle
		euler.y = atan2(shipz.y,shipz.z);   // pitch angle
		euler.z = asin(shipz.x);            // yaw angle
	}

	// Apply offset and normalize output angles
	VECTOR3 output = euler + balloffset;
	//for (int i = 0; i < 3; i++) output.data[i] = posangle(output.data[i]);

	//// Debuggery
	//sprintf(oapiDebugString(),"X %0.2f, Y %0.2f, Z %0.2f, Proj %f", output.x*DEG, output.y*DEG, output.z*DEG, Projection);

	// Return output
	return output;
}

// Calculate target euler angles
VECTOR3 FDAI::GetTgtEulerAngles (NAVHANDLE hNav)
{	
	NAVDATA data;
	OBJHANDLE hObj = NULL;

	VECTOR3 tgteuler, output;
	VECTOR3 dir, sdir;
	VECTOR3 gPos, gVel;
	VECTOR3 tgt_gPos, tgt_gVel, tgt_rVel;

	v->GetGlobalVel(gVel);
	oapiGetNavData (hNav, &data);

	// Process Nav Data
	if (hNav)
	{
		switch(data.type)
		{
		case TRANSMITTER_IDS:
			hObj = data.ids.hVessel;
			break;
		
		case TRANSMITTER_XPDR:
			hObj = data.xpdr.hVessel;
			break;
		
		case TRANSMITTER_VTOL:
			hObj = data.vtol.hBase;
			break;
		
		case TRANSMITTER_VOR:
			hObj = data.vor.hPlanet;
			MATRIX3 Rp;
			oapiGetRotationMatrix( hObj, &Rp);
			oapiGetGlobalVel (hObj, &tgt_gVel);
			
			//tvel += mul ( Rp, _V(-sin(data.vor.lng),0,cos(data.vor.lng)) * PI2/oapiGetPlanetPeriod(hObj)*oapiGetSize(hObj)*cos(data.vor.lat));
			
			tgt_rVel = gVel - tgt_gVel;

			sprintf( oapiDebugString(), "rvel: x=%f, y=%f, z=%f", tgt_rVel.x, tgt_rVel.y, tgt_rVel.z);
			break; 

		} // End "switch(data.type)"

		if (hObj)
		{
			oapiGetGlobalVel (hObj, &tgt_gVel);
			tgt_rvel = gVel-tgt_gVel;
		}
	} // End "if (hNav)"

	// Calculate Angles
	switch (TgtMode) 
	{
	case NO_TARGET: // No target
		tgteuler = _V( 0.0, 0.0, 0.0); // do nothing
		break;

	case FIXED: // fixed offset
		tgteuler = GetEulerAngles(); // starget euler angles same as vessel's 
		break;

	case TGT_DIRECTION: // direction to NAV source
		oapiGetNavPos( hNav, &dir);
		v->GetGlobalPos(sdir);
		dir = tmul( GetFrameRotMatrix(), unit(dir-sdir));
		break;

	case TGT_RVEL: // relative velocity of NAV source
		v->GetGlobalVel (sdir);
		dir = tmul( GetFrameRotMatrix(), unit(tgt_rvel));
		break;
	} // End "switch(TgtMode)" 

		
	if (Projection == 0)
	{
		tgteuler.y = asin(dir.y);
		tgteuler.z = atan2(dir.x, dir.z);
	}
	else
	{
		tgteuler.y = atan2(dir.y, dir.z);
		tgteuler.z = asin(dir.x);
	}
		
	//tgteuler += GetEulerAngles (projection_mode);
	// frame-relative --> ship-relative; is this a good idea?
	tgteuler.y = posangle(tgteuler.y);
	tgteuler.z = posangle(tgteuler.z);

	output = tgteuler += tgtoffset;
	return output;
}

// Calculate rotation rates
VECTOR3 FDAI::GetRotationRates ()
{
	VECTOR3 output = NULL_VECTOR;
	v->GetAngularVel (output);
	return output;
}

// Rotate ADI Ball
void FDAI::AnimateADI (VECTOR3 euler)
{
	double	rho		= euler.x;			// roll angle
	double	theta	=-euler.y;			// pitch angle
	double	phi		= euler.z;			// yaw angle
	double	simdt	= oapiGetSimStep();	// Get length of timestep

	if (mgh_ADI_current != NULL) // Pointer sanity check
	{
		// the following lines were copied from "FDAI.cpp" in the ShuttleA's SDK sample.
  		double dangle_max = FDAI_BALLSPEED * simdt;	// Scale ball movement speed to timestep
 
		// Restrict values of Rho to +/- PI (180 degrees)
 		double drho = rho - curr_rho;		
 		if (drho > PI) drho -= PI2;
 		else if (drho < -PI) drho += PI2;
 
		// Restrict values of Theta to +/- PI (180 degrees)
 		double dtht = theta - curr_theta;
 		if (dtht > PI) dtht -= PI2;
 		else if (dtht < -PI) dtht += PI2;

		// Restrict values of Phi to +/- PI (180 degrees)
 		double dphi = phi-curr_phi;			
 		if (dphi > PI) dphi -= PI2;
 		else if (dphi < -PI) dphi += PI2;
 
		// Limit rate of change to ball's movement speed.
 		double dangle = max (fabs(drho), max (fabs(dtht), fabs(dphi)));
 		if (dangle > dangle_max)
 		{
 			double scale = dangle_max/dangle;
 			rho = curr_rho + drho*scale;
 			theta = curr_theta + dtht*scale;
 			phi = curr_phi + dphi*scale;
 		}
		// End copy/paste

 		// save current angles
 		curr_rho = rho;
 		curr_theta = theta;
 		curr_phi = phi;
 
 		// Construct rotation matrix for Ball transformations
		double	sinr = sin(rho);
		double	cosr = cos(rho);
 		double	sint = sin(theta);
		double	cost = cos(theta);
 		double	sinp = sin(phi);
		double	cosp = cos(phi);

		double a1, b1, c1, a2, b2, c2, a3, b3, c3;

		if (Projection == 0)
		{
			// below are the coefficients of the rows of the rotation matrix M for the ball,
			// given by M = RTP, with
			// MATRIX3 P = {cosp,0,-sinp,  0,1,0,  sinp,0,cosp};
			// MATRIX3 T = {1,0,0,  0,cost,-sint,  0,sint,cost};
			// MATRIX3 R = {cosr,sinr,0,  -sinr,cosr,0,  0,0,1};

			a1 =  cosr*cosp - sinr*sint*sinp;
			b1 =  sinr*cost;
			c1 = -cosr*sinp - sinr*sint*cosp;
			a2 = -sinr*cosp - cosr*sint*sinp;
			b2 =  cosr*cost;
			c2 =  sinr*sinp - cosr*sint*cosp;
			a3 =  cost*sinp;
			b3 =  sint;
			c3 =  cost*cosp;

		}
		else
		{
			// below are the coefficients of the rows of the rotation matrix M for the ball,
			// given by M = ZRPT, with
			// MATRIX3 Z = {0,1,0,  -1,0,0,  0,0,1};
			// MATRIX3 P = {1,0,0,  0,cosp,-sinp,  0,sinp,cosp};  // yaw
			// MATRIX3 T = {cost,0,sint,  0,1,0,  -sint,0,cost};  // pitch
			// MATRIX3 R = {cosr,sinr,0,  -sinr,cosr,0,  0,0,1};  // bank

			a1 = -sinr*cost + cosr*sinp*sint;
			b1 =  cosr*cosp;
			c1 = -sinr*sint - cosr*sinp*cost;
			a2 = -cosr*cost - sinr*sinp*sint;
			b2 = -sinr*cosp;
			c2 = -cosr*sint + sinr*sinp*cost;
			a3 = -cosp*sint;
			b3 =  sinp;
			c3 =  cosp*cost;
		}

		//MATRIX3 rMatrix = { a1,b1,c1, a2,b2,c2, a3,b3,c3};

		// Compensate for difference between panel orientation and spacecraft's local frame if any.
		MATRIX3 M = { a1,b1,c1, a2,b2,c2, a3,b3,c3};
		MATRIX3 rMatrix = mul (bias, M);
	
		// ==============================================================
 		// Due to the way the orbiter core handles MGROUP_ROTATE Parent/child operations it is not currently possible to rotate a mesh 
		// on more than two axises at once. This presents problems for an ADI ball that needs to be able to rotate along all three.
		// Before, the algorithm always applied the last state of the vertices to the same Euler angle-generated 
		// rotation matrix on every time-step. Now, i talways applies the same rotation to the original vertices
		// taken from the template mesh loaded via oapiLoadMeshGlobal(). In addition, now the proper mesh instance 
		// is used instead of the mesh template. Current mesh instance is retrieved via visual callbacks.
		//
		// Special thanks to ASMI and Face on the orbiter forums for helping me with this part. *Tips Hat*
		// ==============================================================

		DWORD l = mgh_ADI_orig->nVtx;	// Get vertex count of ADI ball meshgroup

		// Create a structure for modifying the ADI ball mesh 
		GROUPEDITSPEC gespec_adi;							// Structure declaration
		gespec_adi.flags = GRPEDIT_VTXNML | GRPEDIT_VTXCRD; // Set Flags	// | is how you combine flags in C/C++ :) -ASMI	// Thanks ;) -Hlynka		
		gespec_adi.nVtx = mgh_ADI_orig->nVtx;				// Declare index of vertices

		// allocate memory for the vertices and their indices
		gespec_adi.Vtx = new NTVERTEX[gespec_adi.nVtx];
		gespec_adi.vIdx = new WORD[gespec_adi.nVtx];

		for( WORD j = 0; j < l; j++)	// For each vertex in the ADI ball's meshgroup perform the following operations...
		{
			VECTOR3 p = _V( mgh_ADI_orig->Vtx[j].x, mgh_ADI_orig->Vtx[j].y, mgh_ADI_orig->Vtx[j].z);	// Store the original position vertex in p.
			VECTOR3 n = _V( mgh_ADI_orig->Vtx[j].nx, mgh_ADI_orig->Vtx[j].ny, mgh_ADI_orig->Vtx[j].nz);	// Store the original normal vertex in n.
			p = mul( rMatrix, p);	// multiply p by rMatrix to get new position
			n = mul( rMatrix, n);	// Do the same for the normal vector to prevent lighting artifacts.

			// Write new value of p into the live structure.
			gespec_adi.Vtx[j].x=p.x;
			gespec_adi.Vtx[j].y=p.y;
			gespec_adi.Vtx[j].z=p.z;

			// Write new value of n to the ADI ball live structure.
			gespec_adi.Vtx[j].nx=n.x;
			gespec_adi.Vtx[j].ny=n.y;
			gespec_adi.Vtx[j].nz=n.z;

			//remember about index
			gespec_adi.vIdx[j] = j; 

		} // End vertext operations
 
		// perform the rotation (apply edit spec to mesh)
		oapiEditMeshGroup( dmh_FDAI, 0, &gespec_adi); // C++ uses ampersand operator (&) to get address
	
		// ...and don't forget to clean up (only you can prevent memory leaks :P)
		delete [] gespec_adi.Vtx;
		delete [] gespec_adi.vIdx;
	} // End "if (mgh_ADI_current != NULL)"

	//// Debuggery
	//sprintf(oapiDebugString(),"rho %0.2f theta %0.2f phi %0.2f", rho*DEG, theta*DEG, phi*DEG);
	//sprintf(oapiDebugString()," Pitch %0.2f Roll %0.2f Yaw %0.2f", theta*DEG, rho*DEG, phi*DEG);

} // End FDAI::AnimateADI(VECTOR3 euler)

// Animate Error Needles
void FDAI::AnimateErrorNeedles (VECTOR3 euler)
{
	double pitch_error = euler.x;
	double roll_error = euler.z;
	double yaw_error = euler.y;
	
	// Calculate display values
	double pitch_lvl = 0.5 + pitch_error / ErrorScale;
	double roll_lvl	= 0.5 + roll_error / ErrorScale;
	double yaw_lvl = 0.5 + yaw_error / ErrorScale;

	// Sanity check values
	if (pitch_lvl > 1)		pitch_lvl = 1;
	else if (pitch_lvl < 0)	pitch_lvl = 0;

	if (roll_lvl > 1)		roll_lvl = 1;
	else if (roll_lvl < 0)	roll_lvl = 0;

	if (yaw_lvl > 1)		yaw_lvl = 1;
	else if (yaw_lvl < 0)	yaw_lvl = 0;

	// Animate Needles
	v->SetAnimation (anim_PitchError, pitch_lvl);
	v->SetAnimation (anim_RollError, roll_lvl);
	v->SetAnimation (anim_YawError, yaw_lvl); 
}

// Animate Rate Needles
void FDAI::AnimateRateNeedles (VECTOR3 rotation)
{
	double pitch_rate = rotation.x;
	double roll_rate = rotation.z;
	double yaw_rate = rotation.y;
	
	// Calculate display values
	double pitch_lvl = 0.5 + pitch_rate / RateScale;
	double roll_lvl	= 0.5 + roll_rate / RateScale;
	double yaw_lvl = 0.5 + yaw_rate / RateScale;

	// Sanity check values
	if (pitch_lvl > 1)		pitch_lvl = 1;
	else if (pitch_lvl < 0)	pitch_lvl = 0;

	if (roll_lvl > 1)		roll_lvl = 1;
	else if (roll_lvl < 0)	roll_lvl = 0;

	if (yaw_lvl > 1)		yaw_lvl = 1;
	else if (yaw_lvl < 0)	yaw_lvl = 0;

	// Animate Needles
	v->SetAnimation (anim_PitchRate, pitch_lvl);
	v->SetAnimation (anim_RollRate, roll_lvl);
	v->SetAnimation (anim_YawRate, yaw_lvl); 
}