// ==============================================================
//		ORBITER MODULE: 'SPIDER' LUNAR EXCURSION MODULE
//			A custom Vessel For Orbiter 2010/2011
//  
// AAPO_FDAI.h
// Interface for ADI ball sub-system class
// ==============================================================

#pragma once
#include <math.h>
#include <vector>
#include "AAPO_Common.h"

// ========================================================
// Mesh resource file for FDAI.msh
// Generated with meshc on Mon Jun 03 20:32:55 2013
// ========================================================

// Number of mesh groups:
#define FDAI_NGRP					12

// Number of materials:
#define FDAI_NMAT					5

// Number of textures:
#define FDAI_NTEX					1

// Named mesh groups:
#define FDAI_GRP_ADI_Ball			0
#define FDAI_GRP_Case				1
#define FDAI_GRP_ErrorNeedle_Pitch	2
#define FDAI_GRP_ErrorNeedle_Roll	3
#define FDAI_GRP_ErrorNeedle_Yaw	4
#define FDAI_GRP_Face				5
#define FDAI_GRP_Horizon			6
#define FDAI_GRP_RateNeedle_Pitch	7
#define FDAI_GRP_RateNeedle_Roll	8
#define FDAI_GRP_RateNeedle_Yaw		9
#define FDAI_GRP_Roll_Index			10
#define FDAI_GRP_Scale				11

// ==============================================================
// Constants
// ==============================================================

const double	FDAI_BALLSPEED		= 3.0;					// Max rotation rate of ADI ball
const double	FDAI_NEEDLERANGE	= 0.064;				// Rate and Error needle range of travel.

// ==============================================================
// Class declaration
// ==============================================================
class FDAI 
{
public:
	FDAI (const VESSEL *vessel, const VECTOR3 pos, double biasangle, VECTOR3 biasaxis);
	inline const VESSEL *GetVessel() const { return v; }

	// Public Functions (User Inputs)
	bool	SetBias (double angle, VECTOR3 rotationaxis);
	bool	SetErrorScale (double scale);
	bool	SetNavChan (int newnavchan);
	bool	SetProjMode (int newmode);
	bool	SetRateScale (double scale);
	bool	SetRefMode (int newmode);
	bool	SetTgtMode (int newmode);
	bool	SetBallOffset (VECTOR3 newoffset);
	bool	SetTargetOffset (VECTOR3 newoffset);

	// Callback functions 
	// NOTE: "PostStep()" needs to be called in the "clbkPostStep()" function of the associated vessel, "VisualCreated()" in "clbkVisualCreated()" etc...
	void	VisualCreated (VISHANDLE vis, int refcount);
	void	VisualDestroyed (VISHANDLE vis, int refcount);
	void	PostStep (double simt, double simdt, double mjd);
	
	bool	PowerOn;
	double	RateScale, ErrorScale;

	// Enumerators
	int		Projection;		// Projection mode (0 = yaw range of 360, 1 = pitch range of 360)
	int		NavChan;		// Nav Channel to be used in NAV_SOURCE reference mode
	enum	Reference {ECLIPTIC, EQUATORIAL, ORBIT, SURFACE, NAV_SOURCE} RefMode;	// Reference orientation used to calculate ball rotations. 
	enum	Mode {NO_TARGET, FIXED, TGT_DIRECTION, TGT_RVEL} TgtMode;				// Direction to be indicated by error needles

	// Handles
	OBJHANDLE	oh_rBody;		
	NAVHANDLE	nh_NavTarget;

private:
	const VESSEL	*v;			// Pointer to Vessel
	int				mesh_idx;	// Mesh index	

	// Variables
	VECTOR3	balloffset, tgtoffset, ball_pos;
	MATRIX3	bias;

	VECTOR3	tgt_rpos, tgt_rvel;
	double	curr_rho, curr_theta, curr_phi;

	// Private Functions (Internal Operations)
	MATRIX3	GetFrameRotMatrix();
	VECTOR3	GetEulerAngles();
	VECTOR3	GetTgtEulerAngles(NAVHANDLE hNav);
	VECTOR3	GetRotationRates();

	void	AnimateADI(VECTOR3 euler);
	void	AnimateErrorNeedles(VECTOR3 euler);
	void	AnimateRateNeedles(VECTOR3 rotation);

	inline double posangle (double angle)
	{
		double a = fmod(angle, PI2);
		return (a >= 0.0 ? a : a+PI2);
	}

	// Mesh group pointers
	DEVMESHHANDLE	dmh_FDAI;
	UINT			mesh_FDAI;
	
	MESHGROUPEX *mgh_ADI_orig;		// original (template) ADI ball
	MESHGROUPEX *mgh_ADI_current;	// current visual's ADI ball

	UINT		anim_bias, anim_PitchRate, anim_RollRate, anim_YawRate, anim_PitchError, anim_RollError, anim_YawError;
};