// ==============================================================
//                 ORBITER MODULE: CM_Chute
//                  Part of the ORBITER SDK
//          Copyright (C) 2002-2004 Martin Schweiger
//                   All rights reserved
//
// CM_Chute.cpp
// Control module for CM_Chute vessel class
//
// Notes:
// ==============================================================

#include "CM_Chute.h"

// ==============================================================
// Class Constructor and Destructor
// ==============================================================

CM_Chute::CM_Chute (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
	ChuteStatus = CLOSED;
	deploy_proc = 0.0;
	
	birthdate = oapiGetSimTime ();	// Birthdate = current sim time
	lifespan = 3600;				// Set default lifespan, 3600 seconds (1 hour).
}

CM_Chute::~CM_Chute ()
{
}

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void CM_Chute::clbkSetClassCaps (FILEHANDLE cfg)
{
	// physical vessel parameters
	SetSize (CHUTE_SIZE);
	SetEmptyMass (CHUTE_MASS * 3);
	SetPMI (CHUTE_PMI);
	SetCrossSections (CHUTE_CS);
	SetTouchdownPoints (_V( 0,-3,-CHUTE_LENGTH), _V(-3, 0,-CHUTE_LENGTH), _V( 3, 0,-CHUTE_LENGTH));

	// Aerodynamics
	SetRotDrag (_V( 0.01, 0.01, 0.005));
	CreateAirfoil (LIFT_VERTICAL, _V(0, 0, 0.1), VLiftCoeff, CHUTE_AREA, 0, 0.47);
	CreateAirfoil (LIFT_VERTICAL, _V(0, 0, 0.1), VLiftCoeff, CHUTE_AREA, 0, 0.47);
	CreateAirfoil (LIFT_HORIZONTAL, _V(0, 0, 0.1), HLiftCoeff, CHUTE_AREA, 0, 0.47);

	// Attachment point
	CreateAttachment (true, _V( 0, 0,-CHUTE_LENGTH), _V( 0, 0,-1), _V( 0, 1, 0), "X");
	
	// associate a mesh for the visual
	AddMesh ("AAPO/CM_Chutes", &_V( 0, 0,-CHUTE_LENGTH));
}

// --------------------------------------------------------------
// Read status from scenario file
// --------------------------------------------------------------
void CM_Chute::clbkLoadStateEx (FILEHANDLE scn, void *status)
{
	char *cbuf; 

	while (oapiReadScenario_nextline (scn, cbuf)) 
	{
		// Load animation states
		if (!_strnicmp (cbuf, "STATE", 5))
		{
			sscanf (cbuf + 5, "%i %lf", &ChuteStatus, &deploy_proc);	
		}

		// Load default parameters
		else ParseScenarioLineEx (cbuf, status);	

	} // End "while (oapiReadScenario_nextline (scn, cbuf))"

} // End "clbkLoadStateEx(FILEHANDLE scn, void *status)"

// --------------------------------------------------------------
// Write status to scenario file
// --------------------------------------------------------------
void CM_Chute::clbkSaveState (FILEHANDLE scn)
{
	int i = 0;
	char cbuf[256];
	
	// Write default vessel parameters to file
	VESSEL3::clbkSaveState (scn);

	// Write custom parameters to file
	sprintf (cbuf, "%i %0.4f", ChuteStatus, deploy_proc);
	oapiWriteScenario_string (scn, "STATE", cbuf);
} // End "clbkSaveState(FILEHANDLE scn)"

// --------------------------------------------------------------
// Finalise vessel creation
// --------------------------------------------------------------
void CM_Chute::clbkPostCreation (void)
{
	static UINT groups[CHUTE_NGROUPS] = {0, 1, 2};
	static MGROUP_SCALE deploy ( 0, groups, CHUTE_NGROUPS, _V( 0, 0, 0), _V( 200.0, 200.0, 1000.0));

	anim_deploy = CreateAnimation (1.0);
	AddAnimationComponent (anim_deploy, 0.0, 1.0, &deploy);
	SetAnimation (anim_deploy, deploy_proc);
}

// --------------------------------------------------------------
// Post step processes
// --------------------------------------------------------------
void CM_Chute::clbkPostStep (double simt, double simdt, double mjd)
{
	// Deployment Animation
	double delta = simdt / CHUTE_DTIME;

	// Deployment Animation
	if (ChuteStatus != OPEN)	
	{
		ChuteStatus = OPENING;
		SetAnimation (anim_deploy, deploy_proc);

		if (deploy_proc < 1) deploy_proc += delta;	
		else ChuteStatus = OPEN;						
	}

	//sprintf (oapiDebugString (), "Proc %0.2f", deploy_proc);

	//// Deflation Animation
	//if (GroundContact ())
	//{
	//	delta = simdt / 10;
	//	if (deflate_proc > 0) deflate_proc -= delta;
	//	else deflate_proc = 0;

	//	SetAnimation (anim_deflate, deflate_proc);
	//}

	// Clean up
	OBJHANDLE self = GetHandle ();					// Get vessel handle 
	age = oapiGetSimTime () - birthdate;			// Get age, current sim time - birthdate
	if (age > lifespan) oapiDeleteVessel(self);		// If age is greater than lifespan delete vessel 

	if (GroundContact ()) oapiDeleteVessel(self);	// If vessel touches the ground delete it 
	//double drag = GetDrag () / GetMass ();			// Get decceleration due to drag
	//if (drag > 8 * G) oapiDeleteVessel (self);		// if decceleration is greater than 8 Gs burn up
}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new CM_Chute (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (CM_Chute*) vessel;
}
