// ==============================================================
//                 ORBITER MODULE: CM_Drogue
//                  Part of the ORBITER SDK
//          Copyright (C) 2002-2004 Martin Schweiger
//                   All rights reserved
//
// CM_Drogue.cpp
// Control module for CM_Drogue vessel class
//
// Notes:
// ==============================================================

#include "CM_Drogue.h"

// ==============================================================
// Class Constructor and Destructor
// ==============================================================

CM_Drogue::CM_Drogue (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
	birthdate = oapiGetSimTime ();	// Birthdate = current sim time
	lifespan = 3600;				// Set default lifespan, 3600 seconds (1 hour).

	ChuteStatus = CLOSED;
	deploy_proc = 0;
}

CM_Drogue::~CM_Drogue ()
{
}

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void CM_Drogue::clbkSetClassCaps (FILEHANDLE cfg)
{
	// physical vessel parameters
	SetSize (DROGUE_SIZE);
	SetEmptyMass (DROGUE_MASS);
	SetPMI (DROGUE_PMI);
	SetCrossSections (DROGUE_CS);
	
	// Aerodynamics
	SetRotDrag (_V( 0.01, 0.01, 0.005));
	CreateAirfoil (LIFT_VERTICAL, _V( 0,-0.02, 0.1), VLiftCoeff, DROGUE_AREA, 0, 0.47);
	CreateAirfoil (LIFT_HORIZONTAL, _V( 0, 0, 0.1), HLiftCoeff, DROGUE_AREA, 0, 0.47);

	// Attachment
	CreateAttachment (true, _V( 0, 0,-DROGUE_LENGTH), _V( 0, 0,-1), _V( 0, 1, 0), "X");
	
	// associate a mesh for the visual
	AddMesh ("AAPO/CM_Drogues", &_V( 0, 0,-DROGUE_LENGTH));
}

// --------------------------------------------------------------
// Read status from scenario file
// --------------------------------------------------------------
void CM_Drogue::clbkLoadStateEx (FILEHANDLE scn, void *status)
{
	char *cbuf; 

	while (oapiReadScenario_nextline (scn, cbuf)) 
	{
		// Load animation states
		if (!_strnicmp (cbuf, "STATE", 5))
		{
			sscanf (cbuf + 5, "%i %lf", &ChuteStatus, &deploy_proc);	
		}

		// Load default parameters
		else ParseScenarioLineEx (cbuf, status);	

	} // End "while (oapiReadScenario_nextline (scn, cbuf))"

} // End "clbkLoadStateEx(FILEHANDLE scn, void *status)"

// --------------------------------------------------------------
// Write status to scenario file
// --------------------------------------------------------------
void CM_Drogue::clbkSaveState (FILEHANDLE scn)
{
	int i = 0;
	char cbuf[256];
	
	// Write default vessel parameters to file
	VESSEL3::clbkSaveState (scn);

	// Write custom parameters to file
	sprintf (cbuf, "%i %0.4f", ChuteStatus, deploy_proc);		
	oapiWriteScenario_string (scn, "STATE", cbuf);
} // End "clbkSaveState(FILEHANDLE scn)"

// --------------------------------------------------------------
// Finalise vessel creation
// --------------------------------------------------------------
void CM_Drogue::clbkPostCreation (void)
{
	static UINT groups[DROGUE_NGROUPS] = {0, 1, 2};
	static MGROUP_SCALE deploy ( 0, groups, DROGUE_NGROUPS, _V( 0, 0, 0), _V( 200.0, 200.0, 1000.0));

	anim_deploy = CreateAnimation (1.0);
	AddAnimationComponent (anim_deploy, 0.0, 1.0, &deploy);
	SetAnimation (anim_deploy, deploy_proc);
}

// --------------------------------------------------------------
// Post step processes
// --------------------------------------------------------------
void CM_Drogue::clbkPostStep (double simt, double simdt, double mjd)
{
	double delta = simdt / DROGUE_DTIME;

	// Deployment Animation
	if (ChuteStatus != OPEN)	
	{
		ChuteStatus = OPENING;
		SetAnimation (anim_deploy, deploy_proc);

		if (deploy_proc < 1) deploy_proc += delta;	
		else ChuteStatus = OPEN;						
	}
	// sprintf (oapiDebugString (), "Proc %0.2f", deploy_proc);

	OBJHANDLE parent = GetAttachmentStatus (GetAttachmentHandle (true, 0));

	if (parent == NULL)
	{
		// Clean up
		OBJHANDLE self = GetHandle ();					// Get vessel handle 
		age = oapiGetSimTime () - birthdate;			// Get age, current sim time - birthdate
		if (age > lifespan) oapiDeleteVessel(self);		// If age is greater than lifespan delete the vessel 
		if (GroundContact ()) oapiDeleteVessel(self);	// If vessel touches the ground delete it 
	}
}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new CM_Drogue (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (CM_Drogue*) vessel;
}
