// ==============================================================
//                 ORBITER MODULE: CSM
//                  Part of the ORBITER SDK
//          Copyright (C) 2002-2004 Martin Schweiger
//                   All rights reserved
//
// CSM.cpp
// Control module for CSM vessel class
//
// Notes:
// This is an example for a "minimal" vessel implementation which
// only overloads the clbkSetClassCaps method to define vessel
// capabilities and otherwise uses the default VESSEL class
// behaviour.
// ==============================================================

// Header files
#include "CSM.h"
#include "CSM_Cockpit.h"
#include "CSM_SubSystems.h"
#include "CSM_SwitchDefs.h"

#include "CM_Apex.h"
#include "CM_Drogue.h"
#include "CM_Chute.h"
#include "LES.h"
#include "SLA.h"
#include "SM.h"

// Mesh resource files
#include "mesh_CM_VC.h"
#include "mesh_CM.h"
#include "mesh_SM.h"

// ==============================================================
// Class Constructor and Destructor
// ==============================================================
CSM::CSM (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
	// Config file difficulty modifiers (1.0 = 100% of default value)
	thrust_mod		= 1.0;
	dv_mod			= 1.0;
	consumables_mod = 1.0;

	// Initial vessel state
	LES_attached	= false;
	SLA_attached	= false; 
	SM_attached		= true;
	DockingProbe	= true;
	ApexCovered		= true;
	SM_BayOpen		= false;

	ReEntryConfig	= false;
	AbortInProgress = false;

	// CM animation states
	Drogues_Deployed = false;
	drogue_proc		= 0.0;

	Chutes_Deployed = false;
	chute_proc		= 0.0;

	HatchStatus		= CLOSED;
	hatch_proc		= 0.0;

	// SM animation states
	UmbilicalStatus	= CLOSED;
	umbilical_proc	= 0.0;

	LightStatus		= CLOSED;
	light_proc		= 0.0;

	AntennaStatus	= CLOSED;
	antenna_proc	= 0.0;

	// Load meshes
	mh_adapter		= oapiLoadMeshGlobal ("AAPO/LaunchAdapter");	
	mh_apex			= oapiLoadMeshGlobal ("AAPO/CM_Apex");
	mh_cockpit		= oapiLoadMeshGlobal ("AAPO/CM_VC");
	mh_command		= oapiLoadMeshGlobal ("AAPO/CM");
	mh_probe		= oapiLoadMeshGlobal ("AAPO/DockingProbe");
	mh_sbaycover	= oapiLoadMeshGlobal ("AAPO/SM_BayCover");		
	mh_service		= oapiLoadMeshGlobal ("AAPO/SM");
	mh_tower		= oapiLoadMeshGlobal ("AAPO/AbortTower");

	// Load sub classes
	vc				= new CSM_VC (this);	// Cockpit control class
	csm				= new CSM_SYS (this, vc);	// Cockpit control class
}

CSM::~CSM ()
{
}

// ==============================================================
// Custom Vessel Functions
// ==============================================================

// --------------------------------------------------------------
// Define mesh animations
// --------------------------------------------------------------
void CSM::DefineAnimations (void)
{
	float alpha, beta, gamma, x1, y1, z1, x, y, z;

	// EVA Hatch (Exterior)
	alpha = RAD * 33.0;
	beta = RAD * 21.0;
	gamma = RAD * 120;
	x1 = 0.0;
	y1 =-sin(alpha); 
	z1 = cos(alpha);
	x = x1*cos(beta) - y1*sin(beta);
	y = x1*sin(beta) + y1*cos(beta); 
	z = z1;

	VECTOR3 ref = _V(-0.497, 1.589, 0.516);

	static UINT	meshgroups_HatchExt[5] = {GRP_CM_Hatch_Handle, GRP_CM_Hatch_Inside, GRP_CM_Hatch_Mylar, GRP_CM_Hatch_Outside, GRP_CM_Windows_Hatch};
	static MGROUP_ROTATE	mgt_HatchExt (mesh_CM, meshgroups_HatchExt, 5, ref, _V( x, y, z), gamma); 

	static UINT	meshgroups_HatchVC[5] = {GRP_VC_Hatch_Handle, GRP_VC_Hatch_Inside, GRP_VC_Hatch_Insulation, GRP_VC_Hatch_Mylar, GRP_VC_Windows_Hatch};
	static MGROUP_ROTATE	mgt_HatchVC (mesh_Cockpit, meshgroups_HatchVC, 5, ref, _V( x, y, z), gamma); 

	anim_Hatch = CreateAnimation (0.0);
	AddAnimationComponent (anim_Hatch, 0.3f, 1.0f, &mgt_HatchExt);
	AddAnimationComponent (anim_Hatch, 0.3f, 1.0f, &mgt_HatchVC);

	// Drogue deploy
	static UINT meshgroups_drogues = GRP_CM_ChutesDrogue;
	static MGROUP_SCALE droguedeploy (mesh_CM, &meshgroups_drogues, 1, _V( 0, 0, 1), _V( 0.01, 0.01, 0.01));

	anim_droguedeploy = CreateAnimation (0.0);
	AddAnimationComponent (anim_droguedeploy, 0.0, 1.0, &droguedeploy);

	// Chute deploy
	static UINT meshgroups_chutes = GRP_CM_ChutesMain;
	static MGROUP_SCALE chutedeploy (mesh_CM, &meshgroups_chutes, 1, _V( 0, 0, 1), _V( 0.01, 0.01, 0.01));

	anim_chutedeploy = CreateAnimation (0.0);
	AddAnimationComponent (anim_chutedeploy, 0.0, 1.0, &chutedeploy);

	// Umbilicle
	static UINT				meshgroup_umbilical = GRP_SM_Umbilical;
	static MGROUP_ROTATE	mgt_umbilical (mesh_SM, &meshgroup_umbilical, 1, SM_UMBILICAL_PIVOT, _V( 1.0, 0.0, 0.0), (float) 35*RAD); 

	anim_umbilical = CreateAnimation (1.0);
	AddAnimationComponent (anim_umbilical, 0.0f, 1.0f, &mgt_umbilical);	
	SetAnimation (anim_umbilical, umbilical_proc);

	// Eva Light
	static UINT				meshgroup_light = GRP_SM_EVA_Light;
	static MGROUP_ROTATE	mgt_light (mesh_SM, &meshgroup_light, 1, SM_LIGHT_PIVOT, _V( sin(SM_LIGHT_CANT), cos(SM_LIGHT_CANT), 0.0), (float)-101*RAD); 

	anim_light = CreateAnimation (1.0);
	AddAnimationComponent (anim_light, 0.0f, 1.0f, &mgt_light);	
	SetAnimation (anim_light, light_proc);

	// High Gain Antenna 
	static UINT				meshgroup_HGAstrut = GRP_SM_HGA_Strut;
	static MGROUP_ROTATE	mgt_antenna (mesh_SM, &meshgroup_HGAstrut, 1, SM_HGA_MOUNT, _V( sin(SM_HGA_CANT), cos(SM_HGA_CANT), 0.0), (float) 90*RAD); 

	anim_antenna = CreateAnimation (1.0);
	ach_antenna = AddAnimationComponent (anim_antenna, 0.0f, 1.0f, &mgt_antenna);

	//
	SetAnimation (anim_Hatch, hatch_proc);
	SetAnimation (anim_antenna, antenna_proc);
	SetAnimation (anim_light, light_proc);

} // End "CSM::DefineAnimations"

// --------------------------------------------------------------
// Define local light sources
// --------------------------------------------------------------
void CSM::DefineLighting (void)
{

} // End "CSM::DefineLighting"

// --------------------------------------------------------------
// Define exhaust and particle streams
// --------------------------------------------------------------
void CSM::DefineParticleStreams (void)
{
	// Main propulsion exhaust
	SURFHANDLE tex_main = oapiRegisterExhaustTexture ("Exhaust2");	// Load Main Engine exhaust texture
	SURFHANDLE tex_rcs = oapiRegisterExhaustTexture ("Exhaust_atrcs");	// Load RCS exhaust texture

	// Add Main engine exhaust
	AddExhaust (th_sps, 4.0*thrust_mod, 1, 3.0, tex_main);

	// Add RCS exhaust
	for (int i = 0; i < 12; i++)
	{
		AddExhaust (th_cm_rcs[i], 2.0, 0.1, 0.06, tex_rcs); // Command module RCS thrusters [length, width, offset]
	}
	
	for (int i = 0; i < 16; i++)
	{
		AddExhaust (th_sm_rcs[i], 3.0, 0.2, 0.3, tex_rcs); // Service module RCS thrusters [length, width, offset]
	}

} // End "CSM::DefineParticleStreams"

// --------------------------------------------------------------
// Define sound effects
// --------------------------------------------------------------
void CSM::DefineSounds (void)
{
	// OrbiterSound 4.0
	OrbiterSoundID = ConnectToOrbiterSoundDLL (GetHandle());			// Init sound
	SoundOptionOnOff (OrbiterSoundID, PLAYCOUNTDOWNWHENTAKEOFF, false);	// Disable the stock countdown... it's annoying :)

	LoadSoundFiles (OrbiterSoundID);
} // End "CSM::DefineSounds"

// --------------------------------------------------------------
// Calculate current empty mass of vessel
// --------------------------------------------------------------
double CSM::CSM_EmptyMass (void)
{
	double output = CM_EMPTYMASS;					// Default empty mass is equal to command module's empty mass

	if (SM_attached)	output += SM_EMPTYMASS;		// If service module is attached add it's mass to total
	if (LES_attached)	output += LES_MASS;			// If abort tower is attached add it's mass to total
	if (SLA_attached)	output += SLA_MASS;			// If SLA is attached add it's mass to total
	if (ApexCovered)	output += APEX_MASS;		// If parachute cover is attached add it's mass to total

	// Add mass of crew	to total
	int nsouls = Crew.GetCrewTotalNumber ();
	for (int i = 0; i < nsouls; i ++) output += Crew.GetCrewWeightBySlotNumber (i);

	// Add mass of payload to total
	output += Cargo.GetCargoTotalMass ();

	int attcount = AttachmentCount (false);
	for (int i = 0; i < attcount; i ++)
	{
		OBJHANDLE oh_pyld = GetAttachmentStatus (GetAttachmentHandle (false, i));
		if (oh_pyld != NULL)
		{
			VESSEL *pyld = oapiGetVesselInterface (oh_pyld);
			output += pyld->GetMass ();
		}
	}

	//  Subtract mass of SM bay cover if open
	if ((SM_attached) && (SM_BayOpen)) output -= 100;

	// Subtract mass of parachutes if deployed
	if (Drogues_Deployed)	output -= DROGUE_MASS;
	if (Chutes_Deployed)	output -= CHUTE_MASS;

	// Return total
	return output;
} // End "CSM::CSM_EmptyMass"

// --------------------------------------------------------------
// EVA Hatch control logic
// --------------------------------------------------------------
bool CSM::ActivateHatch (void)
{
	// Sanity check.
	if (hatch_proc < 0)			// If process value is less than 0...
	{
		HatchStatus = CLOSED;	// ...set status to "CLOSED"
		hatch_proc = 0;			// and process value to 0
	}
	else if (hatch_proc > 1)	// If process value is greater than 1...
	{
		HatchStatus = OPEN;		// ...set status to "OPEN"
		hatch_proc = 1;			// and process value to 1
	}

	// Process input
	if (HatchStatus != CLOSED) 
	{
		HatchStatus = CLOSING;
		return true;
	}	
	else
	{
		// check pressure difference
		HatchStatus = OPENING;
		return true;
	}

	SetAnimation (anim_Hatch, hatch_proc);			// Update animation state
	return false;
} // End "CSM::ActivateHatch"

// --------------------------------------------------------------
// High gain atntenna control logic
// --------------------------------------------------------------
bool CSM::DeployHGA (void)
{
	// Sanity check.
	if (antenna_proc < 0)			// If process value is less than 0...
	{
		AntennaStatus = CLOSED;		// ...set status to "CLOSED"
		antenna_proc = 0;			// and process value to 0
	}
	else if (antenna_proc > 1)		// If process value is greater than 1...
	{
		AntennaStatus = OPEN;		// ...set status to "OPEN"
		antenna_proc = 1;			// and process value to 1
	}
	
	// Process input
	if ((AntennaStatus != OPEN) && (Booster_attached == false)) // check to see if in launch configuration
	{
		// Deploy antenna		
		AntennaStatus = OPENING;
		return true;
	}	
	
	SetAnimation (anim_Hatch, hatch_proc);			// Update animation state
	return false;
} // End "CSM::ActivateGear"

// --------------------------------------------------------------
// Abort tower seperation routine
// --------------------------------------------------------------
bool CSM::TowerSep (void)
{
	// Check CG offset 
	VECTOR3 ofs;
	if (ReEntryConfig) ofs = _V( 0, 0, 0);
	else ofs = CSM_CM_OFFSET; 

	OBJHANDLE oh_LES = NULL;

	if ((AbortInProgress) && (GetDockStatus (dh_cm_dock) != NULL))
	{
		oh_LES = GetDockStatus (dh_cm_dock);
		VESSEL *les = oapiGetVesselInterface (oh_LES);			// Get interface for LES vessel
		les->SetThrusterGroupLevel (THGROUP_HOVER, 1.0);		// Fire LES's jettison motors

		csm->ELS_state = 2;
		ApexSep ();
		PlayVesselWave (OrbiterSoundID, SOUND_TOWERJET);		// Play sound effect
		return true;
	}
	else if (LES_attached) // Sanity check, is there an Launch escape system to seperate from?
	{
		// Create service module as seperate vessel
		char name[256];
		strcpy (name, GetName()); 
		strcat (name, "_LES");

		oh_LES = SpawnObject (this, name, "AAPO/Parts/LES", CM_ABORT_OFFSET + ofs, 0.0); 
		VESSEL *les = oapiGetVesselInterface (oh_LES);			// Get interface for LES vessel
		les->SetThrusterGroupLevel (THGROUP_HOVER, 1.0);		// Fire LES's jettison motors
		
		LES_attached = false;									// Update state
		DelMesh (mesh_LES);										// Delete mesh
		PlayVesselWave (OrbiterSoundID, SOUND_TOWERJET);		// Play sound effect
		return true;
	}
	else return false;
} // End "CSM::TowerSep"

// --------------------------------------------------------------
// Spacecraft launch adapter seperation routine
// --------------------------------------------------------------
bool CSM::AdapterSep (void)
{
	OBJHANDLE oh_SLA, oh_LM, oh_LV; 

	if (SLA_attached) // Sanity check, is there an SLA to seperate from?
	{
		char name[256];
		strcpy (name, GetName()); 
		strcat (name, "_SLA");
		
		oh_SLA = SpawnObject (this, name, "AAPO/Parts/SLA", CSM_SLA_OFFSET, 0.1); 
		oh_LM = GetAttachmentStatus (ah_payload);	// get handle for Lunar module
		oh_LV = GetAttachmentStatus (ah_booster);	// get handle for Launch vehicle

		VESSEL *sla = oapiGetVesselInterface (oh_SLA);

		// Shift payload from vessel to SLA
		if (oh_LM != NULL)
		{
			int idx = 0;
			VESSEL *lm = oapiGetVesselInterface (oh_LM);	// get VESSEL interface for payload
			int attcount = lm->AttachmentCount (true);		// get number of attachment points
			
			for (int i = 0; i < attcount; i++)				// Scroll through payload's attachment points...
			{
				if (lm->GetAttachmentStatus (lm->GetAttachmentHandle (true, i)) == GetHandle ()) idx = i; // ...and find the point that we are currently attached to.
			}
			
			sla->AttachChild (oh_LM, sla->GetAttachmentHandle (false, 0), lm->GetAttachmentHandle (true, idx));
		} // End "if (oh_LM != NULL)"

		// Attach SLA to booster
		if (oh_LV != NULL)
		{
			int idx = 0;
			VESSEL *lv = oapiGetVesselInterface (oh_LV);	// get VESSEL interface for launch vehicle
			int attcount = lv->AttachmentCount (false);		// get number of attachment points

			for (int i = 0; i < attcount; i++)				// Scroll through LV's attachment points...
			{
				if (lv->GetAttachmentStatus (lv->GetAttachmentHandle (false, i)) == GetHandle ()) idx = i; // ...and find the point that we are currently attached to.
			}

			lv->DetachChild (lv->GetAttachmentHandle (false, idx), 0.0);	// dettach CSM from LV
			lv->AttachChild (oh_SLA, lv->GetAttachmentHandle (false, idx), sla->GetAttachmentHandle (true, 0)); // Attach CSM
		} // End "if (oh_LV != NULL)"

		// Play sound effect
		PlayVesselWave (OrbiterSoundID, SOUND_SEPERATION); 

		// Update vessel state
		DelMesh (mesh_SLA);
		SLA_attached = false;
		return true;
	} // End "if (SLA_attached)"
	else return false;

} // End "CSM::AdapterSep"

// --------------------------------------------------------------
// Service module seperation routine
// --------------------------------------------------------------
bool CSM::ServiceModuleSep (void)
{
	if (SM_attached) // Sanity check, is there an service module to seperate from?
	{
		// Create service module as seperate vessel
		char name[256];
		strcpy (name, GetName()); 
		strcat (name, "_ServiceModule");
		OBJHANDLE oh_servicemod = SpawnObject (this, name, "AAPO/Parts/SM", CSM_SM_OFFSET, 0.1); 

		// Match service module vessel's propellant levels to current 
		VESSEL *sm = oapiGetVesselInterface (oh_servicemod);
		for (int i = 0; i < SM_TANKCOUNT; i ++) 
		{
			PROPELLANT_HANDLE ph_temp = GetPropellantHandleByIndex (CM_TANKCOUNT + i);
			sm->SetPropellantMass (sm->GetPropellantHandleByIndex (i), GetPropellantMass (ph_temp));
		}

		// Fire thrusters to seperate SM from CM
		if (LES_attached == false)
		{
			sm->SetThrusterGroupLevel (THGROUP_ATT_LEFT, 1.0);
			sm->SetThrusterGroupLevel (THGROUP_ATT_BACK, 1.0);
		}

		// Play sound effect
		PlayVesselWave (OrbiterSoundID, SOUND_STAGESEP);

		// Update vessel state
		DelMesh (mesh_SM);

		if (mesh_SBayCover) // If SIM bay cover is present delete it from CM and add it to Service module
		{
			DelMesh (mesh_SBayCover);
			sm->AddMesh (mh_sbaycover, &SM_SIMBAY_OFFSET);	// SIM Bay cover mesh
		}

		SM_attached = false;
		return true;
	} // End "if (SM_attached)"
	else return false;

} // End "CSM::ServiceModuleSep"

// --------------------------------------------------------------
// Docking probe jettison routine
// --------------------------------------------------------------
bool CSM::ProbeSep (void)
{
	// Check CG offset 
	VECTOR3 ofs;
	if (ReEntryConfig) ofs = _V( 0, 0, 0);
	else ofs = CSM_CM_OFFSET; 

	if (DockingProbe) // Sanity check, is there an probe to jettison?
	{
		char name[256];
		strcpy (name, GetName()); 
		strcat (name, "_DockingRing");
		OBJHANDLE oh_probe = SpawnObject (this, name, "AAPO/Parts/CM_Probe", CM_DOCK_POS + ofs, 0.33);
		VESSEL *probe = oapiGetVesselInterface (oh_probe);

		int idx	= 0;												// dock index
		OBJHANDLE oh_docked = GetDockStatus (dh_cm_dock);			// object handle of docked vessel

		if (oh_docked != NULL)	// make sure handle is valid, if it is...
		{
			VESSEL *docked = oapiGetVesselInterface (oh_docked);	// get VESSEL interface for docked vessel
			int dcount = docked->DockCount ();						// get number of docking ports

			for (int i = 0; i < dcount; i++)	// Cycle through ports
			{
				if (docked->GetDockStatus (docked->GetDockHandle (i)) == GetHandle ())
				{
					idx = i;
					docked->Undock (idx, 0);
					docked->Dock (oh_probe, idx, 0, 1);
				}
			}
		}

		Undock  (ALLDOCKS, 0);			// Release any docked vessels
		PlayVesselWave (OrbiterSoundID, SOUND_EXPLSVDEVICE); // Play sound effect

		DelMesh (mesh_Probe);			// Delete mesh
		DockingProbe = false;			// Update state
		return true;
	}
	else return false;
} // End "CSM::ProbeSep"

// --------------------------------------------------------------
// Command module apex (parachute cover) jettison routine
// --------------------------------------------------------------
bool CSM::ApexSep (void)
{
	// Check CG offset 
	VECTOR3 ofs;
	if (ReEntryConfig) ofs = _V( 0, 0, 0);
	else ofs = CSM_CM_OFFSET; 

	if (fabs(GetAOA ()) < 5 * PI / 6) return false;
	
	if (ApexCovered) // Sanity check, is there a cover to jettison?
	{
		// Create cover as seperate vessel
		char name[256];
		strcpy (name, GetName()); 
		strcat (name, "_ApexCover");
		OBJHANDLE oh_apex = SpawnObject (this, name, "AAPO/Parts/CM_Apex", CM_APEX_OFFSET + ofs, 1.0);
		VESSEL *apex = oapiGetVesselInterface (oh_apex);

		PlayVesselWave (OrbiterSoundID, SOUND_EXPLSVDEVICE); // Play sound effect

		if (DockingProbe)
		{
			int idx	= 0;												
			ProbeSep ();

			strcpy (name, GetName()); 
			strcat (name, "_DockingRing");

			OBJHANDLE oh_probe = oapiGetVesselByName  (name);		// object handle of docking probe

			if (oh_probe != NULL)
			{
				VESSEL *probe = oapiGetVesselInterface (oh_probe);	// get VESSEL interface for probe
				probe->AttachChild (oh_apex, probe->GetAttachmentHandle (false, 0), apex->GetAttachmentHandle (true, 0));

			}
		}

		// Update vessel state
		DelMesh (mesh_Apex);
		ApexCovered = false;
		return true;

	} // End "if (ApexCovered)"
	else return false;
} // End "CSM::ApexSep"

// --------------------------------------------------------------
// SIM Bay cover jettison routine
// --------------------------------------------------------------
bool CSM::OpenBayCover (void)
{
	if ((SM_attached) && (SM_BayOpen == false))
	{
		char name[256];
		strcpy (name, GetName()); 
		strcat (name, "_BayCover");
		OBJHANDLE oh_cover = SpawnObject (this, name, "AAPO/Parts/SM_BayCover", SM_SIMBAY_OFFSET + CSM_SM_OFFSET, 0.0);
		VESSEL *cover = oapiGetVesselInterface (oh_cover);

		double alpha = RAD * 45;
		double beta = RAD *-38;

		VECTOR3 vforce = RotateVector ( _V(0, sin(alpha),-cos(alpha)), beta,_V(0, 0, 1));
		cover->AddForce ( vforce * 1500, _V( 0, 0, 0.8));

		PlayVesselWave (OrbiterSoundID, SOUND_SMSEP); // Play sound effect

		DelMesh (mesh_SBayCover);
		SM_BayOpen = true;
		return true;
	}
	else
	{
		if (mesh_SBayCover) DelMesh (mesh_SBayCover);
		return false;
	}
}

// --------------------------------------------------------------
// Deploy stabilzation drogues
// --------------------------------------------------------------
void CSM::DeployDrogues (void)
{
	if ((ApexCovered == false) && (Drogues_Deployed == false)) // Sanity check
	{
		// Set time accel to 1 to prevent glitching
		if ((oapiGetFocusInterface () == this) && (oapiGetTimeAcceleration () > 1)) oapiSetTimeAcceleration (1);

		// Clear existing chutes and drag elements
		if (CSM::GetAttachmentStatus (ah_chutes) != NULL)
		{
			DetachChild (ah_chutes, 0.1);
			ClearVariableDragElements ();
			PlayVesselWave (OrbiterSoundID, SOUND_EXPLSVDEVICE); // Play sound effect
		}

		// Create drogues
		char name[256];
		strcpy (name, GetName()); 
		strcat (name, "_Drogues");
		OBJHANDLE oh_drogue = SpawnObject (this, name, "AAPO/Parts/CM_Drogue", CM_CHUTE_ATTACH, 0.1); 
		VESSEL *drogue = oapiGetVesselInterface (oh_drogue);
		AttachChild (oh_drogue, ah_chutes, drogue->GetAttachmentHandle (true, 0));

		//PlayVesselWave (OrbiterSoundID, SOUND_EXPLSVDEVICE); // Play sound effect

		// Create drag elements
		for (int i = 0; i < 2; i++) CreateVariableDragElement (&drogue_proc, CM_CHUTE_DROGUEAREA, CM_CHUTE_ATTACH);
		
		SetAnimation (anim_droguedeploy, 1);
		Drogues_Deployed = true;
	}

} // End "CSM::DeployDrogues"

// --------------------------------------------------------------
// Deploy main parachutes
// --------------------------------------------------------------
void CSM::DeployChutes (void)
{
	int i = 0; 

	if ((Chutes_Deployed == false) && (!ApexCovered))
	{
		// Set time accel to 1 to prevent glitching
		if ((oapiGetFocusInterface () == this) && (oapiGetTimeAcceleration () > 1)) oapiSetTimeAcceleration (1);
		
		// Clear existing chutes and drag elements
		if (CSM::GetAttachmentStatus (ah_chutes) != NULL)
		{
			DetachChild (ah_chutes, 0.1);
			ClearVariableDragElements ();
			PlayVesselWave (OrbiterSoundID, SOUND_EXPLSVDEVICE); // Play sound effect
		}
	
		// Create chutes
		char name[256];
		strcpy (name, GetName()); 
		strcat (name, "_Parachutes");
		OBJHANDLE oh_mains = SpawnObject (this, name, "AAPO/Parts/CM_Chute", CM_CHUTE_ATTACH, 0.1); 
		VESSEL *mains = oapiGetVesselInterface (oh_mains);
		AttachChild (oh_mains, ah_chutes, mains->GetAttachmentHandle (true, 0));		

		// Create drag elements
		for (i = 0; i < 3; i++) CreateVariableDragElement (&chute_proc, CM_CHUTE_MAINAREA, CM_CHUTE_ATTACH);
	
		SetAnimation (anim_chutedeploy, 1);
		Chutes_Deployed = true;
	}
} // End "CSM::DeployChutes"

// --------------------------------------------------------------
// Activate Launch Escape System
// --------------------------------------------------------------
void CSM::LaunchAbort (void)
{
	if (LES_attached)
	{
		OBJHANDLE	oh_LV, oh_SLA, oh_SM, oh_LM;
		oh_LM	= GetAttachmentStatus (ah_payload);				// get handle for Lunar module
		oh_LV	= GetAttachmentStatus (ah_booster);				// get handle for Launch vehicle
	
		if (SM_attached)
		{
			ServiceModuleSep ();

			char name[256];
			strcpy (name, GetName()); 
			strcat (name, "_ServiceModule");
			oh_SM = oapiGetObjectByName (name);
			VESSEL *sm = oapiGetVesselInterface (oh_SM);

			if (SLA_attached) // Sanity check, is there an SLA to seperate from?
			{
				AdapterSep ();

				char name[256];
				strcpy (name, GetName()); 
				strcat (name, "_SLA");
				oh_SLA = oapiGetObjectByName (name);
				VESSEL *sla	= oapiGetVesselInterface (oh_SLA);

				// Attach SM to SLA
				sla->AttachChild (oh_SM, sla->GetAttachmentHandle (false, 1), sm->GetAttachmentHandle (true, 0));

			} // End "if (SLA_attached)"
			else if (oh_LV != NULL)
			{
				VESSEL *lv = oapiGetVesselInterface (oh_LV);
				lv->AttachChild (oh_SM, lv->GetAttachmentHandle (false, 0), sm->GetAttachmentHandle (true, 0));
			}
		} // End "if (SM_attached)"

		// Create LES as seperate vessel
		char name[256];
		strcpy (name, GetName()); 
		strcat (name, "_LES");
		OBJHANDLE oh_LES = SpawnObject (this, name, "AAPO/Parts/LES", CM_ABORT_OFFSET, 0.0); 

		Dock (oh_LES, 0, 0, 1);
		VESSEL *les = oapiGetVesselInterface (oh_LES);			// Get interface for LES vessel
		les->SetThrusterGroupLevel (THGROUP_MAIN, 1.0);			// Fire LES's jettison motors

		DelMesh (mesh_LES);										// Delete mesh
		LES_attached = false;									// Update state
		AbortInProgress = true;

	} // End "if (LES_attached)"
} // End "LaunchAbort ()"

// --------------------------------------------------------------
// Set Re-Entry configuration (Update physical parameters and flight model)
// --------------------------------------------------------------
void CSM::SetReEntryConfig (void)
{
	if (ReEntryConfig == false) // If Re-Entry Config has not been set...
	{
		// Shift CoG
		ShiftCG (CSM_CM_OFFSET);								

		// Update physical parameters
		SetSize (CM_SIZE);										// Update size
		SetPMI (CM_PMI);										// Update moments of inertia
		SetCrossSections (CM_CS);								// Update cross sections
		SetTouchdownPoints (CM_TDP[0], CM_TDP[1], CM_TDP[2]);	// Update touchdown points
		SetCW ( CM_CW[0], CM_CW[1], CM_CW[2], CM_CW[3]);		// Update wind resistance coefficients

		// Update Aerodynamic model
		SetRotDrag (CM_RD);	
		CreateAirfoil (LIFT_VERTICAL,-CM_COP, &VLiftCoeff, 3.5, 11.95, 1.0);
		CreateAirfoil (LIFT_HORIZONTAL,_V(0.0, 0.0, 0.0), &HLiftCoeff, 3.5, 11.95, 1.0);	

		// Update config
		ReEntryConfig = true;									
	}

	// Mesh clean-up
	if (mesh_SM) DelMesh (mesh_SM);
	if (mesh_SBayCover) DelMesh (mesh_SBayCover);

	// TEMPORARY THRUST CONTROL CONTROL GROUPS
	THRUSTER_HANDLE	th_group[4];

	// RCS control groups (Rotation)
	th_group[0]	= th_cm_rcs[2];
	th_group[1] = th_cm_rcs[3];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHUP);

	th_group[0]	= th_cm_rcs[0];
	th_group[1] = th_cm_rcs[1];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHDOWN);

	th_group[0]	= th_cm_rcs[4];
	th_group[1] = th_cm_rcs[7];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKLEFT); 

	th_group[0]	= th_cm_rcs[5];
	th_group[1] = th_cm_rcs[6];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKRIGHT);

	th_group[0]	= th_cm_rcs[10];
	th_group[1] = th_cm_rcs[11];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0]	= th_cm_rcs[8];
	th_group[1] = th_cm_rcs[9];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);	
}

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void CSM::clbkSetClassCaps (FILEHANDLE cfg)
{
	int i = 0;

	// Read dificulty modifiers from config file
	oapiReadItem_float (cfg, "THRUST_MODIFIER", thrust_mod);			// Max thrust multiplier 
	oapiReadItem_float (cfg, "DV_MODIFIER", dv_mod);					// Exhaust velocity multiplier
	oapiReadItem_float (cfg, "CONSUMABLES_MODIFIER", consumables_mod);	// Consumables consumption rate (1 = 100% of historical rate)

	// Initialize UMMU 2.5
	UmmuSetClassCaps (cfg);

	// Set physical parameters
	SetSize (CSM_SIZE);
	SetEmptyMass (CM_EMPTYMASS + SM_EMPTYMASS);
	SetPMI (CSM_PMI);
	SetCrossSections (CSM_CS);
	SetRotDrag (CSM_RD);
	//SetTouchdownPoints (CSM_TDP[0], CSM_TDP[1], CSM_TDP[2]);
	SetSurfaceFrictionCoeff (1, 1);
	SetCW (CSM_CW[0], CSM_CW[1], CSM_CW[2], CSM_CW[3]);

	// docking port definitions
	dh_cm_dock		= CreateDock (CM_DOCK_POS + CSM_CM_OFFSET, CM_DOCK_DIR, CM_DOCK_ROT);

	// Attachment point definitions
	ah_booster		= CreateAttachment (true, SLA_ATTACH_BOOSTPOS + CSM_SLA_OFFSET, SLA_ATTACH_BOOSTDIR, SLA_ATTACH_BOOSTROT, "X");	// for attachment to launch vehicle
	ah_payload		= CreateAttachment (false, SLA_DOCK_POS + CSM_SLA_OFFSET, SLA_DOCK_DIR, SLA_DOCK_ROT, "X");	// for attachment of lunar module to SLA
	ah_abort		= CreateAttachment (true, CM_DOCK_POS + CSM_CM_OFFSET, CM_DOCK_DIR, CM_DOCK_ROT, "X");
	ah_chutes		= CreateAttachment (false, CM_CHUTE_ATTACH + CSM_CM_OFFSET, _V( 0, 0, 1), _V( 0, 1, 0), "X");
	
	// Command Module propellant resources
	ph_cm_O2		= CreatePropellantResource (CM_O2MASS);
	ph_cm_H2O		= CreatePropellantResource (CM_WATERMASS);
	ph_cm_Waste		= CreatePropellantResource (CM_WASTEWATERMASS);
	for (i = 0; i < 2; i ++) ph_cm_rcs[i] = CreatePropellantResource (CM_RCS_FUELMASS);
	
	SetPropellantMass (ph_cm_Waste, 0.0); // Waste water tank starts empty

	// Service Module propellant resources
	ph_sm_O2 = CreatePropellantResource (SM_O2MASS);
	ph_sm_H2 = CreatePropellantResource (SM_H2MASS);
	ph_sps	= CreatePropellantResource (SM_FUELMASS);
	for (int i = 0; i < 4; i ++) ph_sm_rcs[i] = CreatePropellantResource (SM_RCS_FUELMASS);

	// Main engine
	th_sps	= CreateThruster (SM_ENGINEPOS + CSM_SM_OFFSET, _V( 0, 0, 1), SM_MAXTHRUST * thrust_mod, ph_sps, SM_ISP * dv_mod);	

	// Command Module RCS engines
	double alpha, beta;
	VECTOR3 shift; 
	alpha	= RAD * (90 - 32);
	
	// Command module Pitch
	th_cm_rcs[0]	= CreateThruster (CM_RCS_PITCHPOS[0] + CSM_CM_OFFSET, _V( 0,-1, 0), CM_RCS_MAXTHRUST, ph_cm_rcs[0], CM_RCS_ISP * dv_mod);	// forward
	th_cm_rcs[1]	= CreateThruster (CM_RCS_PITCHPOS[1] + CSM_CM_OFFSET, _V( 0,-1, 0), CM_RCS_MAXTHRUST, ph_cm_rcs[1], CM_RCS_ISP * dv_mod);	
	th_cm_rcs[2]	= CreateThruster (CM_RCS_PITCHPOS[2] + CSM_CM_OFFSET, _V( 0,-sin(alpha),-cos(alpha)), CM_RCS_MAXTHRUST, ph_cm_rcs[0], CM_RCS_ISP * dv_mod);	// aft
	th_cm_rcs[3]	= CreateThruster (CM_RCS_PITCHPOS[3] + CSM_CM_OFFSET, _V( 0,-sin(alpha),-cos(alpha)), CM_RCS_MAXTHRUST, ph_cm_rcs[1], CM_RCS_ISP * dv_mod); 
	
	// Command module Roll
	th_cm_rcs[4]	= CreateThruster (CM_RCS_ROLLPOS[0] + CSM_CM_OFFSET, _V( 0,-1, 0), CM_RCS_MAXTHRUST, ph_cm_rcs[0], CM_RCS_ISP * dv_mod);	// port
	th_cm_rcs[5]	= CreateThruster (CM_RCS_ROLLPOS[1] + CSM_CM_OFFSET, _V( 1, 0, 0), CM_RCS_MAXTHRUST, ph_cm_rcs[0], CM_RCS_ISP * dv_mod); 
	th_cm_rcs[6]	= CreateThruster (CM_RCS_ROLLPOS[2] + CSM_CM_OFFSET, _V( 0,-1, 0), CM_RCS_MAXTHRUST, ph_cm_rcs[1], CM_RCS_ISP * dv_mod);	// stbd	
	th_cm_rcs[7]	= CreateThruster (CM_RCS_ROLLPOS[3] + CSM_CM_OFFSET, _V(-1, 0, 0), CM_RCS_MAXTHRUST, ph_cm_rcs[1], CM_RCS_ISP * dv_mod); 

	// Command module Yaw
	th_cm_rcs[8]	= CreateThruster (CM_RCS_YAWPOS[0] + CSM_CM_OFFSET, _V( sin(alpha), 0,-cos(alpha)), CM_RCS_MAXTHRUST, ph_cm_rcs[0], CM_RCS_ISP * dv_mod);	// port
	th_cm_rcs[9]	= CreateThruster (CM_RCS_YAWPOS[1] + CSM_CM_OFFSET, _V( sin(alpha), 0,-cos(alpha)), CM_RCS_MAXTHRUST, ph_cm_rcs[1], CM_RCS_ISP * dv_mod); 
	th_cm_rcs[10]	= CreateThruster (CM_RCS_YAWPOS[2] + CSM_CM_OFFSET, _V(-sin(alpha), 0,-cos(alpha)), CM_RCS_MAXTHRUST, ph_cm_rcs[1], CM_RCS_ISP * dv_mod);	// stbd
	th_cm_rcs[11]	= CreateThruster (CM_RCS_YAWPOS[3] + CSM_CM_OFFSET, _V(-sin(alpha), 0,-cos(alpha)), CM_RCS_MAXTHRUST, ph_cm_rcs[0], CM_RCS_ISP * dv_mod); 

	// Service Module RCS engines
	alpha	= RAD * 10;
	beta	= RAD * 9.25;
	shift	= _V( 0.0, 0.0, 0.137);

	// Service module Quadrant A (12 o'clock)
	th_sm_rcs[ 0]	= CreateThruster (SM_RCS_ENGINEPOS[0] + CSM_SM_OFFSET - shift, RotateVector (_V( 0,-sin(alpha), cos(alpha)), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[0], SM_RCS_ISP * dv_mod);	// Fwd
	th_sm_rcs[ 1]	= CreateThruster (SM_RCS_ENGINEPOS[0] + CSM_SM_OFFSET + shift, RotateVector (_V( 0,-sin(alpha),-cos(alpha)), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[0], SM_RCS_ISP * dv_mod);	// Back
	th_sm_rcs[ 2]	= CreateThruster (SM_RCS_ENGINEPOS[0] + CSM_SM_OFFSET - shift, RotateVector (_V( cos(alpha),-sin(alpha), 0), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[0], SM_RCS_ISP * dv_mod);	// Right
	th_sm_rcs[ 3]	= CreateThruster (SM_RCS_ENGINEPOS[0] + CSM_SM_OFFSET + shift, RotateVector (_V(-cos(alpha),-sin(alpha), 0), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[0], SM_RCS_ISP * dv_mod);	// Left
	
	// Service module  Quadrant B (3 o'clock)
	th_sm_rcs[ 4]	= CreateThruster (SM_RCS_ENGINEPOS[1] + CSM_SM_OFFSET - shift, RotateVector (_V( sin(alpha), 0, cos(alpha)), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[1], SM_RCS_ISP * dv_mod);	// Fwd
	th_sm_rcs[ 5]	= CreateThruster (SM_RCS_ENGINEPOS[1] + CSM_SM_OFFSET + shift, RotateVector (_V( sin(alpha), 0,-cos(alpha)), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[1], SM_RCS_ISP * dv_mod);	// Back
	th_sm_rcs[ 6]	= CreateThruster (SM_RCS_ENGINEPOS[1] + CSM_SM_OFFSET - shift, RotateVector (_V( sin(alpha), cos(alpha), 0), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[1], SM_RCS_ISP * dv_mod);	// Up
	th_sm_rcs[ 7]	= CreateThruster (SM_RCS_ENGINEPOS[1] + CSM_SM_OFFSET + shift, RotateVector (_V( sin(alpha),-cos(alpha), 0), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[1], SM_RCS_ISP * dv_mod);	// Down

	// Service module  Quadrant C (6 o'clock)
	th_sm_rcs[ 8]	= CreateThruster (SM_RCS_ENGINEPOS[2] + CSM_SM_OFFSET - shift, RotateVector (_V( 0, sin(alpha), cos(alpha)), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[2], SM_RCS_ISP * dv_mod);	// Fwd
	th_sm_rcs[ 9]	= CreateThruster (SM_RCS_ENGINEPOS[2] + CSM_SM_OFFSET + shift, RotateVector (_V( 0, sin(alpha),-cos(alpha)), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[2], SM_RCS_ISP * dv_mod);	// Back
	th_sm_rcs[10]	= CreateThruster (SM_RCS_ENGINEPOS[2] + CSM_SM_OFFSET + shift, RotateVector (_V( cos(alpha), sin(alpha), 0), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[2], SM_RCS_ISP * dv_mod);	// Right
	th_sm_rcs[11]	= CreateThruster (SM_RCS_ENGINEPOS[2] + CSM_SM_OFFSET - shift, RotateVector (_V(-cos(alpha), sin(alpha), 0), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[2], SM_RCS_ISP * dv_mod);	// Left

	// Service module  Quadrant D (9 o'clock)
	th_sm_rcs[12]	= CreateThruster (SM_RCS_ENGINEPOS[3] + CSM_SM_OFFSET - shift, RotateVector (_V(-sin(alpha), 0, cos(alpha)), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[3], SM_RCS_ISP * dv_mod);	// Fwd
	th_sm_rcs[13]	= CreateThruster (SM_RCS_ENGINEPOS[3] + CSM_SM_OFFSET + shift, RotateVector (_V(-sin(alpha), 0,-cos(alpha)), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[3], SM_RCS_ISP * dv_mod);	// Back
	th_sm_rcs[14]	= CreateThruster (SM_RCS_ENGINEPOS[3] + CSM_SM_OFFSET + shift, RotateVector (_V(-sin(alpha), cos(alpha), 0), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[3], SM_RCS_ISP * dv_mod);	// Up
	th_sm_rcs[15]	= CreateThruster (SM_RCS_ENGINEPOS[3] + CSM_SM_OFFSET - shift, RotateVector (_V(-sin(alpha),-cos(alpha), 0), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[3], SM_RCS_ISP * dv_mod);	// Down
	
	// Associate meshes for the visual
	mesh_Cockpit	= AddMesh (mh_cockpit, &CSM_CM_OFFSET);		// Virtual cockpit mesh

	mesh_CM			= AddMesh (mh_command, &CSM_CM_OFFSET);							// Command module mesh
	mesh_Apex		= AddMesh (mh_apex, &(CSM_CM_OFFSET + CM_APEX_OFFSET));			// CM Parachute cover mesh
	mesh_Probe		= AddMesh (mh_probe, &(CM_DOCK_POS + CSM_CM_OFFSET));			// CM Docking probe mesh

	mesh_SM			= AddMesh (mh_service, &CSM_SM_OFFSET);							// Service module mesh
	mesh_SBayCover	= AddMesh (mh_sbaycover, &(CSM_SM_OFFSET + SM_SIMBAY_OFFSET));	// Service module SIM Bay cover

	mesh_LES		= AddMesh (mh_tower, &(CSM_CM_OFFSET + CM_ABORT_OFFSET));		// Abort tower mesh
	mesh_SLA		= AddMesh (mh_adapter, &CSM_SLA_OFFSET);						// Launch adapter mesh

	// Set mesh visibility modes 
	SetMeshVisibilityMode (mesh_Cockpit, MESHVIS_VC);	
	SetMeshVisibilityMode (mesh_SM, MESHVIS_ALWAYS);
	SetMeshVisibilityMode (mesh_LES, MESHVIS_ALWAYS);
}

// --------------------------------------------------------------
// Read status from scenario file
// --------------------------------------------------------------
void CSM::clbkLoadStateEx (FILEHANDLE scn, void *status)
{
	int i = 0;
	char *cbuf; 
	int temp = 0;

	while (oapiReadScenario_nextline (scn, cbuf)) 
	{
		// Check for presence of Abort Tower  
		if (!_strnicmp (cbuf, "LES_ATTACHED", 12))	// *Boolean parser courtesy of dbeachy1*
		{
			temp = 0;								// initialize temp in case sscanf fails to parse the integer; default to 0 (false) if the parse fails.  
			sscanf (cbuf+12, "%i", &temp);			// write 4 bytes to temp -- no need to typecast the pointer since it's just a void * anyway
			LES_attached = (temp != 0);				// to be tidy, a bool's value should either be true (1) or false (0)
		}

		// Check for presence of Launch Adapter
		else if (!_strnicmp (cbuf, "SLA_ATTACHED", 12)) // 
		{
			temp = 0;					  
			sscanf (cbuf+12, "%i", &temp);  
			SLA_attached = (temp != 0);
		}

		// Check for presence of Service Module
		else if (!_strnicmp( cbuf, "SM_ATTACHED", 11))
		{
			temp = 0;					 
			sscanf (cbuf+11, "%i", &temp);
			SM_attached = (temp != 0);
		}

		// Is SM bay open?
		else if (!_strnicmp( cbuf, "SM_BAYOPEN", 10))
		{
			temp = 0;					 
			sscanf (cbuf+10, "%i", &temp);
			SM_BayOpen = (temp != 0);
		}

		// Check for presence of Docking Probe
		else if (!_strnicmp( cbuf, "DOCKING_PROBE", 13))
		{
			temp = 0;					 
			sscanf (cbuf+13, "%i", &temp);
			DockingProbe = (temp != 0);
		}

		// Check for presence of CM forward heat sheild (Apex Cover)
		else if (!_strnicmp (cbuf, "APEX_COVER", 10))
		{
			temp = 0;
			sscanf (cbuf+10, "%i", &temp);
			ApexCovered = (temp != 0);
		}

		else if (!_strnicmp (cbuf, "DROGUES_DEPLOYED", 16))
		{
			temp = 0;
			sscanf (cbuf+16, "%i %lf", &temp, &drogue_proc);
			Drogues_Deployed = (temp != 0);
		}

		else if (!_strnicmp (cbuf, "CHUTES_DEPLOYED", 15))
		{
			temp = 0;
			sscanf (cbuf+15, "%i %lf", &temp, &chute_proc);
			Chutes_Deployed = (temp != 0);
		}

		// Load animation states
        else if (!_strnicmp (cbuf, "HATCH", 5)) // find line labeled "HATCH" in scn file
		{
			sscanf (cbuf+5, "%i %lf", &HatchStatus, &hatch_proc);		// Read values stored there to HatchStatus and hatch_proc
		}		
		else if (!_strnicmp (cbuf, "HGA", 3)) // find line labeled "GEAR" in scn file
		{
			sscanf (cbuf+3, "%i %lf", &AntennaStatus, &antenna_proc);	// Read values stored there to GearStatus and gear_proc
		}
	
		// Load cockpit panel state
		else if (vc->ParseScenarioLine (cbuf)) continue;	

		// Load CSM systems state
		//else if (csm->ParseScenarioLine (cbuf)) continue;						

		// Load UMMUs
		else if (Crew.LoadAllMembersFromOrbiterScenario (cbuf)) continue;

		// Load UCGO cargoes
		else if (Cargo.LoadCargoFromScenario (cbuf)) continue;

		// Load default parameters
		else ParseScenarioLineEx (cbuf, status);

	} // End "while (oapiReadScenario_nextline (scn, cbuf))"

} // End "CSM::clbkLoadStateEx"

// --------------------------------------------------------------
// Write status to scenario file
// --------------------------------------------------------------
void CSM::clbkSaveState (FILEHANDLE scn)
{
	int i = 0;
	char cbuf[256];
	
	// Write default vessel parameters to scenario file
	VESSEL3::clbkSaveState (scn);

	// Write custom parameters to scenario file	
	sprintf (cbuf, "%i %0.4f", HatchStatus, hatch_proc);		// EVA Hatch status and animation state
	oapiWriteScenario_string (scn, "HATCH", cbuf);

	sprintf (cbuf, "%i %0.4f", LightStatus, light_proc);		// EVA light status and animation state
	oapiWriteScenario_string (scn, "LIGHT", cbuf);

	sprintf (cbuf, "%i %0.4f", AntennaStatus, antenna_proc);	// High Gain Antenna status and animation state
	oapiWriteScenario_string (scn, "HGA", cbuf);
	
	sprintf (cbuf, "%i", LES_attached);						
	oapiWriteScenario_string (scn, "LES_ATTACHED", cbuf);

	sprintf (cbuf, "%i", SLA_attached);						
	oapiWriteScenario_string (scn, "SLA_ATTACHED", cbuf);

	sprintf (cbuf, "%i", SM_attached);						
	oapiWriteScenario_string (scn, "SM_ATTACHED", cbuf);

	sprintf (cbuf, "%i", SM_BayOpen);						
	oapiWriteScenario_string (scn, "SM_BAYOPEN", cbuf);

	sprintf (cbuf, "%i", DockingProbe);						
	oapiWriteScenario_string (scn, "DOCKING_PROBE", cbuf);

	sprintf (cbuf, "%i", ApexCovered);						
	oapiWriteScenario_string (scn, "APEX_COVER", cbuf);

	sprintf (cbuf, "%i %0.4f", Drogues_Deployed, drogue_proc);						
	oapiWriteScenario_string (scn, "DROGUES_DEPLOYED", cbuf);

	sprintf (cbuf, "%i %0.4f", Chutes_Deployed, chute_proc);						
	oapiWriteScenario_string (scn, "CHUTES_DEPLOYED", cbuf);

	// Save cockpit panel state
	vc->SaveState (scn);

	// Save subsystems status
	//csm->SaveState (scn);

	// Save UMmu state
	Crew.SaveAllMembersInOrbiterScenarios (scn);

	// Save UCGO cargo state
	Cargo.SaveCargoToScenario(scn);

} // End "CSM::clbkSaveState"

// --------------------------------------------------------------
// Finalise vessel creation
// --------------------------------------------------------------
void CSM::clbkPostCreation ()
{
	DefineAnimations ();
	DefineLighting ();
	DefineParticleStreams ();
	DefineSounds ();
	
	vc->InitVC (mesh_Cockpit);
	csm->InitCSM ();

	if (SM_attached == false) SetReEntryConfig ();
	if (SM_BayOpen) DelMesh (mesh_SBayCover);

	if (LES_attached == false) DelMesh (mesh_LES);
	if (SLA_attached == false) DelMesh (mesh_SLA);	
	if (DockingProbe == false) DelMesh (mesh_Probe);
	if (ApexCovered == false) DelMesh (mesh_Apex);

	// Create drag elements as required
	if (Chutes_Deployed) for (int i = 0; i < 3; i++) CreateVariableDragElement (&chute_proc, CM_CHUTE_MAINAREA, CM_CHUTE_ATTACH);
	else if (Drogues_Deployed) for (int i = 0; i < 2; i++) CreateVariableDragElement (&drogue_proc, CM_CHUTE_DROGUEAREA, CM_CHUTE_ATTACH);


		// TEMPORARY THRUST CONTROL CONTROL GROUPS
		THRUSTER_HANDLE	th_group[4];

		// RCS control groups (Rotation)
		th_group[0]	= th_sm_rcs[1];
		th_group[1] = th_sm_rcs[8];
		CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHUP);

		th_group[0]	= th_sm_rcs[0];
		th_group[1] = th_sm_rcs[9];
		CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHDOWN);

		th_group[0]	= th_sm_rcs[7];
		th_group[1] = th_sm_rcs[14];
		CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKLEFT);

		th_group[0]	= th_sm_rcs[2];
		th_group[1] = th_sm_rcs[11];
		CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKRIGHT);

		th_group[0]	= th_sm_rcs[4];
		th_group[1] = th_sm_rcs[13];
		CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

		th_group[0]	= th_sm_rcs[5];
		th_group[1] = th_sm_rcs[12];
		CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);

		// RCS control groups (Translation)
		th_group[0] = th_sm_rcs[6];
		th_group[1] = th_sm_rcs[14];
		CreateThrusterGroup (th_group, 2, THGROUP_ATT_UP);

		th_group[0] = th_sm_rcs[7];
		th_group[1] = th_sm_rcs[15];
		CreateThrusterGroup (th_group, 2, THGROUP_ATT_DOWN);

		th_group[0] = th_sm_rcs[2];
		th_group[1] = th_sm_rcs[10];
		CreateThrusterGroup (th_group, 2, THGROUP_ATT_RIGHT);

		th_group[0] = th_sm_rcs[3];
		th_group[1] = th_sm_rcs[11];
		CreateThrusterGroup (th_group, 2, THGROUP_ATT_LEFT);
		
		th_group[0] = th_sm_rcs[0];
		th_group[1] = th_sm_rcs[4];
		th_group[2] = th_sm_rcs[8];
		th_group[3] = th_sm_rcs[12];
		CreateThrusterGroup (th_group, 4, THGROUP_ATT_FORWARD);

		th_group[0] = th_sm_rcs[1];
		th_group[1] = th_sm_rcs[5];
		th_group[2] = th_sm_rcs[9];
		th_group[3] = th_sm_rcs[13];
		CreateThrusterGroup (th_group, 4, THGROUP_ATT_BACK);

		// Main engine
		CreateThrusterGroup (&th_sps, 1, THGROUP_MAIN);
		// END TEMPORARY THRUSTER GROUPS

} // End "CSM::clbkPostCreation"

// --------------------------------------------------------------
// Pre-Step processes
// --------------------------------------------------------------
void CSM::clbkPreStep (double simt, double simdt, double mjd)
{
	// Update Vessel State
	SetEmptyMass (CSM_EmptyMass ());
	Booster_attached = (GetAttachmentStatus (ah_booster) != NULL);	// if GetAttachmentStatus returns a value other than NULL boster_attached returns true. 

	if (SLA_attached)		SetAttachmentParams (ah_booster, SLA_ATTACH_BOOSTPOS + CSM_SLA_OFFSET, SLA_ATTACH_BOOSTDIR, SLA_ATTACH_BOOSTROT);
	else if (SM_attached)	SetAttachmentParams (ah_booster, SM_ATTACH_POS + CSM_SM_OFFSET, SM_ATTACH_DIR, SM_ATTACH_ROT);
	else					SetAttachmentParams (ah_booster, CM_ATTACH_POS, CM_ATTACH_DIR, CM_ATTACH_ROT);

	// If Service Module is attached...
	if (SM_attached) 
	{
		SetDefaultPropellantResource (ph_sps);		// Set main fuel tank as default propellant resource
		SetISP (GetThrusterIsp0 (th_sps));			// Set main engine ISP as default
	}
	else // If Service Module IS NOT attached... 
	{
		// Update vessel parameters 
		SetReEntryConfig ();

		// Remove SM propellant resources
		SetPropellantMass (ph_sm_O2, 0.0);
		SetPropellantMass (ph_sm_H2, 0.0);
		SetPropellantMass (ph_sps, 0.0);
		for (int i = 0; i < 4; i++) SetPropellantMass (ph_sm_rcs[i], 0.0);
	}

	// 
	csm->PreStep (simt, simdt, mjd);
} // End "CSM::clbkPreStep"

// --------------------------------------------------------------
// Post-Step processes
// --------------------------------------------------------------
void CSM::clbkPostStep (double simt, double simdt, double mjd)
{
	double delta = simdt; 

	if (AbortInProgress)
	{
		if  ((GetAtmPressure() > 37600) && (fabs(GetAOA ()) > 5 * PI / 6)) TowerSep ();
		//sprintf (oapiDebugString (), "Pressure %0.2f, AOA %0.2f", GetAtmPressure(), fabs(GetAOA ()) * DEG);
	}

	if (SM_attached)
	{

	} // End "if (SM_attached)"
	else
	{
		if (GetAttachmentStatus (ah_chutes) == NULL)
		{
			SetRotDrag (CM_RD);
			ClearVariableDragElements ();
		}
		else
		{
			SetRotDrag (_V( 10.0, 10.0, 7.0));		
			
			VECTOR3 airstream;
			GetShipAirspeedVector (airstream);
			normalise (airstream);
			
			double alpha = 0;
			double beta = 0;
			if (GetAOA () != 0.0)
			{
				alpha = asin (airstream.x);
				beta = asin (airstream.y);
			}
			//sprintf (oapiDebugString (), "Alpha %0.2f, Beta %0.2f, AOA %0.2f", alpha * DEG, beta * DEG, GetAOA () * DEG);

			VECTOR3 dir, rot, temp;
		
			temp = RotateVector (_V( 0, 0, 1), alpha, _V( 0, 1, 0));
			dir = RotateVector (temp, beta, _V( 1, 0, 0));

			temp = RotateVector (_V( 0, 1, 0), alpha, _V( 0, 1, 0));
			rot = RotateVector (temp, beta, _V( 1, 0, 0));

			SetAttachmentParams (ah_chutes, CM_CHUTE_ATTACH, dir, rot);
		} // End "if (GetAttachmentStatus (ah_chutes) == NULL)"
	} // End "else"

	// Animate High Gain Antenna
	if (AntennaStatus > OPEN)	
	{
		double	delta = simdt / 1;							// delta = simdt divided by duration of animation in seconds

		if (AntennaStatus == OPENING)
		{
			if (antenna_proc < 1) antenna_proc += delta;	// If process value is less than 1 add delta to process value
			else AntennaStatus = OPEN;						// If process value is greater than 1 GearStatus = OPEN
		}
		else if (AntennaStatus == CLOSING)
		{
			if (antenna_proc > 0) antenna_proc -= delta;	// If process value is Greater than 0 subtract delta from process value
			else AntennaStatus = CLOSED;					// If process value is less than 0 GearStatus = CLOSED
		}
		SetAnimation (anim_antenna, antenna_proc);
	}

	// Animate EVA Hatch
	if (HatchStatus > OPEN)	
	{
		double	delta = simdt / 3;							// delta = simdt divided by duration of animation in seconds

		if (HatchStatus == OPENING)
		{
			if (hatch_proc < 1) hatch_proc += delta;	// If process value is less than 1 add delta to process value
			else HatchStatus = OPEN;					// If process value is greater than 1 set status to OPEN

			if (hatch_proc < 0.2) PlayVesselWave (OrbiterSoundID, SOUND_HATCHOPEN); // Play sound effect
		}
		else if (HatchStatus == CLOSING)
		{
			if (hatch_proc > 0) hatch_proc -= delta;	// If process value is Greater than 0 subtract delta from process value
			else HatchStatus = CLOSED;					// If process value is less than 0 set status to CLOSED

			if (hatch_proc < 0.3) PlayVesselWave (OrbiterSoundID, SOUND_HATCHCLOSE); // Play sound effect
		}
		
		SetAnimation (anim_Hatch, hatch_proc);			// Update animation state
	}

	//
	if (Drogues_Deployed)
	{
		delta = simdt / DROGUE_DTIME;
		if (drogue_proc < 1) drogue_proc += delta;
		else drogue_proc = 1;
	}

	//
	if (Chutes_Deployed)
	{
		delta = simdt / CHUTE_DTIME;
		if (chute_proc < 1) chute_proc += delta;
		else chute_proc = 1;
	}

	// 
	vc->PostStep (simt, simdt, mjd);
	csm->PostStep (simt, simdt, mjd);

	//
	CSM::UmmuPostStep (simt, simdt, mjd);

} // End "CSM::clbkPostStep"

// --------------------------------------------------------------
// Called after a vessel visual has been created by the renderer.
// --------------------------------------------------------------
void CSM::clbkVisualCreated (VISHANDLE vis, int refcount)
{
	// Set up cockpit visuals
	vc->VisualCreated  (vis, refcount);

	// Set up UCGO visuals
	Cargo.SetUcgoVisual (vis);	
}

// --------------------------------------------------------------
// Called before a vessel visual is destroyed.
// --------------------------------------------------------------
void CSM::clbkVisualDestroyed (VISHANDLE vis, int refcount)
{
	vc->VisualDestroyed (vis, refcount);
}

// --------------------------------------------------------------
// Setup the virtual cockpit HUD
// --------------------------------------------------------------
bool CSM::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
	if (oapiGetFocusInterface () != this) return false;

	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	// Ummu display messages
	if (dHudMessageDelay > 0) skp->Text (5, hps->H / 60 * 15, cUmmuHudDisplay, strlen (cUmmuHudDisplay));
	
	return true;
}

// --------------------------------------------------------------
// Docking event notification
// --------------------------------------------------------------
void CSM::clbkDockEvent (int dock, OBJHANDLE mate)
{
	if (DockingProbe == false) Undock (dock); // Do not allow vessels to dock if docking probe is not present
}

// --------------------------------------------------------------
// Process keyboard inputs
// --------------------------------------------------------------
int CSM::clbkConsumeBufferedKey (DWORD key, bool down, char *kstate)
{
	if (!down) return 0;

	// Grapple nearest UCGO Payload when [C] is pressed.
	if (key == OAPI_KEY_C && !KEYMOD_SHIFT(kstate) && !KEYMOD_CONTROL (kstate) && !KEYMOD_ALT(kstate))
	{ 
		UcgoGrappleCargo (iSelectedCargo);
		return 1;
	}

	// Jettison docking probe when [CTRL] + [Alt] + [D] is pressed.
	if (key == OAPI_KEY_D && !KEYMOD_SHIFT(kstate) && KEYMOD_CONTROL (kstate) && KEYMOD_ALT(kstate))
	{ 
		ProbeSep ();	
		return 1;
	}

	// Transfer selected UMMU when [E] is pressed.
	if (key == OAPI_KEY_E && !KEYMOD_SHIFT(kstate) && !KEYMOD_CONTROL(kstate) && !KEYMOD_ALT(kstate))
	{ 
		UmmuTransferCrewMember (iSelectedUmmuMember);
		return 1;									
	}

	// Activate Launch Escape System, or deploy parachutes when [CTRL] + [E] is pressed.
	if (key == OAPI_KEY_E && !KEYMOD_SHIFT(kstate) && KEYMOD_CONTROL(kstate) && !KEYMOD_ALT(kstate))
	{ 
		if (LES_attached) LaunchAbort ();
		else if (!Drogues_Deployed) DeployDrogues ();
		else DeployChutes ();
		return 1;									
	}

	// Deploy High Gain Antenna when [G] is pressed.
	if (key == OAPI_KEY_G  &&!KEYMOD_SHIFT(kstate) && !KEYMOD_CONTROL (kstate) && !KEYMOD_ALT(kstate)) // [K] is down, no [shift], no [ctrl], no [alt]
	{
		DeployHGA ();	
		return 1;			
	}

	// Release UCGO Payload when [J] is pressed.
	if (key == OAPI_KEY_J && !KEYMOD_SHIFT(kstate) && !KEYMOD_CONTROL (kstate) && !KEYMOD_ALT(kstate))
	{ 
		if (SM_BayOpen == false) OpenBayCover ();
		else UcgoReleaseCargo (iSelectedCargo);
		return 1;
	}

	// Jettison service module when [CTRL] + [J] is pressed.
	if (key == OAPI_KEY_J && !KEYMOD_SHIFT(kstate) && KEYMOD_CONTROL (kstate) && !KEYMOD_ALT(kstate))
	{ 
		if (LES_attached) TowerSep ();				
		else if (SLA_attached) AdapterSep ();
		else if (GetAttachmentStatus (ah_booster) != NULL) DetachParent (this, 0.01);
		else ServiceModuleSep ();
		return 1;
	}

	// Open hatch when [K] is pressed.
	if (key == OAPI_KEY_K  &&!KEYMOD_SHIFT(kstate) && !KEYMOD_CONTROL (kstate) && !KEYMOD_ALT(kstate)) // [K] is down, no [shift], no [ctrl], no [alt]
	{
		ActivateHatch ();
		return 1;
	}

	// Add new UMMU to vessel when [M] is pressed.
	if (key == OAPI_KEY_M && !KEYMOD_SHIFT(kstate) && !KEYMOD_CONTROL(kstate) && !KEYMOD_ALT(kstate))
	{
		AddUMmuToVessel (true);
		return 1;
	}

	// Select crewmember for EVA or tranfer when [S] is pressed.
	if (key == OAPI_KEY_S && !KEYMOD_SHIFT(kstate) && !KEYMOD_CONTROL(kstate) && !KEYMOD_ALT(kstate))
	{
		UmmuSelectCrewMember ();
		return 1;
	}

	// Show name of vessel and number of souls on board when [Ctrl] + [S] is pressed.
	if (key == OAPI_KEY_S && !KEYMOD_SHIFT(kstate) && KEYMOD_CONTROL(kstate) && !KEYMOD_ALT(kstate))
	{
		sprintf (SendHudMessage (),"Vessel: %s, Souls on board: %i, Cargo mass: %0.1f kg",
		GetName(), Crew.GetCrewTotalNumber (), Cargo.GetCargoTotalMass());
		return 1;
	}

	return 0;	// if no keys are pressed the function returns '0' and nothing happens

} // End "CSM::clbkConsumeBufferedKey"

// --------------------------------------------------------------
// Load generic "Glass" cockpit
// --------------------------------------------------------------
bool CSM::clbkLoadGenericCockpit ()
{
	// Check CG offset 
	VECTOR3 ofs;
	if (ReEntryConfig) ofs = _V( 0, 0, 0);
	else ofs = CSM_CM_OFFSET; 

	// Set camera parameters
	SetCameraOffset (_V(-0.55, 1.33, 0.20) + ofs);
	SetCameraDefaultDirection (_V( 0, 0, 1));
	oapiCameraSetCockpitDir ( 0, 0);								

	return true;
}

// --------------------------------------------------------------
// Load virtual cockpit
// --------------------------------------------------------------
bool CSM::clbkLoadVC (int id)
{
	// Check CG offset 
	VECTOR3 ofs;
	if (ReEntryConfig) ofs = _V( 0, 0, 0);
	else ofs = CSM_CM_OFFSET; 

	// Set camera parameters
	double alpha =-20 * RAD;
	SetCameraRotationRange (RAD*80, RAD*80, RAD*60, RAD*40);		// camera range of motion (Left, Right, Up, Down)
	SetCameraDefaultDirection ( _V( 0, sin(alpha), cos(alpha)));	// default camera direction (x,y,z)
	oapiCameraSetCockpitDir ( 0, 0);								// Center camera

	// Register Camera (view) properties
	switch (id)
	{
	case 0: // Left Seat
		SetCameraOffset (_V(-0.64, 1.16, 0.2) + ofs);	// Set camera position (x,y,z)
		oapiVCSetNeighbours (-1, 1,-1,-1);				// Set adjacent cameras (Left, Right, Top, Bottom)
		break;

	case 1: // Middle Seat
		SetCameraOffset (_V( 0.00, 1.16, 0.2) + ofs);	// Set camera position (x,y,z)
		oapiVCSetNeighbours ( 0, 2,-1,-1);				// Set adjacent cameras (Left, Right, Top, Bottom)
		break;

	case 2: // Right Seat
		SetCameraOffset (_V( 0.64, 1.16, 0.2) + ofs);	// Set camera position (x,y,z)
		oapiVCSetNeighbours ( 1,-1,-1,-1);				// Set adjacent cameras (Left, Right, Top, Bottom)
		break;
	} // end "switch (id)"
	
	// Register MFDs
	static VCMFDSPEC MFD0 = {mesh_Cockpit, GRP_VC_MFD_Screen};	// Surface on which to display MFD. (Mesh, Mesh Group)
	oapiVCRegisterMFD (0, &MFD0);								// Register MFD in orbiter's interface [registry # and VCMFDSPEC]. (registry numbers start at 0 and count up)

	//SURFHANDLE sh_PanelElements = oapiGetTextureHandle (mh_cockpit, 1);

	vc->RegisterActiveAreas (ofs);

	return true;
} // End "CSM::clbkLoadVC"

// --------------------------------------------------------------
// Respond to user inputs
// --------------------------------------------------------------
bool CSM::clbkVCMouseEvent (int id, int event, VECTOR3 &p)
{
	// Check to see if any one is on board before passing inputs to cockpit control class.
	if (Crew.GetCrewTotalNumber () > 0) return vc->MouseEvent (id, event, p);	
	else return false;
}

// --------------------------------------------------------------
// Respond to virtual cockpit area redraw requests
// --------------------------------------------------------------
bool CSM::clbkVCRedrawEvent (int id, int event, SURFHANDLE surf)					
{
	return vc->RedrawEvent (id, event, surf);
	return false;
}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new CSM (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (CSM*)vessel;
}