// ==============================================================
//                    ORBITER MODULE: CSM
//                  Part of the ORBITER SDK
//               Copyright (C) 2012 Greg Hlynka
//                    All rights reserved
//
// CSM.h
// Interface for CSM vessel class
//
// Notes: This file exists
// ==============================================================

#pragma once
#include "orbitersdk.h"
#include "AAPO_Common.h"

// ==============================================================
// Combined Vessel (Command + Service Module) Constants
// ==============================================================
const VECTOR3	CSM_CM_OFFSET		= { 0.00,-0.30, 2.00};		// Offset of command module from combined CoG
const VECTOR3	CSM_SM_OFFSET		= { 0.00, 0.00,-1.81};		// Offset of service module from combined CoG
const VECTOR3	CSM_SLA_OFFSET		= { 0.00, 0.00,-7.33};		// Offset of spacecraft launch adapter from combined CoG

const double	CSM_SIZE			= 2.41;						// Vessel's mean radius [m]
const VECTOR3	CSM_CS				= { 30.90, 30.80, 14.86};	// x,y,z cross sections [m^2]
const VECTOR3	CSM_PMI				= { 6.11, 5.96, 2.71};		// Principal moments of inertia (mass-normalised) [m^2]
const VECTOR3	CSM_RD				= { 0.70, 0.70, 0.30};		// Rotational drag coefficients
const VECTOR3	CSM_COM				= { 0.0, 0.0, 0.0};			// Center of mass
const double	CSM_CW[4]			= { 0.1, 0.3, 1.4, 1.4};	// Wind resistance coefficients (forward, aft, lateral, vertical)
const VECTOR3	CSM_TDP[3]			= {{ 0.00,-3.30,-11.58}, {-2.86, 4.25,-11.58}, { 2.86, 4.25,-11.58}};	// touchdown points [m]

// ==============================================================
// Command Module Constants
// ==============================================================
const VECTOR3	CM_APEX_OFFSET		= { 0.00, 0.30, 1.34};		// Position of Apex (parachute cover) relative to command module CoG
const VECTOR3	CM_ABORT_OFFSET		= { 0.00, 0.30, 4.75};		// Position of Abort Tower relative to command module CoG

const double	CM_SIZE				= 1.96;						// Mean radius [m]
const VECTOR3	CM_CS				= { 7.42, 7.40, 12.06};		// Cross sections [m^2]
const VECTOR3	CM_PMI				= { 1.41, 1.41, 1.86};		// Principal moments of inertia (mass-normalised) [m^2] without service module
const VECTOR3	CM_RD				= { 0.07, 0.07, 0.003};		// Rotation drag coefficients without service module
const VECTOR3	CM_COP				= { 0.00, 0.30,-1.03};		// Command Module center of pressure / lift relative to CoG

const double	CM_CW[4]			= { 5.5, 0.1, 3.4, 3.4};	// Wind resistance coefficients (forward, back, lateral, vertical)
const VECTOR3	CM_TDP[3]			= {{ 0.00,-2.00,-0.66}, {-1.00, 1.73,-0.66}, { 1.00, 1.73,-0.66}};		// touchdown points [m]

const double	CM_EMPTYMASS		= 5806.00;					// Empty mass [kg] (Encyclopedia Astronautica) 
const double	CM_O2MASS			= 1.67;						// Command module O2 reserve mass [kg] (3.7 lbs, Apollo Operations Handbook pg 2.7-7)
const double	CM_WATERMASS		= 14.96;					// Command module water reserve [kg] (33 lbs, Wikipedia)
const double	CM_WASTEWATERMASS	= 26.31;					// Command module waste water storage capacity [kg] (58 lbs, Wikipedia)
const double	CM_RCS_FUELMASS		= 122.47 / 2;				// Command module RCS fuel mass [kg] (270 lbs split between two tanks, Wikipedia)
const int		CM_TANKCOUNT		= 5;						// Number of command module  propellant recources

const double	CM_RCS_ISP			= 2844.90;					// Command module RCS specific impulse [m/s] (290 s, Encyclopedia Astronautica)
const double	CM_RCS_MAXTHRUST	= 413.68;					// Command module RCS max thrust [n] (93 lbf, Apollo CSM News Reference pg 196)

const VECTOR3	CM_RCS_PITCHPOS[4]	= {{ 0.077, 1.152, 1.182}, {-0.077, 1.152, 1.182}, { 0.111, 2.130,-0.311}, {-0.111, 2.130,-0.311}};
const VECTOR3	CM_RCS_ROLLPOS[4]	= {{-1.331, 1.460,-0.215}, {-1.160, 1.631,-0.215}, { 1.331, 1.460,-0.215}, { 1.160, 1.631,-0.215}};
const VECTOR3	CM_RCS_YAWPOS[4]	= {{-1.830, 0.411,-0.311}, {-1.830, 0.189,-0.311}, { 1.830, 0.411,-0.311}, { 1.830, 0.189,-0.311}};

const VECTOR3	CM_DOCK_POS			= { 0.00, 0.30, 1.77};		// Position docking port [m]
const VECTOR3	CM_DOCK_DIR			= { 0, 0, 1};				// Docking port approach direction
const VECTOR3	CM_DOCK_ROT			= { 0, 1, 0};				// Docking port alignment direction

const VECTOR3	CM_ATTACH_POS		= { 0.00, 0.30,-1.0};		// Position of aft attachment point [m]
const VECTOR3	CM_ATTACH_DIR		= { 0, 0,-1};				// Aft attachment point direction
const VECTOR3	CM_ATTACH_ROT		= { 0, 1, 0};				// Aft attachment point rotation

const VECTOR3	CM_CHUTE_ATTACH		= { 0.00, 0.75, 1.45};		// Parachute attachment point
const double	CM_CHUTE_DROGUEAREA	= PI * pow (2.5, 2);		// Stabilization drogue cross section [m^2] ("PI * R squared" solved for diameter of 16.5 ft, Apollo CSM News Reference pg 94)
const double	CM_CHUTE_MAINAREA	= PI * pow (12.725, 2);		// Main parachute cross section [m^2] ("PI * R squared" solved for diameter of 83.5 ft, Apollo CSM News Reference pg 95)

// UMMU & UCCGO Support
const int		CM_UMMU_CREWSIZE	= 3;						// I added an additional slot in case someone wants to reenact Apollo 13 ;)
const VECTOR3	CM_UMMU_HATCHPOS	= { 0.00, 1.7, 0.60};		// Position of EVA hatch
const VECTOR3	CM_UMMU_CREWROT		= { 0, 0, 0};				// Orientation of UMMU on EVA

// ==============================================================
// Command/Service Module class interface
// ==============================================================

class CSM: public VESSEL3 {
public:
	CSM (OBJHANDLE hVessel, int flightmodel);
	~CSM ();
	
	// Member classes
	friend class	CSM_VC;		// Virtual cockpit class
	CSM_VC			*vc;		// Virtual cockpit interface

	friend class	CSM_SYS;	// CSM subsystem control class
	CSM_SYS			*csm;		// Subsystem interface

	friend class	FDAI;		// Flight Director Attitude Indicator (FDAI) intrument subclass.
	friend class	SM;			// Service Module class		

	// Custom vessel functions
	void	DefineAnimations (void);			// Define mesh animations
	void	DefineLighting (void);				// Define local light sources
	void	DefineParticleStreams (void);		// Define exhaust and particle streams
	void	DefineSounds (void);				// Define sound effects

	double	CSM_EmptyMass (void);				// Calculate current empty mass of vessel
	bool	ActivateHatch (void);				// EVA Hatch control logic
	bool	DeployHGA (void);					// High-Gain Antenna deployment logic
	
	bool	TowerSep (void);					// Jettison Launch escape tower
	bool	AdapterSep (void);					// Seperate from Spacecraft launch adapter
	bool	ServiceModuleSep (void);			// Jettison Service Module
	bool	ProbeSep (void);					// Jettison Docking Ring / Probe
	bool	ApexSep (void);						// Jettison Fwd Heat Shield / Parachute Cover
	bool	OpenBayCover (void);				// Open SM instrument bay cover
	void	DeployDrogues (void);				// Deploy stabilzation drogues
	void	DeployChutes (void);				// Deploy landing parachutes 
	void	LaunchAbort (void);					// Activate launch abort system.	

	void	SetLaunchConfig (void);				// 
	void	SetFlightConfig (void);				// 
	void	SetReEntryConfig (void);			// 

	// Overloaded callback functions
	void	clbkSetClassCaps (FILEHANDLE cfg);										// Set the capabilities of the vessel class
	void	clbkLoadStateEx (FILEHANDLE scn, void *status);							// Read status from scenario file
	void	clbkSaveState (FILEHANDLE scn);											// Write status to scenario file
	void	clbkPostCreation ();													// Finalise vessel creation
	void	clbkPreStep (double simt, double simdt, double mjd);					// Pre-Step processes
	void	clbkPostStep (double simt, double simdt, double mjd);					// Post-Step processes
	//bool	clbkPlaybackEvent (double simt, double event_t, const char *event_type, const char *event);
	void	clbkVisualCreated (VISHANDLE vis, int refcount);						// Called after a vessel visual has been created by the renderer.
	void	clbkVisualDestroyed (VISHANDLE vis, int refcount);						// Called before a vessel visual is destroyed. 
	bool	clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	//void	clbkRenderHUD (int mode, const HUDPAINTSPEC *hps, SURFHANDLE hTex);
	//void	clbkRCSMode (int mode);
	//void	clbkADCtrlMode (DWORD mode);
	//void	clbkHUDMode (int mode);
	//void	clbkMFDMode (int mfd, int mode);
	//void	clbkNavMode (int mode, bool active);
	void	clbkDockEvent (int dock, OBJHANDLE mate);								// Docking event notification.
	//int	clbkConsumeDirectKey (char *kstate);
	int		clbkConsumeBufferedKey (DWORD key, bool down, char *kstate);			// Process keyboard inputs
	bool	clbkLoadGenericCockpit ();
	//bool	clbkLoadPanel (int id);
	//bool	clbkLoadPanel2D (int id, PANELHANDLE hPanel, DWORD viewW, DWORD viewH);
	//bool	clbkPanelMouseEvent (int id, int event, int mx, int my, void *context);
	//bool	clbkPanelRedrawEvent (int id, int event, SURFHANDLE surf, void *context);
	bool	clbkLoadVC (int id);													// Load virtual cockpit mode
	bool	clbkVCMouseEvent (int id, int event, VECTOR3 &p);						// Respond to virtual cockpit mouse events
	bool	clbkVCRedrawEvent (int id, int event, SURFHANDLE surf);					// Respond to virtual cockpit area redraw requests
	//int  clbkGeneric (int msgid, int prm, void *context);

	// Implimentation of OrbiterSound 4.0
	int		OrbiterSoundID;
	
	// Implimentation of UMmu 2.5
	UMMUCREWMANAGMENT Crew;															// UMmu class interface
	int iSelectedUmmuMember;														// Select crew member to eva
	char cUmmuHudDisplay[256];														// UMmu hud char variable
	double dHudMessageDelay;														// UMmu hud display delay
	char *SendHudMessage (void);													// UMmu hud display function

	void	UmmuSetClassCaps (FILEHANDLE cfg);										// Set the capabilities of the vessel class
	void	UmmuPostStep (double simt, double simdt, double mjd);					// Post-Step processes
	
	// The Add Mmu without scenery editor declaration see PDF doc
	char	cAddUMmuToVessel[255];
	void	AddUMmuToVessel (BOOL bStartAdding = false);

	bool	UmmuGetAirlockState ();
	bool	UmmuSelectCrewMember ();
	bool	UmmuTransferCrewMember (int slot);

	// UCGO 2.5 CLASS HANDLE FUNCTION AND VARIABLES
	UCGO	Cargo;			// Cargo class handle
	int		iSelectedCargo;	// for the selection of cargos -1 by default
	bool	UcgoGrappleCargo (int slot);
	bool	UcgoReleaseCargo (int slot);

	// Thrusters and Propellant handles
	PROPELLANT_HANDLE	ph_cm_rcs[2], ph_sm_rcs[4], ph_sps;							// Functional Propellant tanks
	PROPELLANT_HANDLE	ph_cm_O2, ph_cm_H2O, ph_cm_Waste, ph_sm_O2, ph_sm_H2;		// Represenative propellant tanks (consumables)
	
	THRUSTER_HANDLE		th_cm_rcs[12], th_sm_rcs[16], th_sps;						// Functional thrusters
	THRUSTER_HANDLE		th_vent;													// Represenative thrusters (particle streams)

	AIRFOILHANDLE		afh_vert, afh_horz;											// Aerodynamic forces

	ATTACHMENTHANDLE	ah_booster, ah_abort, ah_payload, ah_chutes;
	DOCKHANDLE			dh_cm_dock;
	OBJHANDLE			oh_LV, oh_SLA, oh_SM, oh_LM;

private:
	double	thrust_mod, dv_mod, consumables_mod;
	
	bool	ReEntryConfig, FlightConfig, LaunchConfig, AbortInProgress;
	bool	Booster_attached, LES_attached, SLA_attached, SM_attached, DockingProbe, ApexCovered, SM_BayOpen;
	bool	Drogues_Deployed, Chutes_Deployed;

	doorstate	HatchStatus, AntennaStatus, LightStatus, UmbilicalStatus;
	double			drogue_proc, chute_proc, hatch_proc, antenna_proc, light_proc, umbilical_proc;

	// Animations
	UINT						anim_droguedeploy, anim_chutedeploy, anim_Hatch;
	UINT						anim_light, anim_umbilical, anim_antenna;
	ANIMATIONCOMPONENT_HANDLE	ach_antenna;

	// Meshes
	MESHHANDLE	mh_adapter, mh_cockpit, mh_command, mh_probe, mh_apex, mh_sbaycover, mh_service, mh_tower;	
	UINT		mesh_CM, mesh_Probe, mesh_Apex;
	UINT		mesh_SM, mesh_SBayCover;		
	UINT		mesh_LES, mesh_SLA;
	UINT		mesh_Cockpit;		
	
};  // End "class CSM"
