// ==============================================================
//                 ORBITER MODULE: LM
//                  Part of the ORBITER SDK
//          Copyright (C) 2002-2004 Martin Schweiger
//                   All rights reserved
//
// LM.cpp
// Control module for LM vessel class
//
// Notes:
// This is an example for a "minimal" vessel implementation which
// only overloads the clbkSetClassCaps method to define vessel
// capabilities and otherwise uses the default VESSEL class
// behaviour.
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"

// ==============================================================
// Some vessel parameters
// ==============================================================
const double  PB_SIZE       = 3.5;             // mean radius [m]
const VECTOR3 PB_CS         = {10.5,15.0,5.8}; // x,y,z cross sections [m^2]
const VECTOR3 PB_PMI        = {2.28,2.31,0.79};// principal moments of inertia (mass-normalised) [m^2]
const VECTOR3 PB_RD         = {0.025,0.025,0.02};//{0.05,0.1,0.05};  // rotation drag coefficients

// ==============================================================
// Vessel class interface
// ==============================================================

class LM: public VESSEL3 {
public:
	LM (OBJHANDLE hVessel, int flightmodel);
	~LM ();

	void	clbkSetClassCaps (FILEHANDLE cfg);						// Set vessel class caps
	void	clbkPostStep (double simt, double simdt, double mjd);	// Post-Step processes

private:
	double birthdate, age, lifespan;
};

// ==============================================================
// Class Constructor and Destructor
// ==============================================================

LM::LM (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
	birthdate = oapiGetSimTime ();	// Birthdate = current sim time
	lifespan = 3600;				// Set default lifespan, 3600 seconds (1 hour).
}

LM::~LM ()
{
}

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void LM::clbkSetClassCaps (FILEHANDLE cfg)
{
	// physical vessel parameters
	SetSize (1.0);
	SetEmptyMass (100);
	SetPMI (PB_PMI);
	SetCrossSections (PB_CS);
	SetRotDrag (PB_RD);

	// camera parameters
	SetCameraOffset ( _V( 0.0, 0.0, 0.0));

	// associate a mesh for the visual
	AddMesh ("AAPO/CSM_ApexCover");
}

// --------------------------------------------------------------
// Post step processes
// --------------------------------------------------------------
void LM::clbkPostStep (double simt, double simdt, double mjd)
{
	OBJHANDLE self = GetHandle ();					// Get vessel handle 

	if (GroundContact ()) oapiDeleteVessel(self);	// If vessel touches the ground delete it 

	age = oapiGetSimTime () - birthdate;			// Get age, current sim time - birthdate
	if (age > lifespan) oapiDeleteVessel(self);		// If age is greater than lifespan delete vessel 

	double drag = GetDrag () / GetMass ();			// Get decceleration due to drag
	if (drag > 8 * G) oapiDeleteVessel (self);		// if decceleration is greater than 8 Gs burn up
}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new LM (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (LM*)vessel;
}
