// ==============================================================
//                 Orbiter Module: CSM_VC
//    part of the Apollo Aplications Prject for Orbiter (AAPO)
//               copyright (c) 2012 greg hlynka
//                    all rights reserved
//
// csm_cockpit.cpp
// control module for Lunar Module virtual cockpit and displays
//
// notes:
// ==============================================================

#include "CSM.h"
#include "CSM_Cockpit.h"
#include "CSM_SubSystems.h"
#include "CSM_SwitchDefs.h"

#include "AAPO_FDAI.h"

// Mesh resource files
#include "mesh_CM_VC.h"
#include "mesh_CM.h"
#include "mesh_SM.h"

// ==============================================================
// LM Cockpit class interface
// ==============================================================

// --------------------------------------------------------------
// Constructor
// --------------------------------------------------------------
CSM_VC::CSM_VC (CSM *vessel)
{
	v = vessel;
	int i = 0;

	for (i = 0; i < P1_SWITCHCOUNT; i++) P1_Switch[i] = (switchstate)0;
	for (i = 0; i < P2_SWITCHCOUNT; i++) P2_Switch[i] = (switchstate)0;
	for (i = 0; i < P3_SWITCHCOUNT; i++) P3_Switch[i] = (switchstate)0;
	for (i = 0; i < DIALCOUNT; i++) RotarySwitch[i] = (switchstate)0;

	fdai0 = new FDAI (vessel, PANEL_FDAIPOS + CSM_CM_OFFSET, MAIN_PANEL_TILT, _V(1, 0, 0));

	// Load VC display source
	//sh_displaysource	= oapiLoadTexture ("AAPO/PanelElements/CSM_instruments.dds");
}

// --------------------------------------------------------------
// Destructor
// --------------------------------------------------------------
CSM_VC::~CSM_VC ()
{
	int i = 0;

	for (i = 0; i < P1_SWITCHCOUNT; i++) delete mgt_P1switch[i];
	for (i = 0; i < P2_SWITCHCOUNT; i++) delete mgt_P2switch[i];
	for (i = 0; i < P3_SWITCHCOUNT; i++) delete mgt_P3switch[i];
	for (i = 0; i < DIALCOUNT; i++) delete mgt_Rotary[i];
}

// ==============================================================
// Public functions
// ==============================================================

// --------------------------------------------------------------
// Initialize VC
// --------------------------------------------------------------
void CSM_VC::InitVC (UINT mesh)
{
	OrbiterSoundID = v->OrbiterSoundID;

	// Animate Panel 1 Switches
	static UINT meshgroup_P1switches[P1_SWITCHCOUNT];

	for (int i = 0; i < P1_SWITCHCOUNT; i++)
	{
		meshgroup_P1switches[i] = (GRP_VC_Switch_P1_000 + i);

		mgt_P1switch[i] = new MGROUP_ROTATE (mesh, &meshgroup_P1switches[i], 1, P1_SWITCH_POS[i], _V( 1, 0, 0), (float)PI/4);
		anim_P1switch[i] = v->CreateAnimation (0.5);
		v->AddAnimationComponent (anim_P1switch[i], 0.0f, 1.0f, mgt_P1switch[i]);
	}

	// Animate Panel 2 Switches
	static UINT meshgroup_P2switches[P2_SWITCHCOUNT];

	for (int i = 0; i < P2_SWITCHCOUNT; i++)
	{
		meshgroup_P2switches[i] = (GRP_VC_Switch_P2_000 + i);

		mgt_P2switch[i] = new MGROUP_ROTATE (mesh, &meshgroup_P2switches[i], 1, P2_SWITCH_POS[i], _V( 1, 0, 0), (float)PI/4);
		anim_P2switch[i] = v->CreateAnimation (0.5);
		v->AddAnimationComponent (anim_P2switch[i], 0.0f, 1.0f, mgt_P2switch[i]);
	}

	// Animate Panel 3 Switches
	static UINT meshgroup_P3switches[P3_SWITCHCOUNT];

	for (int i = 0; i < P3_SWITCHCOUNT; i++)
	{
		meshgroup_P3switches[i] = (GRP_VC_Switch_P3_000 + i);

		mgt_P3switch[i] = new MGROUP_ROTATE (mesh, &meshgroup_P3switches[i], 1, P3_SWITCH_POS[i], _V( 1, 0, 0), (float)PI/4);
		anim_P3switch[i] = v->CreateAnimation (0.5);
		v->AddAnimationComponent (anim_P3switch[i], 0.0f, 1.0f, mgt_P3switch[i]);
	}

	// Animate rotary switches
	static UINT meshgroup_Rotary[DIALCOUNT]; 
	
	for (int i = 0; i < DIALCOUNT; i++)
	{
		meshgroup_Rotary[i] = (GRP_VC_RSwitch_000 + i);
		mgt_Rotary[i] = new MGROUP_ROTATE (mesh, &meshgroup_Rotary[i], 1, DIAL_POS[i], _V( 0, sin(MAIN_PANEL_TILT),-cos(MAIN_PANEL_TILT)), (float) 2 * PI);

		anim_Rotary[i] = v->CreateAnimation (0.0);
		v->AddAnimationComponent (anim_Rotary[i], 0.0f, 1.0f, mgt_Rotary[i]);
	}
}

// --------------------------------------------------------------
// Register VC active areas
// --------------------------------------------------------------
void CSM_VC::RegisterActiveAreas (VECTOR3 ofs)
{
	// Register MFD Buttons
	VECTOR3 MFD_buttons [11];
	VECTOR3 MFD_position = MFD0_POS + ofs;																							// Location of MFD within VC
	for (int i = 0; i < 11; i++) MFD_buttons[i] = MFD_position + RotateVector (MFD_BUTTON_POS[i], MAIN_PANEL_TILT, _V( 1, 0, 0));	// Rotate button coordinates to match panel tilt

	//oapiVCRegisterArea (AID_MFD0_LBUTTONS, _R( 0, 96, 32, 240), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN, PANEL_MAP_BACKGROUND, sh_PanelElements);
	//oapiVCSetAreaClickmode_Quadrilateral (AID_MFD0_LBUTTONS, MFD_buttons[0], MFD_buttons[1], MFD_buttons[2], MFD_buttons[3]);

	//oapiVCRegisterArea (AID_MFD0_RBUTTONS, _R( 32, 96, 64, 240), PANEL_REDRAW_USER, PANEL_MOUSE_LBDOWN, PANEL_MAP_BACKGROUND, sh_PanelElements);
	//oapiVCSetAreaClickmode_Quadrilateral (AID_MFD0_RBUTTONS, MFD_buttons[4], MFD_buttons[5], MFD_buttons[6], MFD_buttons[7]);

	oapiVCRegisterArea (AID_MFD0_PWR, _R(0,0,0,0), PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN, PANEL_MAP_NONE, NULL);
	oapiVCSetAreaClickmode_Spherical (AID_MFD0_PWR, MFD_buttons[8], 0.01);

	oapiVCRegisterArea (AID_MFD0_SEL, _R(0,0,0,0), PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN, PANEL_MAP_NONE, NULL);
	oapiVCSetAreaClickmode_Spherical (AID_MFD0_SEL, MFD_buttons[9], 0.01);

	oapiVCRegisterArea (AID_MFD0_MNU, _R(0,0,0,0), PANEL_REDRAW_NEVER, PANEL_MOUSE_LBDOWN, PANEL_MAP_NONE, NULL); 
	oapiVCSetAreaClickmode_Spherical (AID_MFD0_MNU, MFD_buttons[10], 0.01);

	// Register main panel switches
	for (int i = 0; i < P1_SWITCHCOUNT; i++)	
	{
		oapiVCRegisterArea (AID_SWITCH_P1_000 + i, PANEL_REDRAW_MOUSE, PANEL_MOUSE_DOWN);
		oapiVCSetAreaClickmode_Spherical (AID_SWITCH_P1_000 + i, P1_SWITCH_POS[i] + ofs, 0.012);
	}

	for (int i = 0; i < P2_SWITCHCOUNT; i++)	
	{
		oapiVCRegisterArea (AID_SWITCH_P2_000 + i, PANEL_REDRAW_MOUSE, PANEL_MOUSE_DOWN);
		oapiVCSetAreaClickmode_Spherical (AID_SWITCH_P2_000 + i, P2_SWITCH_POS[i] + ofs, 0.012);
	}

	for (int i = 0; i < P3_SWITCHCOUNT; i++)	
	{
		oapiVCRegisterArea (AID_SWITCH_P3_000 + i, PANEL_REDRAW_MOUSE, PANEL_MOUSE_DOWN);
		oapiVCSetAreaClickmode_Spherical (AID_SWITCH_P3_000 + i, P3_SWITCH_POS[i] + ofs, 0.012);
	}

	for (int i = 0; i < DIALCOUNT; i++)	
	{
		oapiVCRegisterArea (AID_ROTARYSWITCH_000 + i, PANEL_REDRAW_MOUSE, PANEL_MOUSE_DOWN);
		oapiVCSetAreaClickmode_Spherical (AID_ROTARYSWITCH_000 + i, DIAL_POS[i] + ofs, 0.02);
	}	
}

// --------------------------------------------------------------
// Respond to user inputs
// --------------------------------------------------------------
bool CSM_VC::MouseEvent (int id, int event, VECTOR3 &p)
{

	// Process MFD Buttons
	switch (id) 
	{
	case AID_MFD0_LBUTTONS:	// Left side
	case AID_MFD0_RBUTTONS:	// Right side
		PlayVesselWave (OrbiterSoundID, SOUND_BTN);
		double dp;								// Because MFD buttons are defined as a group we must break it down to determine which specific button was pressed.
		if (modf (p.y*23.0/4.0, &dp) < 0.75)	// by casting "dp" to an "int" we get a "rounded out" value for the click point's vertical position. This value will corrispond to 1 of the 6 buttons on each side of the MFD.
		{
			int bt = (int)dp + (id == AID_MFD0_LBUTTONS ? 0 : 6);	// Because rounding out only accounts for 6 of the 12 buttons, we add either 0 or 6 to 'dp' depending on which side of the screen the mouse is on.
			oapiProcessMFDButton (0, bt, event);					// 'bt' now returns a value between 1 and 12 corrisponding to one of the MFD's 12 buttons, OAPI does it's thing.
			return true;											// we're done
		}
		break;
	
	case AID_MFD0_PWR:		// Power
		PlayVesselWave (OrbiterSoundID, SOUND_BTN);
		oapiToggleMFD_on (0);
		return true;

	case AID_MFD0_SEL:		// Select
		PlayVesselWave (OrbiterSoundID, SOUND_BTN);
		oapiSendMFDKey (0, OAPI_KEY_F1);
		return true;

	case AID_MFD0_MNU:		// Menu
		PlayVesselWave (OrbiterSoundID, SOUND_BTN);
		oapiSendMFDKey (0, OAPI_KEY_GRAVE);
		return true;
	}

	// Process main panel switches
	for (int i = 0; i < P1_SWITCHCOUNT; i ++)
	{
		int npos;

		switch (i) 
		{
		case 02: case 03: case 04: case 26: case 27:   
			npos = 3;
			break;
		default:
			npos = 2;
			break;
		}

		if (id == AID_SWITCH_P1_000 + i) return ActivateSwitch (event, &P1_Switch[i], npos, SOUND_SWITCH, OrbiterSoundID);
	}

	for (int i = 0; i < P2_SWITCHCOUNT; i ++)
	{
		if (id == AID_SWITCH_P2_000 + i) return ActivateSwitch (event, &P2_Switch[i], 2, SOUND_SWITCH, OrbiterSoundID);
	}

	for (int i = 0; i < P3_SWITCHCOUNT; i ++)
	{
		if (id == AID_SWITCH_P3_000 + i) return ActivateSwitch (event, &P3_Switch[i], 2, SOUND_SWITCH, OrbiterSoundID);
	}

	for (int i = 0; i < DIALCOUNT; i ++)
	{
		if (id == AID_ROTARYSWITCH_000 + i) return ActivateSwitch (event, &RotarySwitch[i], 12, SOUND_ROTARY, OrbiterSoundID);
	}
	return false;
}

// --------------------------------------------------------------
// Respond to redraw requests
// --------------------------------------------------------------
bool CSM_VC::RedrawEvent (int id, int event, SURFHANDLE surf)
{
	
	// Animate switches
	for (int i = 0; i < P1_SWITCHCOUNT; i ++)
	{
		if (id == AID_SWITCH_P1_000 + i) return v->SetAnimation (anim_P1switch[i], (float)P1_Switch[i] / 2);
	}

	for (int i = 0; i < P2_SWITCHCOUNT; i ++)
	{
		if (id == AID_SWITCH_P2_000 + i) return v->SetAnimation (anim_P2switch[i], (float)P2_Switch[i] / 2);
	}

	for (int i = 0; i < P3_SWITCHCOUNT; i ++)
	{
		if (id == AID_SWITCH_P3_000 + i) return v->SetAnimation (anim_P3switch[i], (float)P3_Switch[i] / 2);
	}

	for (int i = 0; i < DIALCOUNT; i ++)
	{
		if (id == AID_ROTARYSWITCH_000 + i) return v->SetAnimation (anim_Rotary[i], (float)RotarySwitch[i] / 12);
	}
	return false;
}

// --------------------------------------------------------------
// Read status from scenario file
// --------------------------------------------------------------
bool CSM_VC::ParseScenarioLine (char *line)
{
	int i = 0;
	int state = 0;

	// Load Panel 1 switch positions
	if (!_strnicmp (line, "PANEL1_SWITCHES", 15)) 
	{
		line += 16;
		for (i = 0; (i < P1_SWITCHCOUNT) && (*line); i++) P1_Switch[i] = (switchstate)(*line++ - '0');
		return true;
	}

	// Load Panel 2 switch positions
	else if (!_strnicmp (line, "PANEL2_SWITCHES", 15)) 
	{
		line += 16;
		for (i = 0; (i < P2_SWITCHCOUNT) && (*line); i++) P2_Switch[i] = (switchstate)(*line++ - '0');
		return true;
	}

	// Load Panel 3 switch positions
	else if (!_strnicmp (line, "PANEL3_SWITCHES", 15)) 
	{
		line += 16;
		for (i = 0; (i < P3_SWITCHCOUNT) && (*line); i++) P3_Switch[i] = (switchstate)(*line++ - '0');
		return true;
	}

	// Loadrotary switch positions
	else if (!_strnicmp (line, "ROTARY_SWITCHES", 15))
	{
		line += 16;
		for (i = 0; (i < DIALCOUNT) && (*line); i++) RotarySwitch[i] = (switchstate)(*line++ - '0');
		return true;
	}

	else return false;
} // End "LM_COCKPIT::LoadState"

// --------------------------------------------------------------
// Write status to scenario file
// --------------------------------------------------------------
void CSM_VC::SaveState (FILEHANDLE scn)
{
	int i = 0;
	char switchpos[256];

	// Save Panel 1 state
	for (i = 0; i < P1_SWITCHCOUNT; i++) switchpos[i] = P1_Switch[i] + '0';
	switchpos[i] = '\0';
	oapiWriteScenario_string(scn, "PANEL1_SWITCHES", switchpos);

	// Save Panel 2 state
	for (i = 0; i < P2_SWITCHCOUNT; i++) switchpos[i] = P2_Switch[i] + '0';
	switchpos[i] = '\0';
	oapiWriteScenario_string(scn, "PANEL2_SWITCHES", switchpos);

	// Save Panel 3 state
	for (i = 0; i < P3_SWITCHCOUNT; i++) switchpos[i] = P3_Switch[i] + '0';
	switchpos[i] = '\0';
	oapiWriteScenario_string(scn, "PANEL3_SWITCHES", switchpos);

	//
	for (i = 0; i < DIALCOUNT; i++) switchpos[i] = RotarySwitch[i] + '0';
	switchpos[i] = '\0';
	oapiWriteScenario_string(scn, "ROTARY_SWITCHES", switchpos);
}

// --------------------------------------------------------------
// Post-Step processes
// --------------------------------------------------------------
void CSM_VC::PostStep (double simt, double simdt, double mjd)
{
	// Update FDAI
	switch (SWITCH_FDAI_SELECT) // 
	{
	case UP:
		fdai0->SetRefMode (0); // Set Reference frame to "ECC" (Ecliptic, AKA Orbiter Global)
		break;
	case DOWN:
		fdai0->SetRefMode (3); // Set Reference frame to "SRF" (Surface)
		break;
	case CENTER:
		fdai0->SetRefMode (2); // Set Reference frame to "ORB" (Orbital velocity, equivilent to default "Orbit" HUD mode)
		break;
	}

	switch (SWITCH_FDAI_SOURCE)
	{
	case UP:
		//fdai0->SetRefMode (4); // Set Reference frame to "ECC" (Ecliptic, AKA Orbiter Global)
		break;
	case DOWN:
		//fdai0->SetRefMode (4); // Set Reference frame to "SRF" (Surface)
		break;
	case CENTER:
		//fdai0->SetRefMode (4); // Set Reference frame to "ORB" (Orbital velocity, equivilent to default "Orbit" HUD mode)
		break;
	}

	switch (SWITCH_FDAI_MODE)
	{
	case UP:
		fdai0->SetTgtMode (2); // Set Target mode to "Direction" 
		break;
	case DOWN:
		fdai0->SetTgtMode (3); // Set Target mode to "rVel" 
		break;
	case CENTER:
		fdai0->SetTgtMode (1); // Set Target mode to "fixed"
		break;
	}

	// Set FDAI display scales
	switch (SWITCH_FDAI_SCALE)
	{
	case UP: // Fine
		fdai0->SetErrorScale (RAD * 5);
		fdai0->SetRateScale (RAD * 1); 
		break;
	case CENTER: // Medium
		fdai0->SetErrorScale (RAD * 10);
		fdai0->SetRateScale (RAD * 5);
		break;
	case DOWN: // Coarse
		fdai0->SetErrorScale (RAD * 15);
		fdai0->SetRateScale (RAD * 10);
		break;
	}

	fdai0->PostStep (simt, simdt, mjd);
}

// --------------------------------------------------------------
// Called after a vessel visual has been created by the renderer.
// --------------------------------------------------------------
void CSM_VC::VisualCreated (VISHANDLE vis, int refcount)
{
	fdai0->VisualCreated (vis, refcount);
}

// --------------------------------------------------------------
// Called before a vessel visual is destroyed.
// --------------------------------------------------------------
void CSM_VC::VisualDestroyed (VISHANDLE vis, int refcount)
{
	fdai0->VisualDestroyed (vis, refcount);
}