// ==============================================================
//                 Orbiter Module: LM_Cockpit
//    part of the Apollo Aplications Prject for Orbiter (AAPO)
//               copyright (c) 2012 greg hlynka
//                    all rights reserved
//
// lm_cockpit.cpp
// control module for Lunar Module virtual cockpit and displays
//
// notes:
// ==============================================================

#include "CSM.h"
#include "CSM_SubSystems.h"
#include "CSM_Cockpit.h"
#include "CSM_SwitchDefs.h"

#include "SM.h"

// Mesh resource files
#include "mesh_CM_VC.h"
#include "mesh_CM.h"
#include "mesh_SM.h"

// ==============================================================
// LM Cockpit class interface
// ==============================================================

// --------------------------------------------------------------
// Constructor
// --------------------------------------------------------------
CSM_SYS::CSM_SYS (CSM *vessel, CSM_VC *cockpit)
{
	v = vessel;
	vc = cockpit;

	// Propellant Management
	
	// Propulsion / Attitude Control
	
	// Earth Landing System
	ELS_state	= 0;
	ELS_count	= 0.0;

	// Enviromental Control and Life Support

	// Power Management

	// Misc variables and functions
	AbortInProgress = false;
}

// --------------------------------------------------------------
// Destructor
// --------------------------------------------------------------
CSM_SYS::~CSM_SYS ()
{

}

// ==============================================================
// Public functions
// ==============================================================

// --------------------------------------------------------------
// Initialize CSM
// --------------------------------------------------------------
void CSM_SYS::InitCSM ()
{
	int i = 0;

	// Engine Gimbal Animations
	static UINT				meshgroup_engine[2] = { GRP_SM_SPS_Nozzle, GRP_SM_SPS_NozzleExtention};
	static MGROUP_ROTATE	mgt_gimbalx (v->mesh_SM, meshgroup_engine, 2, SM_ENGINEPOS, _V( 1.0, 0.0, 0.0), (float) 12*RAD); 
	static MGROUP_ROTATE	mgt_gimbaly (v->mesh_SM, meshgroup_engine, 2, SM_ENGINEPOS, _V( 0.0,-1.0, 0.0), (float) 12*RAD); 

	anim_gimbalx	= v->CreateAnimation (0.5);
	v->AddAnimationComponent (anim_gimbalx, 0.0f, 1.0f, &mgt_gimbalx);

	anim_gimbaly	= v->CreateAnimation (0.5);
	v->AddAnimationComponent (anim_gimbaly, 0.0f, 1.0f, &mgt_gimbaly);


	// HGA Antenna Animations
	static UINT				meshgroup_HGApivot = GRP_SM_HGA_Pivot;
	static MGROUP_ROTATE	mgt_HGAyaw (v->mesh_SM, &meshgroup_HGApivot, 1, SM_HGA_PIVOT, _V( cos(SM_HGA_CANT), sin(SM_HGA_CANT), 0.0), (float) 360*RAD); 
	
	anim_HGAyaw = v->CreateAnimation (0.0);
	ach_HGAyaw = v->AddAnimationComponent (anim_HGAyaw, 0.0f, 1.0f, &mgt_HGAyaw, v->ach_antenna);

	static UINT				meshgroup_HGAntenna[5] = {GRP_SM_HGA_Antennas, GRP_SM_HGA_Body, GRP_SM_HGA_Reflectors};
	static MGROUP_ROTATE	mgt_HGApitch (v->mesh_SM, meshgroup_HGAntenna, 3, SM_HGA_PIVOT, _V( cos(SM_HGA_CANT), sin(SM_HGA_CANT), 0.0), (float) 90*RAD); 
	
	anim_HGApitch = v->CreateAnimation (1.0);
	ach_HGApitch = v->AddAnimationComponent (anim_HGApitch, 0.0f, 1.0f, &mgt_HGApitch, ach_HGAyaw);

} // End "CSM_SYS::InitLM"

// --------------------------------------------------------------
// Read status from scenario file
// --------------------------------------------------------------
bool CSM_SYS::ParseScenarioLine (char *line)
{

	return false;
} // End "CSM_SYS::LoadState"

// --------------------------------------------------------------
// Write status to scenario file
// --------------------------------------------------------------
void CSM_SYS::SaveState (FILEHANDLE scn)
{
	int i = 0;
	char cbuf[256];

} // End "CSM_SYS::SaveState"

// --------------------------------------------------------------
// Pre-Step processes
// --------------------------------------------------------------
void CSM_SYS::PreStep (double simt, double simdt, double mjd)
{		
	// Propellant Management
	PropellantPreStep (simt, simdt, mjd);

	// Propulsion / Attitude Control
	ThrusterPreStep (simt, simdt, mjd);

	// Earth Landing System
	// **Check if ELS is armed**

	// Enviromental Control and Life Support
	EnviromentPreStep (simt, simdt, mjd);

	// Power Management


} // End "CSM_SYS::PreStep"
 
// --------------------------------------------------------------
// Post-Step processes
// --------------------------------------------------------------
void CSM_SYS::PostStep (double simt, double simdt, double mjd)
{		
	// Propellant Management
	PropellantPostStep (simt, simdt, mjd);

	// Propulsion / Attitude Control
	ThrusterPostStep (simt, simdt, mjd);

	// Earth Landing System
	if (ELS_SysArmed ()) ELS_TimeStep (simt, simdt, mjd);

	// Enviromental Control and Life Support
	EnviromentPostStep (simt, simdt, mjd);

	// Power Management


} // End "CSM_SYS::PostStep"

// ==============================================================
// Private Functions
// ==============================================================

// --------------------------------------------------------------
// Propellant Management
// --------------------------------------------------------------
void	CSM_SYS::PropellantPreStep (double simt, double simdt, double mjd)
{
}

void	CSM_SYS::PropellantPostStep (double simt, double simdt, double mjd)
{
}

// --------------------------------------------------------------
// Propulsion / Attitude Control
// --------------------------------------------------------------
void CSM_SYS::ThrusterPreStep (double simt, double simdt, double mjd)
{	
	// Engine Gimbal and CG balancing
	VECTOR3 cgvec, tvec, cgpos, enginepos;
	v->GetSuperstructureCG (cgpos);														// The position of combined CG of the stage and any docked modules in vessel frame
	v->GetThrusterRef (v->th_sps, enginepos);											// The position of main engine in vessel frame

	cgvec = _V(cgpos.x - enginepos.x, cgpos.y - enginepos.y, cgpos.z - enginepos.z);	// Get the vector between them...  
	tvec = -cgvec;
	normalise (tvec);		

	pitchinput = v->GetThrusterGroupLevel (THGROUP_ATT_PITCHUP) - v->GetThrusterGroupLevel (THGROUP_ATT_PITCHDOWN);
	rollinput = v->GetThrusterGroupLevel (THGROUP_ATT_BANKLEFT) - v->GetThrusterGroupLevel (THGROUP_ATT_BANKRIGHT);
	yawinput = v->GetThrusterGroupLevel (THGROUP_ATT_YAWRIGHT) - v->GetThrusterGroupLevel (THGROUP_ATT_YAWLEFT);

	//sprintf (oapiDebugString (), "X %f, Y %f, Z %f", pitchinput, rollinput, yawinput);
}

void CSM_SYS::ThrusterPostStep (double simt, double simdt, double mjd)
{
	//GimbalSPS (simdt, pitchinput, yawinput);
	VECTOR3 thrustdir;
	thrustdir.y = pitchinput * sin (RAD * 3);
	thrustdir.x = yawinput * sin (RAD * 3);

	//sprintf (oapiDebugString (), "Pitch %f, Yaw %f", pitchinput, yawinput);

	// Restrict X and Y to valid range
	if (thrustdir.y > sin (SM_GIMBAL_RANGE)) thrustdir.y = sin (SM_GIMBAL_RANGE);  
	else if (thrustdir.y < sin (-SM_GIMBAL_RANGE)) thrustdir.y = sin (-SM_GIMBAL_RANGE);  

	if (thrustdir.x > sin (SM_GIMBAL_RANGE)) thrustdir.x = sin (SM_GIMBAL_RANGE);  
	else if (thrustdir.x < sin (-SM_GIMBAL_RANGE)) thrustdir.x = sin (-SM_GIMBAL_RANGE);  

	// Animate engine gimbal
	double pitchgimbal = 0.5 + (asin (thrustdir.y) / (2 * SM_GIMBAL_RANGE));
	double yawgimbal = 0.5 + (asin (thrustdir.x) / (2 * SM_GIMBAL_RANGE));

	//sprintf (oapiDebugString (), "X %f, Y %f", yawgimbal, pitchgimbal);

	if (gimbalx_proc != pitchgimbal)
	{
		double delta = simdt / SM_GIMBAL_RATE;
		double diff = gimbalx_proc - pitchgimbal;
	
		if (fabs (diff) < delta) gimbalx_proc = pitchgimbal;
		else if (diff < 0) gimbalx_proc += delta;
		else gimbalx_proc -= delta;
	}

	if (gimbaly_proc != yawgimbal)
	{
		double delta = simdt / SM_GIMBAL_RATE;
		double diff = gimbaly_proc - yawgimbal;
		
		if (fabs (diff) < delta) gimbaly_proc = yawgimbal;
		else if (diff < 0) gimbaly_proc += delta;
		else gimbaly_proc -= delta;
	}
	v->SetAnimation (anim_gimbalx, gimbalx_proc);
	v->SetAnimation (anim_gimbaly, gimbaly_proc);

	// Update thrust vector
	thrustdir.x = sin (SM_GIMBAL_RANGE) - (2 * SM_GIMBAL_RANGE * gimbaly_proc); 
	thrustdir.y = sin (SM_GIMBAL_RANGE) - (2 * SM_GIMBAL_RANGE * gimbalx_proc); 
	thrustdir.z = sqrt (1 - pow (thrustdir.x, 2) - pow (thrustdir.y, 2));
	v->	SetThrusterDir (v->th_sps, thrustdir);

}

void CSM_SYS::GimbalSPS (double simdt, double pitch, double yaw)
{
	double	delta = simdt / SM_GIMBAL_RATE;
	
	VECTOR3	thrustv, new_thrustv;

	// Sanity check, restrict inputs to values between -1 and 1.
	if (pitch > 1.0) pitch = 1.0;
	else if (pitch <-1.0) pitch =-1.0;

	if (yaw > 1.0) yaw = 1.0;
	else if (yaw <-1.0) yaw =-1.0;

	// Get current SPS thrust vector, and compare to desired
	v->GetThrusterDir(v->th_sps, thrustv);	
	
	double pitch_diff = pitch * sin (SM_GIMBAL_RANGE) - (asin (thrustv.y) / SM_GIMBAL_RANGE);
	double yaw_diff = yaw * sin (SM_GIMBAL_RANGE) - (asin (thrustv.x) / SM_GIMBAL_RANGE);

	// Calculate new thrust vector
	if (fabs(pitch_diff) > delta)
	{
		if (pitch_diff > 0) new_thrustv.y = thrustv.y + delta;
		else new_thrustv.y = thrustv.y - delta;
	}
	else new_thrustv.y = thrustv.y + pitch_diff; 

	if (fabs(yaw_diff) > delta)
	{
		if (yaw_diff > 0) new_thrustv.x = thrustv.x + delta;
		else new_thrustv.x = thrustv.x - delta;
	}
	else new_thrustv.x = thrustv.x + pitch_diff; 

	new_thrustv.z = sqrt (1 - pow (new_thrustv.x, 2) - pow (new_thrustv.y, 2));

	v->SetAnimation (anim_gimbalx, 0.5 + (asin (new_thrustv.y) / (2 * SM_GIMBAL_RANGE)));
	v->SetAnimation (anim_gimbaly, 0.5 + (asin (new_thrustv.x) / (2 * SM_GIMBAL_RANGE)));

	// Update thrust vector
	v->SetThrusterDir (v->th_sps, new_thrustv);

	////sprintf (oapiDebugString (), "X %f, Y %f, Z %f", thrustdir.x, thrustdir.y, thrustdir.z);
}

// --------------------------------------------------------------
// Earth Landing System
// --------------------------------------------------------------
void CSM_SYS::ELS_TimeStep (double simt, double simdt, double mjd)
{
	// See Apollo CSM news Ref page 93 for details
	switch (ELS_state)
	{
		case 0: // Standby (Orbital Flight) Config
			ELS_count = 0;
			if ((v->GetAtmPressure() > 0) && (v->GetDockStatus (v->GetDockHandle(0)) == NULL)) ELS_state ++;
			break;

		case 1: // We have entered the atmosphere
			if ((v->GetMachNumber () <= 2.0) || (v->GetAtmPressure() > 37600))// Jettison Apex cover at 24,000 ft. (37.6 KPa) 
			{
				if (v->ApexSep ()) ELS_state ++;
			}
			break;

		case 2: // Wait 1.5 seconds
			ELS_count += simdt;
			if (ELS_count >= 1.5)
			{
				ELS_count = 0;	// Reset countdown
				ELS_state ++;	// Procede to next step
			}
			break;
		case 3: // Deploy drogue chutes
			v->DeployDrogues ();
			if (v->drogue_proc >= 1.0) ELS_state ++;
			break;

		case 4: // Wait 2.5 seconds
			ELS_count += simdt;
			if (ELS_count >= 2.5)
			{
				ELS_count = 0;	// Reset countdown
				ELS_state ++;	// Procede to next step
			}
			break;

		case 5: // If airspeed is less than 175 mph (78.23 m/s) and barometric of altitude is less than 10,000 ft deploy main parachutes
			if ((v->GetAirspeed() < 78.23) && (v->GetAtmPressure() > 70000)) v->DeployChutes ();
			if (v->chute_proc >= 1.0) ELS_state ++;
			break;

		case 6: // Final Descent 
			break;

		case 99: // Splashdown!
			v->DetachChild (v->ah_chutes, 0.1);
			if (v->GroundContact () == false) ELS_state = 0;
			break;
	} // End "switch (ELS_state)"

	if (v->GroundContact ()) ELS_state = 99;

	sprintf (oapiDebugString (), "ELS Step %0.0f, Count %0.2f", (float)ELS_state, (float)ELS_count);
}

bool CSM_SYS::ELS_SysArmed ()
{
	// if (ElectricalPower) // Check electrical power **not yet implimented**
	if (vc->SWITCH_ELS_LOGIC == UP) return true; // if control switch is in "ON" position return true; 
	else
	{
		//ELS_state = 0;
		ELS_count = 0.0;
		return false;
	}
}

bool CSM_SYS::ELS_InAuto ()
{
	// if (ElectricalPower) // Check electrical power **not yet implimented**

	if (vc->SWITCH_ELS_AUTO == UP) return true; // if "AUTO" is selected return true; 
	else return false;
}

// --------------------------------------------------------------
// Enviromental Control and Life Support
// --------------------------------------------------------------
void	CSM_SYS::EnviromentPreStep (double simt, double simdt, double mjd)
{
}	

void	CSM_SYS::EnviromentPostStep (double simt, double simdt, double mjd)
{
}

// --------------------------------------------------------------
// Power Management
// --------------------------------------------------------------


// --------------------------------------------------------------
// Misc variables and functions
// --------------------------------------------------------------

