// ==============================================================
//                 ORBITER MODULE: LES
//                  Part of the ORBITER SDK
//          Copyright (C) 2002-2004 Martin Schweiger
//                   All rights reserved
//
// LES.cpp
// Control module for LES vessel class
//
// Notes:
// This is an example for a "minimal" vessel implementation which
// only overloads the clbkSetClassCaps method to define vessel
// capabilities and otherwise uses the default VESSEL class
// behaviour.
// ==============================================================

#include "LES.h"
#include "mesh_AbortTower.h"

// ==============================================================
// Some vessel parameters
// ==============================================================
const double  PB_SIZE       = 3.5;				// mean radius [m]
const VECTOR3 PB_CS         = {10.5,15.0,5.8};	// x,y,z cross sections [m^2]
const VECTOR3 PB_PMI        = {2.28,2.31,0.79};	// principal moments of inertia (mass-normalised) [m^2]
const VECTOR3 PB_RD         = {0.025,0.025,0.02};//{0.05,0.1,0.05};  // rotation drag coefficients

const VECTOR3 PB_TDP[3]     = {{0,-1.5,2},{-1,-1.5,-1.5},{1,-1.5,-1.5}}; // touchdown points [m]
//const VECTOR3 PB_COP        = {0,0,0};//{0,0,-0.1};      // centre of pressure for airfoils [m]

const VECTOR3 PB_COP        = { 0, 0, 4.51};//{0,0,-0.1};      // centre of pressure for airfoils [m]

const double  PB_VLIFT_C    = 2.0;             // chord length [m]
const double  PB_VLIFT_S    = 2.0;             // wing area [m^2]
const double  PB_VLIFT_A    = 2.5;             // wing aspect ratio
const double  PB_HLIFT_C    = 2.0;             // chord length [m]
const double  PB_HLIFT_S    = 1.5;             // wing area [m^2]
const double  PB_HLIFT_A    = 2.0;             // wing aspect ratio

// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
double LiftCoeff (double aoa)
{
	int i;
	const int nlift = 9;
	static const double AOA[nlift] = {-180*RAD,-60*RAD,-30*RAD,-1*RAD,15*RAD,20*RAD,25*RAD,60*RAD,180*RAD};
	static const double CL[nlift]  = {       0,      0,   -0.1,     0,   0.2,  0.25,   0.2,     0,      0};
	static const double SCL[nlift] = {(CL[1]-CL[0])/(AOA[1]-AOA[0]), (CL[2]-CL[1])/(AOA[2]-AOA[1]),
		                              (CL[3]-CL[2])/(AOA[3]-AOA[2]), (CL[4]-CL[3])/(AOA[4]-AOA[3]),
									  (CL[5]-CL[4])/(AOA[5]-AOA[4]), (CL[6]-CL[5])/(AOA[6]-AOA[5]),
									  (CL[7]-CL[6])/(AOA[7]-AOA[6]), (CL[8]-CL[7])/(AOA[8]-AOA[7])};
	for (i = 0; i < nlift-1 && AOA[i+1] < aoa; i++);
	return CL[i] + (aoa-AOA[i])*SCL[i];
}

// ==============================================================
// Shuttle-PB class interface
// ==============================================================

LES::LES (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
	birthdate = oapiGetSimTime ();	// Birthdate = current sim time
	lifespan = 3600;				// Set default lifespan, 3600 seconds (1 hour).

	if (Canards) canard_proc = 1.0;
	else canard_proc = 0.0;

	AbortInProgress = false;
	abort_count = 0.0;
}

LES::~LES ()
{
}

// ==============================================================
// Airfoil lift/drag functions
// ==============================================================

void LES::vlift (VESSEL *v, double aoa, double M, double Re,
	void *context, double *cl, double *cm, double *cd)
{
	static const double clp[] = {  // lift coefficient from -pi to pi in 10deg steps
		-0.1,-0.5,-0.4,-0.1,0,0,0,0,0,0,0,0,0,0,-0.2,-0.6,-0.6,-0.4,0.2,0.5,0.9,0.8,0.2,0,0,0,0,0,0,0,0,0,0.1,0.4,0.5,0.3,-0.1,-0.5
	};
	static const double aoa_step = 10.0*RAD;
	double a, fidx, saoa = sin(aoa);
	a = modf((aoa+PI)/aoa_step, &fidx);
	int idx = (int)(fidx+0.5);
	*cl = clp[idx]*(1.0-a) + clp[idx+1]*a;     // linear interpolation
	*cm = 0.0; //-0.03*sin(aoa-0.1);
	*cd = 0.03 + 0.4*saoa*saoa;                // profile drag
	*cd += oapiGetInducedDrag (*cl, 1.0, 0.5); // induced drag
	*cd += oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);  // wave drag
}

void LES::hlift (VESSEL *v, double aoa, double M, double Re,
	void *context, double *cl, double *cm, double *cd)
{
	static const double clp[] = {  // lift coefficient from -pi to pi in 45deg steps
		0,0.4,0,-0.4,0,0.4,0,-0.4,0,0.4
	};
	static const double aoa_step = 45.0*RAD;
	double a, fidx;
	a = modf((aoa+PI)/aoa_step, &fidx);
	int idx = (int)(fidx+0.5);
	*cl = clp[idx]*(1.0-a) + clp[idx+1]*a;     // linear interpolation
	*cm = 0.0;
	*cd = 0.03;
	*cd += oapiGetInducedDrag (*cl, 1.5, 0.6); // induced drag
	*cd += oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);  // wave drag
}

// ==============================================================
// Overloaded callback functions
// ==============================================================

bool LES::DetachSpacecraft (void)
{	
	SetThrusterGroupLevel (THGROUP_HOVER, 1.0);
	return Undock (0, 0);
}

bool LES::DeployCanards (void)
{
	if (Canards == false)
	{
		CanardStatus = OPENING;
		CreateControlSurface (AIRCTRL_ELEVATOR, 1.33, 0.7, _V( 0, 0, 5.51), AIRCTRL_AXIS_XNEG);
		Canards = true;
		return true;
	}
	else return false;
}

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void LES::clbkSetClassCaps (FILEHANDLE cfg)
{
	double PropMass = LES_FUELMASS + LES_JETT_FUELMASS + LES_PITCH_FUELMASS;

	// physical vessel parameters
	SetSize (LES_SIZE);
	SetEmptyMass (LES_MASS - PropMass);
	SetPMI (LES_PMI);
	SetCrossSections (LES_CS);
	SetRotDrag (LES_RD);
	SetCW (LES_CW[0], LES_CW[1], LES_CW[2], LES_CW[3]);
	SetTouchdownPoints (LES_TDP[0], LES_TDP[1], LES_TDP[2]);
	SetISP (LES_ISP);
	//LES::SetAlbedoRGB
	//LES::SetPitchMomentScale
	//LES::SetYawMomentScale
	
	// airfoil definitions
	CreateAirfoil (LIFT_VERTICAL, LES_COP, &VLiftCoeff, 3.5, LES_CS.z, 1.0);
	CreateAirfoil (LIFT_HORIZONTAL, _V(0.0, 0.0, 0.0), &HLiftCoeff, 3.5, LES_CS.z, 1.0);	

	// docking port definition
	SetDockParams (LES_DOCK_POS, LES_DOCK_DIR, LES_DOCK_ROT);

	//CreateAirfoil3 (LIFT_VERTICAL,   PB_COP, vlift, NULL, PB_VLIFT_C, PB_VLIFT_S, PB_VLIFT_A);
	//CreateAirfoil3 (LIFT_HORIZONTAL, PB_COP, hlift, NULL, PB_HLIFT_C, PB_HLIFT_S, PB_HLIFT_A);

	// propellant resources
	ph_main		= CreatePropellantResource (LES_FUELMASS);
	ph_jett		= CreatePropellantResource (LES_JETT_FUELMASS);
	ph_pitch	= CreatePropellantResource (LES_PITCH_FUELMASS);

	SetDefaultPropellantResource (ph_main);
	
	double alpha, thrust;


	// main engine
	alpha = LES_NOZZLE_ANGLE;
	thrust = LES_MAXTHRUST / cos(alpha);
	
	th_main[0]	= CreateThruster (LES_ENGINEPOS, _V( 0,-sin(alpha), cos(alpha)), thrust * 0.8, ph_main);	// 12 o'clock
	th_main[1]	= CreateThruster (LES_ENGINEPOS, _V( sin(alpha), 0, cos(alpha)), thrust * 1.0, ph_main);	// 3 o'clock
	th_main[2]	= CreateThruster (LES_ENGINEPOS, _V( 0, sin(alpha), cos(alpha)), thrust * 1.2, ph_main);	// 6 o'clock
	th_main[3]	= CreateThruster (LES_ENGINEPOS, _V(-sin(alpha), 0, cos(alpha)), thrust * 1.0, ph_main);	// 9 o'clock
	CreateThrusterGroup (th_main, 4, THGROUP_MAIN);

	// Jettison motor
	alpha = LES_JETTISON_ANGLE;
	thrust = LES_JETT_MAXTHRUST / cos(alpha);

	th_jett[0] = CreateThruster (LES_JETT_ENGINEPOS, _V( 0, sin(alpha), cos(alpha)), thrust, ph_jett);
	th_jett[1] = CreateThruster (LES_JETT_ENGINEPOS, _V( 0,-sin(alpha), cos(alpha)), thrust, ph_jett);
	CreateThrusterGroup (th_jett, 2, THGROUP_HOVER);

	// Pitch motor
	th_pitch = CreateThruster (LES_PITCH_ENGINEPOS, _V( 0,-1, 0), LES_PITCH_MAXTHRUST, ph_pitch);

	// camera parameters
	SetCameraOffset ( _V( 0.0, 0.8, 0.0));

	// associate a mesh for the visual
	AddMesh ("AAPO/AbortTower");
}

// --------------------------------------------------------------
// Finalise vessel creation
// --------------------------------------------------------------
void LES::clbkPostCreation ()
{
	// Animations
	static UINT				meshgroup_port = GRP_LES_Body_CanardPort;
	static UINT				meshgroup_stbd = GRP_LES_Body_CanardStbd;

	static MGROUP_ROTATE	mgt_portcanard (0, &meshgroup_port, 1, _V(-0.3250, 0.0000, 5.3633), _V( 0.0, 0.0, 1.0), (float) 135*RAD); 
	static MGROUP_ROTATE	mgt_stbdcanard (0, &meshgroup_stbd, 1, _V( 0.3250, 0.0000, 5.3633), _V( 0.0, 0.0, 1.0), (float)-135*RAD); 

	anim_canards = CreateAnimation (0.0);
	AddAnimationComponent (anim_canards, 0.0f, 1.0f, &mgt_portcanard);	
	AddAnimationComponent (anim_canards, 0.0f, 1.0f, &mgt_stbdcanard);
	SetAnimation (anim_canards, canard_proc);


	// Exhaust streams
	PARTICLESTREAMSPEC contrail_main =
	{
		0, 5.0, 16, 200, 0.15, 1.0, 5, 3.0, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 2,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};

	PARTICLESTREAMSPEC exhaust_main =
	{
		0, 2.0, 20, 200, 0.05, 0.1, 8, 1.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-5, 0.1
	};

	SURFHANDLE tex = oapiRegisterExhaustTexture ("AAPO/Exhaust2");

	for (int i = 0; i < 4; i ++)
	{
		double ofs = 1.5;
		VECTOR3 pos, dir; 
		GetThrusterRef (th_main[i], pos);
		GetThrusterDir (th_main[i], dir);
		AddExhaustStream (th_main[i], pos + (dir * -ofs), &contrail_main);
		AddExhaustStream (th_main[i], pos + (dir * -ofs), &exhaust_main);
		AddExhaust (th_main[i], 8, 1, ofs, tex);
	}

	PARTICLESTREAMSPEC contrail_hover =
	{
		0, 5.0, 8, 200, 0.15, 1.0, 5, 3.0, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 2,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};
	PARTICLESTREAMSPEC exhaust_hover = {
		0, 2.0, 10, 200, 0.05, 0.05, 8, 1.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-5, 0.1
	};

	for (int i = 0; i < 2; i ++)
	{
		double ofs = 0.75;
		VECTOR3 pos, dir; 
		GetThrusterRef (th_jett[i], pos);
		GetThrusterDir (th_jett[i], dir);
		AddExhaustStream (th_jett[i], pos + (dir * -ofs), &contrail_hover);
		AddExhaustStream (th_jett[i], pos + (dir * -ofs), &exhaust_hover);
		AddExhaust (th_jett[i], 3, 0.1, ofs, tex);
	}

	AddExhaust (th_pitch, 3, 0.2, 0.3, tex);

}

// --------------------------------------------------------------
// Post-Step processes and Animations
// --------------------------------------------------------------
void LES::clbkPostStep (double simt, double simdt, double mjd)
{
	// Thruster Management
	if (GetThrusterGroupLevel (THGROUP_MAIN) > 0.0) AbortInProgress = true;
	if (GetThrusterGroupLevel (THGROUP_HOVER) > 0.0) SetThrusterLevel (th_pitch, 1.0);

	if (AbortInProgress)
	{
		SetThrusterGroupLevel (THGROUP_MAIN, 1.0);
		SetThrusterLevel (th_pitch, 1.0);
		abort_count += simdt;
	}

	if (abort_count > 10)
	{
		DeployCanards ();	
		SetControlSurfaceLevel  (AIRCTRL_ELEVATOR, -1);
	}

	//// Automatically Detach from spacecraft
	//if (abort_count > 14)
	//{
	//	// Detach CM if barometric altitude is less than 10,000 ft 
	//	if  (GetAtmPressure() > 70000) DetachSpacecraft ();

	//	// Detach CM if heatshield is oriented in direction of flight
	//	if (fabs(GetAOA ()) > 5 * PI / 6) DetachSpacecraft ();
	//}
	////sprintf (oapiDebugString (), "Abort timer: %0.2f, AOA %0.2f", abort_count, GetAOA () * DEG);

	// Animate Canards
	if (CanardStatus > OPEN)	
	{
		double	delta = simdt / 0.3;				// delta = simdt divided by duration of animation in seconds

		if (CanardStatus == OPENING)
		{
			if (canard_proc < 1) canard_proc += delta;	// If process value is less than 1 add delta to process value
			else CanardStatus = OPEN;					// If process value is greater than 1 GearStatus = OPEN
		}
		
		if (CanardStatus == CLOSING)
		{
			if (canard_proc > 0) canard_proc -= delta;	// If process value is Greater than 0 subtract delta from process value
			else CanardStatus = CLOSED;					// If process value is less than 0 GearStatus = CLOSED
		}

		SetAnimation (anim_canards, canard_proc);
	}

	// Clean up
	OBJHANDLE self = GetHandle ();					// Get vessel handle 
	if (GroundContact ()) oapiDeleteVessel(self);	// If LES touches the ground delete it 

	if (GetFuelMass () <= 0.0)
	{
		age = oapiGetSimTime () - birthdate;			// age = current simtime - birthdate.
		if (age > lifespan) oapiDeleteVessel(self);		// If age is greater than lifespan delete LES
	}
	
	double drag = GetDrag () / GetMass ();			// Get decceleration due to drag
	if (drag > 16 * G) oapiDeleteVessel (self);		// if decceleration is greater than 16 Gs burn up
}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new LES (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (LES*)vessel;
}
