// ==============================================================
//                    ORBITER MODULE: CSM
//                  Part of the ORBITER SDK
//               Copyright (C) 2012 Greg Hlynka
//                    All rights reserved
//
// CSM.h
// Interface for CSM vessel class
//
// Notes: This file exists
// ==============================================================

#pragma once
#define STRICT
#define ORBITER_MODULE

// Orbiter SDK files
#include "orbitersdk.h"
#include "AAPO_Common.h"

// Mesh resource files
#include "mesh_AbortTower.h"

// ==============================================================
// Vessel constants
// ==============================================================
const double	LES_SIZE			= {5.03};					// Vessel's mean radius [m]
const VECTOR3	LES_CS				= { 12.7, 12.6, 12.4};		// x,y,z cross sections [m^2]
const VECTOR3	LES_PMI				= { 12.1, 12.1, 0.8};		// Principal moments of inertia (mass-normalised) [m^2]
const VECTOR3	LES_RD				= { 0.4, 0.4, 0.05};			// Rotation drag coefficients without service module
const VECTOR3	LES_COP				= { 0.00, 0.00,-5.04};		// LES's center of pressure / lift relative to CoG

const VECTOR3	LES_COM				= { 0.0,-0.02,-0.51};		// Center of mass
const double	LES_CW[4]			= { 0.1, 0.3, 1.4, 1.4};	// Wind resistance coefficients (forward, aft, lateral, vertical)
const VECTOR3	LES_TDP[3]			= {{ 0.00,-3.30,-6.25}, {-2.86, 4.25,-6.25}, { 2.86, 4.25,-6.25}};	// touchdown points [m]

const double	LES_MASS			= 3628.74;					// Total mass [kg] (8000 lbs)
const double	LES_FUELMASS		= 1421.26;					// fuel mass [kg]
const double	LES_JETT_FUELMASS	= 92.98;					// jettison motor fuel mass [kg] (205 lbs)
const double	LES_PITCH_FUELMASS	= 4.04;						// pitch motor fuel mass [kg] (8.9 lbs)
const double	LES_ISP				= 285.6 * G;				// specific impulse [m/s] (285.6 sec estimated based on Ammonia-Perchlorate Solid Rocket performance)
const double	LES_MAXTHRUST		= 653888.58 / 4;			// max thrust [n] (147,000 lbf, Apollo CSM News Reference pg 142)
const double	LES_JETT_MAXTHRUST	= 140118.98 / 2;			// max thrust [n] (31,500 lbf, Apollo CSM News Reference pg 142)
const double	LES_PITCH_MAXTHRUST	= 17792.89;					// max thrust [n] (31,500 lbf, Apollo CSM News Reference pg 142)
const VECTOR3	LES_ENGINEPOS		= { 0.00, 0.00,-0.19};		// main engine location
const VECTOR3	LES_JETT_ENGINEPOS	= { 0.00, 0.00, 4.62};		// jettison motor location
const VECTOR3	LES_PITCH_ENGINEPOS	= { 0.00, 0.00, 5.21};		// pitch motor location

const double	LES_NOZZLE_ANGLE	= 35 * RAD;
const double	LES_JETTISON_ANGLE	= 30 * RAD;

const VECTOR3	LES_DOCK_POS		= { 0.00, 0.00,-2.98};		// docking port location [m]
const VECTOR3	LES_DOCK_DIR		= { 0, 0,-1};				// docking port approach direction
const VECTOR3	LES_DOCK_ROT		= { 0, 1, 0};				// docking port alignment direction

// ==============================================================
// Vessel class interface
// ==============================================================

class LES: public VESSEL3 {
public:
	LES (OBJHANDLE hVessel, int flightmodel);
	~LES ();

	//
	bool	DeployCanards ();
	bool	DetachSpacecraft ();

	//
	void	clbkSetClassCaps (FILEHANDLE cfg);
	void	clbkPostCreation ();
	void	clbkPostStep (double simt, double simdt, double mjd);					// Post-Step processes

	// Implimentation of OrbiterSound 4.0
	int		OrbiterSoundID;

	PROPELLANT_HANDLE	ph_main, ph_jett, ph_pitch; 
	THRUSTER_HANDLE		th_main[4], th_jett[2], th_pitch; 

private:
	double	birthdate, age, lifespan, abort_count;
	bool	AbortInProgress, Canards;

	doorstate	CanardStatus;

	UINT	anim_canards;
	double	canard_proc;

	static void vlift (VESSEL *v, double aoa, double M, double Re,
		void *context, double *cl, double *cm, double *cd);
	static void hlift (VESSEL *v, double aoa, double M, double Re,
		void *context, double *cl, double *cm, double *cd);
};
