// ==============================================================
//                 ORBITER MODULE: LM
//                  Part of the ORBITER SDK
//          Copyright (C) 2002-2004 Martin Schweiger
//                   All rights reserved
//
// LM.cpp
// Control module for LM vessel class
//
// Notes:
// This is an example for a "minimal" vessel implementation which
// only overloads the clbkSetClassCaps method to define vessel
// capabilities and otherwise uses the default VESSEL class
// behaviour.
// ==============================================================

#include "LM.h"
#include "LM_DescentStage.h"
#include "LM_Cockpit.h"
#include "LM_Systems.h"

// ==============================================================
// Lunar Module class interface
// ==============================================================

// --------------------------------------------------------------
// Constructor
// --------------------------------------------------------------
LM::LM (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
	// Config file difficulty modifiers (1.0 = 100% of default value)
	thrust_mod		= 1.0;
	dv_mod			= 1.0;
	consumables_mod = 1.0;

	// Set initial vessel state
	DescentStage_attached	= true;
	Booster_attached		= false;
	AscentConfig			= false;

	UMmuCrewCount	= 0;
	CrewPos			= 0;

	HatchStatus		= CLOSED;
	hatch_proc		= 0.0;

	GearStatus		= OPEN;
	gear_proc		= 1.0;	
	
	radar_proc[0]	= 0.0;
	radar_proc[1]	= 0.5;
	sband_proc[0]	= 0.0;
	sband_proc[1]	= 0.5;

	// Load sub classes
	vc				= new LM_COCKPIT (this);	// Cockpit control class
	lm				= new LM_SYSTEMS (this, vc);	// LM sub-system control class

//	// Load meshes
	//mh_Commander
	//mh_Pilot
	mh_descent		= oapiLoadMeshGlobal("AAPO/LM_DescentStage");
	mh_ascent		= oapiLoadMeshGlobal("AAPO/LM_AscentStage");
	mh_cockpit		= oapiLoadMeshGlobal("AAPO/LM_VC");
}

// --------------------------------------------------------------
// Destructor
// --------------------------------------------------------------
LM::~LM ()
{
	// Delete member classes
	delete	vc;
	delete	lm;

	// Delete animations
	for (int i = 0; i < 4; i++)
	{
		delete mgt_Leg[i]; 
		delete mgt_Strut[i];
		delete mgt_Downlock[i];
	}

	// Delete Local light sources

	// Delete particle stream specs
}

// ==============================================================
// Custom Vessel Functions
// ==============================================================

// --------------------------------------------------------------
// Define mesh animations
// --------------------------------------------------------------
void LM::DefineAnimations (void)
{
	// Component Handles
	ANIMATIONCOMPONENT_HANDLE	ach_Hatch, ach_HatchHandle, ach_SBandPitch, ach_SBandYaw, ach_RadarPitch, ach_RadarYaw;
	ANIMATIONCOMPONENT_HANDLE	ach_GearLeg[4], ach_GearStrut[4], ach_GearLock[4];
	ANIMATIONCOMPONENT_HANDLE	ach_HatchVC, ach_HatchHandleVC;

	// EVA Hatch (Exterior)
	static UINT	meshgroup_Hatch = AS_GRP_EvaHatch;
	static UINT	meshgroup_Handle = AS_GRP_HatchHandle;

	static MGROUP_ROTATE	mgt_Hatch (mesh_Ascent, &meshgroup_Hatch, 1, _V( 0.394,-0.578, 1.661), _V( 0.0, 1.0, 0.0), (float)-85*RAD); 
	static MGROUP_ROTATE	mgt_HatchHandle (mesh_Ascent, &meshgroup_Handle, 1, _V(-0.40279,-0.55598, 1.69760), _V( 0.0, 0.0, 1.0), (float) 90*RAD);

	anim_Hatch = CreateAnimation (0.0);
	ach_Hatch = AddAnimationComponent (anim_Hatch, 0.3f, 1.0f, &mgt_Hatch);
	ach_HatchHandle = AddAnimationComponent (anim_Hatch, 0.0f, 0.2f, &mgt_HatchHandle, ach_Hatch);

	// EVA Hatch (VC)
	static UINT	meshgroup_HatchVC	= VC_GRP_EvaHatch;
	static UINT	meshgroup_HandleVC	= VC_GRP_HatchHandle;

	static MGROUP_ROTATE	mgt_HatchVC (mesh_Cockpit, &meshgroup_HatchVC, 1, _V( 0.394,-0.578, 1.661), _V( 0.0, 1.0, 0.0), (float)-85*RAD);
	static MGROUP_ROTATE	mgt_HatchHandleVC (mesh_Cockpit, &meshgroup_HandleVC, 1, _V(-0.40279,-0.55598, 1.69760), _V( 0.0, 0.0, 1.0), (float) 90*RAD);

	ach_HatchVC			= AddAnimationComponent (anim_Hatch, 0.3f, 1.0f, &mgt_HatchVC);
	ach_HatchHandleVC	= AddAnimationComponent (anim_Hatch, 0.0f, 0.2f, &mgt_HatchHandleVC, ach_HatchVC);

	// Steerable S-Band Antenna
	static UINT meshgroup_SBandPivot		= AS_GRP_SBandPivot;
	static UINT meshgroup_SBandAntenna[2]	= {AS_GRP_SBandAntenna, AS_GRP_SBandDish};

	static MGROUP_ROTATE	mgt_SBand_pivot (mesh_Ascent, &meshgroup_SBandPivot, 1, LM_SBAND_PIVOT, LM_SBAND_AXIS, (float)LM_SBAND_PITCHRANGE);
	static MGROUP_ROTATE	mgt_SBand_Antenna (mesh_Ascent, meshgroup_SBandAntenna, 2, LM_SBAND_PIVOT, _V( 0, 0, 1), (float)LM_SBAND_YAWRANGE);
	
	anim_SBandPitch		= CreateAnimation (0.0);
	anim_SBandYaw		= CreateAnimation (0.5);
	ach_SBandPitch		= AddAnimationComponent (anim_SBandPitch, 0.0f, 1.0f, &mgt_SBand_pivot);
	ach_SBandYaw		= AddAnimationComponent (anim_SBandYaw, 0.0f, 1.0f, &mgt_SBand_Antenna, ach_SBandPitch);

	// Radar Antenna
	static UINT meshgroup_RadarPivot		= AS_GRP_RadarPivot;
	static UINT meshgroup_RadarAntenna[2]	= {AS_GRP_RadarAntenna, AS_GRP_RadarDish};

	static MGROUP_ROTATE	mgt_Radar_pivot (mesh_Ascent, &meshgroup_RadarPivot, 1, LM_RADAR_PIVOT, _V(-1, 0, 0), (float)LM_RADAR_PITCHRANGE);
	static MGROUP_ROTATE	mgt_Radar_Antenna (mesh_Ascent, meshgroup_RadarAntenna, 2, LM_RADAR_PIVOT, _V( 0, 1, 0), (float)LM_RADAR_YAWRANGE);

	anim_RadarPitch		= CreateAnimation (0.0);
	anim_RadarYaw		= CreateAnimation (0.5);
	ach_RadarPitch		= AddAnimationComponent (anim_RadarPitch, 0.0f, 1.0f, &mgt_Radar_pivot);
	ach_RadarYaw		= AddAnimationComponent (anim_RadarYaw, 0.0f, 1.0f, &mgt_Radar_Antenna, ach_RadarPitch);

	// Landing Gear
	static UINT meshgroup_Legs[4][3] = {
		{DS_GRP_LandingFoot_FWD,	DS_GRP_ShockStrut_FWD,	DS_GRP_PrimaryStrut_FWD}, 
		{DS_GRP_LandingFoot_AFT,	DS_GRP_ShockStrut_AFT,	DS_GRP_PrimaryStrut_AFT},
		{DS_GRP_LandingFoot_PORT,	DS_GRP_ShockStrut_PORT,	DS_GRP_PrimaryStrut_PORT}, 
		{DS_GRP_LandingFoot_STBD,	DS_GRP_ShockStrut_STBD,	DS_GRP_PrimaryStrut_STBD}};
	static UINT meshgroup_Struts[4] = { DS_GRP_SecondaryStruts_FWD, DS_GRP_SecondaryStruts_AFT, DS_GRP_SecondaryStruts_PORT, DS_GRP_SecondaryStruts_STBD};
	static UINT meshgroup_Locks[4] = { DS_GRP_Downlock_FWD, DS_GRP_Downlock_AFT, DS_GRP_Downlock_PORT, DS_GRP_Downlock_STBD};
	static UINT meshgroup_Ladder = DS_GRP_Ladder;

	anim_Gear = CreateAnimation (1.0);

	for (int i = 0; i < 4; i++)
	{
		// Animation components
		mgt_Leg[i]		= new MGROUP_ROTATE (mesh_Descent, &meshgroup_Legs[i][0], 3, DES_LEG_PIVOT[i], DES_LEG_AXIS[i], (float) 45*RAD);		// Animate landing legs
		mgt_Strut[i]	= new MGROUP_ROTATE (mesh_Descent, &meshgroup_Struts[i], 1, DES_STRUT_PIVOT[i], DES_LEG_AXIS[i], (float)-63*RAD);		// Animate Support Struts attatched to legs
		mgt_Downlock[i]	= new MGROUP_ROTATE (mesh_Descent, &meshgroup_Locks[i], 1, DES_LOCK_PIVOT[i], DES_LEG_AXIS[i], (float) 150*RAD);	// Animate Locking mechanism joining support struts to body

		// Add individual components to 'anim_Gear' 
		ach_GearLeg[i]		= AddAnimationComponent (anim_Gear, 0.0, 1.0, mgt_Leg[i]);
		ach_GearStrut[i]	= AddAnimationComponent (anim_Gear, 0.0, 1.0, mgt_Strut[i], ach_GearLeg[i]);
		ach_GearLock[i]		= AddAnimationComponent (anim_Gear, 0.0, 1.0, mgt_Downlock[i], ach_GearStrut[i]);
	}

	static MGROUP_ROTATE mgt_Ladder (mesh_Descent, &meshgroup_Ladder, 1, DES_LEG_PIVOT[0], DES_LEG_AXIS[0], (float) 45*RAD); // Apply front leg's animation state to ladder.
	AddAnimationComponent (anim_Gear, 0.0, 1, &mgt_Ladder);

	// Apply initial process values
	SetAnimation (anim_Hatch, hatch_proc);				
	SetAnimation (anim_Gear, gear_proc);
	SetAnimation (anim_RadarPitch, radar_proc[0]); SetAnimation (anim_RadarYaw, radar_proc[1]);
	SetAnimation (anim_SBandPitch, sband_proc[0]); SetAnimation (anim_SBandYaw, sband_proc[1]);

} // End "LM::DefineAnimations"

// --------------------------------------------------------------
// Define local light sources
// --------------------------------------------------------------
void LM::DefineLighting (void)
{
	// Tracking Light (strobe)
	static VECTOR3	STROBE_POS		= { 0.00, 1.84, 2.28};
	static VECTOR3	STROBE_COL		= { 0.9, 1.0, 1.0}; 

	bspec_strobe.shape = BEACONSHAPE_STAR;
	bspec_strobe.pos = &STROBE_POS;
	bspec_strobe.col = &STROBE_COL;
	bspec_strobe.size = 0.5;
	bspec_strobe.falloff = 0.2;
	bspec_strobe.period = 1.0;
	bspec_strobe.duration = 0.02;
	bspec_strobe.tofs = 0.0;
	bspec_strobe.active = 1;
	AddBeacon (&bspec_strobe);

	// Position Lights (Beacons)
	static VECTOR3 BEACON_POS[5]	= {{-2.52, 1.04, 0.20}, { 1.91, 0.74, 0.22}, {-0.28, 1.92, 2.27}, { 0.28, 1.92, 2.27}, { 0.00, 2.3,-1.81}};	
	static VECTOR3 BEACON_COL[5]	= {{ 1.0, 0.5, 0.5}, { 0.5, 1.0, 0.5}, { 1.0, 1.0, 0.5}, { 1.0, 1.0, 0.5}, { 1.0, 1.0, 1.0}};					// Red, Green, Yellow x2, and White

	for (int i = 0; i < 5; i++)
	{
		bspec_pos[i].shape = BEACONSHAPE_DIFFUSE;
		bspec_pos[i].pos = BEACON_POS+i;
		bspec_pos[i].col = BEACON_COL+i;
		bspec_pos[i].size = 0.1;
		bspec_pos[i].falloff = 0.4;
		bspec_pos[i].period = 0;
		bspec_pos[i].duration = 1;
		bspec_pos[i].tofs = 1;
		bspec_pos[i].active = 1;
		AddBeacon (&bspec_pos[i]);
	}

	// Flood Lights

} // End "LM::DefineLighting"

// --------------------------------------------------------------
// Define exhaust and particle streams
// --------------------------------------------------------------
void LM::DefineParticleStreams (void)
{
	// Main propulsion exhaust
	SURFHANDLE tex_mps = oapiRegisterExhaustTexture ("Exhaust2");								// Load Main Engine exhaust texture
	AddExhaust (th_ascent, 3.0*thrust_mod, 0.4*thrust_mod, (0.2 + thrust_mod/2), tex_mps);		// Ascent engine exhaust [length*thrust_mod, width*thrust_mod, offset]
	AddExhaust (th_descent, 4.0*thrust_mod, 0.8*thrust_mod, (0.9 + thrust_mod/2), tex_mps);		// Descent engine exhaust [length*thrust_mod, width*thrust_mod, offset]
	
	// RCS exhaust
	SURFHANDLE tex_rcs = oapiRegisterExhaustTexture ("Exhaust_atrcs");	// Load RCS exhaust texture

	for (int i = 0; i < 8; i++)
	{
		AddExhaust(th_rcs_a[i], 2.0, 0.3, 0.35, tex_rcs); // RCS "A-loop" thrusters exhaust [length, width, offset]
		AddExhaust(th_rcs_b[i], 2.0, 0.3, 0.35, tex_rcs); // RCS "B-loop" thrusters exhaust [length, width, offset]
	}

} // End "LM::DefineParticleStreams"

// --------------------------------------------------------------
// Define sound effects
// --------------------------------------------------------------
void LM::DefineSounds (void)
{
	// OrbiterSound 4.0
	OrbiterSoundID = ConnectToOrbiterSoundDLL (GetHandle());			// Init sound
	SoundOptionOnOff (OrbiterSoundID, PLAYCOUNTDOWNWHENTAKEOFF, FALSE);	// Disable the stock countdown... it's annoying :)

	LoadSoundFiles (OrbiterSoundID);

} // End "LM::DefineSounds"

// --------------------------------------------------------------
// Calculate current empty mass of vessel
// --------------------------------------------------------------
double LM::LM_EmptyMass (void)
{
	// Default empty mass is equal to ascent stage's empty mass
	double output = ASC_EMPTYMASS;								

	// If descent stage is attached add it's mass to total
	if (DescentStage_attached) output += DES_EMPTYMASS;	

	// Add mass of crew	to total
	int nsouls = Crew.GetCrewTotalNumber ();
	for (int i = 0; i < nsouls; i ++) output += Crew.GetCrewWeightBySlotNumber (i);

	// Add mass of cargo to total
	//**not implimented**

	return output;	// Return total
} // End "LM::LM_EmptyMass"

// --------------------------------------------------------------
// EVA Hatch control logic
// --------------------------------------------------------------
bool LM::ActivateHatch (void)
{
	// Sanity check.
	if (hatch_proc < 0)			// If process value is less than 0...
	{
		HatchStatus = CLOSED;	// ...set status to "CLOSED"
		hatch_proc = 0;			// and process value to 0
	}
	else if (hatch_proc > 1)	// If process value is greater than 1...
	{
		HatchStatus = OPEN;		// ...set status to "OPEN"
		hatch_proc = 1;			// and process value to 1
	}

	// Close hatch
	if (HatchStatus != CLOSED) 
	{
		HatchStatus = CLOSING;
		return true;
	}	
	else
	{
		// check pressure difference
		HatchStatus = OPENING;
		return true;
	}

	SetAnimation (anim_Hatch, hatch_proc);			// Update animation state
	return false;
} // End "LM::ActivateHatch"

// --------------------------------------------------------------
// Landing Gear control logic
// --------------------------------------------------------------
bool LM::ActivateGear (void)
{
	// Sanity check.
	if (gear_proc < 0)			// If process value is less than 0...
	{
		GearStatus = CLOSED;	// ...set status to "CLOSED"
		gear_proc = 0;			// and process value to 0
	}
	else if (gear_proc > 1)		// If process value is greater than 1...
	{
		GearStatus = OPEN;		// ...set status to "OPEN"
		gear_proc = 1;			// and process value to 1
	}
	
	// Deploy landing gear
	if ((GearStatus != OPEN) && (Booster_attached == false))
	{
		// check to see if in launch configuration
		GearStatus = OPENING;
		PlayVesselWave (OrbiterSoundID, SOUND_SMSEP);
		return true;
	}	
	else return false;
} // End "LM::ActivateGear"

// --------------------------------------------------------------
// Jettison descent stage
// --------------------------------------------------------------
void LM::StageSep (void)
{
	if (DescentStage_attached) // Sanity check, is there a descent stage to jettison?
	{
		// Create descent stage as seperate vessel
		char name[255];
		strcpy (name, GetName()); 
		strcat (name, "_DescentStage");
		OBJHANDLE oh_ds = SpawnObject (this, name, "AAPO/Parts/DescentStage", LM_DES_OFFSET, 0.1);

		// Match descent stage's propellant levels to LM's descent tanks  
		VESSEL *ds = oapiGetVesselInterface (oh_ds);		// get interface for descent stage
		ds->SetPropellantMass (ds->GetPropellantHandleByIndex (0), GetPropellantMass (ph_descent));
		ds->SetPropellantMass (ds->GetPropellantHandleByIndex (1), GetPropellantMass (ph_descent_O2));
		ds->SetPropellantMass (ds->GetPropellantHandleByIndex (2), GetPropellantMass (ph_descent_H2O));

		// Play sound effect
		PlayVesselWave (OrbiterSoundID, SOUND_STAGESEP);	

		// Update vessel state
		DescentStage_attached = false;
	} // End "if (DescentStage_Attached)"

} // End "LM::StageSep"

// --------------------------------------------------------------
// Set Re-Entry configuration (Update physical parameters and flight model)
// --------------------------------------------------------------
void LM::SetAscentConfig (void)
{
	if (AscentConfig == false)// If Re-Entry Config has not already been set...
	{
		// Shift CoG
		ShiftCG (LM_ASC_OFFSET);								

		// Update physical parameters
		SetSize (ASC_SIZE);											// Update size
		SetEmptyMass (ASC_EMPTYMASS);								// Update mass
		SetPMI (ASC_PMI);											// Update moments of inertia
		SetCrossSections (ASC_CS);									// Update cross sections
		SetRotDrag (ASC_RD);										// Update rotational drag coefficients
		SetTouchdownPoints (ASC_TDP[0], ASC_TDP[1], ASC_TDP[2]);	// Update touchdown points
		SetCW ( ASC_CW[0], ASC_CW[1], ASC_CW[2], ASC_CW[3]);		// Update wind resistance coefficients

		// Update config
		DelMesh (mesh_Descent);
		DelAttachment (ah_booster);
		AscentConfig = true;									
	}
}

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void LM::clbkSetClassCaps (FILEHANDLE cfg)
{
	// Read dificulty modifiers from config file
	oapiReadItem_float (cfg, "THRUST_MODIFIER", thrust_mod);			// Max thrust multiplier 
	oapiReadItem_float (cfg, "DV_MODIFIER", dv_mod);					// Exhaust velocity multiplier
	oapiReadItem_float (cfg, "CONSUMABLES_MODIFIER", consumables_mod);	// Consumables consumption rate (1 = 100% of historical rate)
	
	// Initialize UMMU
	UmmuSetClassCaps (cfg);

	// Physical parameters
	SetSize (LM_SIZE);
	SetEmptyMass (ASC_EMPTYMASS + DES_EMPTYMASS);
	SetPMI (LM_PMI);
	SetCrossSections (LM_CS);
	SetRotDrag (LM_RD);
	SetTouchdownPoints (LM_TDP[0], LM_TDP[1], LM_TDP[2]);
	SetSurfaceFrictionCoeff (1, 1);

	// Docking port definition
	dh_lm_dock		= CreateDock (ASC_DOCK_POS + LM_ASC_OFFSET, ASC_DOCK_DIR, ASC_DOCK_ROT);
	
	// Attachment point definition
	ah_booster		= CreateAttachment (true, _V( 0.00,-0.27, 0.00), _V( 0,-1, 0), _V( 0, 0, 1), "X");	// for attachment to launch vehicle

	// Propellant resources
	
	for (int i = 0; i < 2; i ++)
	{
		ph_rcs[i] = CreatePropellantResource (ASC_RCS_FUELMASS/2);		// RCS propellant tanks
	}
	ph_ascent		= CreatePropellantResource (ASC_FUELMASS);			// Ascent stage propellant tank
	ph_ascent_O2	= CreatePropellantResource (ASC_O2MASS);			// Ascent stage O2 tank
	ph_ascent_H2O	= CreatePropellantResource (ASC_WATERMASS);			// Ascent stage water tank

	ph_descent		= CreatePropellantResource (DES_FUELMASS);			// Descent stage propellant tank
	ph_descent_O2	= CreatePropellantResource (DES_O2MASS);			// Descent stage O2 tank
	ph_descent_H2O	= CreatePropellantResource (DES_WATERMASS);			// Descent stage water tank

	// RCS engines: Quadrant 1 (Port Fwd)
	th_rcs_a[0]	= CreateThruster (ASC_RCS_ENGINEPOS[0] + LM_ASC_OFFSET, _V( 0,-1, 0), ASC_RCS_MAXTHRUST, ph_rcs[0], ASC_RCS_ISP * dv_mod); // Down
	th_rcs_a[1]	= CreateThruster (ASC_RCS_ENGINEPOS[1] + LM_ASC_OFFSET, _V( 0, 0,-1), ASC_RCS_MAXTHRUST, ph_rcs[0], ASC_RCS_ISP * dv_mod); // Back
	th_rcs_b[0]	= CreateThruster (ASC_RCS_ENGINEPOS[0] + LM_ASC_OFFSET, _V( 0, 1, 0), ASC_RCS_MAXTHRUST, ph_rcs[1], ASC_RCS_ISP * dv_mod); // Up
	th_rcs_b[1]	= CreateThruster (ASC_RCS_ENGINEPOS[1] + LM_ASC_OFFSET, _V( 1, 0, 0), ASC_RCS_MAXTHRUST, ph_rcs[1], ASC_RCS_ISP * dv_mod); // Right

	// RCS engines: Quadrant 2 (Port Aft)
	th_rcs_a[2]	= CreateThruster (ASC_RCS_ENGINEPOS[2] + LM_ASC_OFFSET, _V( 0, 1, 0), ASC_RCS_MAXTHRUST, ph_rcs[0], ASC_RCS_ISP * dv_mod); // Up
	th_rcs_a[3]	= CreateThruster (ASC_RCS_ENGINEPOS[3] + LM_ASC_OFFSET, _V( 0, 0, 1), ASC_RCS_MAXTHRUST, ph_rcs[0], ASC_RCS_ISP * dv_mod); // Fwd
	th_rcs_b[2]	= CreateThruster (ASC_RCS_ENGINEPOS[2] + LM_ASC_OFFSET, _V( 0,-1, 0), ASC_RCS_MAXTHRUST, ph_rcs[1], ASC_RCS_ISP * dv_mod); // Down 
	th_rcs_b[3]	= CreateThruster (ASC_RCS_ENGINEPOS[3] + LM_ASC_OFFSET, _V( 1, 0, 0), ASC_RCS_MAXTHRUST, ph_rcs[1], ASC_RCS_ISP * dv_mod); // Right

	// RCS engines: Quadrant 3 (Stbd Aft)
	th_rcs_a[4]	= CreateThruster (ASC_RCS_ENGINEPOS[4] + LM_ASC_OFFSET, _V( 0,-1, 0), ASC_RCS_MAXTHRUST, ph_rcs[0], ASC_RCS_ISP * dv_mod); // Down
	th_rcs_a[5]	= CreateThruster (ASC_RCS_ENGINEPOS[5] + LM_ASC_OFFSET, _V(-1, 0, 0), ASC_RCS_MAXTHRUST, ph_rcs[0], ASC_RCS_ISP * dv_mod); // Left
	th_rcs_b[4]	= CreateThruster (ASC_RCS_ENGINEPOS[4] + LM_ASC_OFFSET, _V( 0, 1, 0), ASC_RCS_MAXTHRUST, ph_rcs[1], ASC_RCS_ISP * dv_mod); // Up
	th_rcs_b[5]	= CreateThruster (ASC_RCS_ENGINEPOS[5] + LM_ASC_OFFSET, _V( 0, 0, 1), ASC_RCS_MAXTHRUST, ph_rcs[1], ASC_RCS_ISP * dv_mod); // Fwd

	// RCS engines: Quadrant 4 (Stbd Fwd)
	th_rcs_a[6]	= CreateThruster (ASC_RCS_ENGINEPOS[6] + LM_ASC_OFFSET, _V( 0, 1, 0), ASC_RCS_MAXTHRUST, ph_rcs[0], ASC_RCS_ISP * dv_mod); // Up
	th_rcs_a[7]	= CreateThruster (ASC_RCS_ENGINEPOS[7] + LM_ASC_OFFSET, _V(-1, 0, 0), ASC_RCS_MAXTHRUST, ph_rcs[0], ASC_RCS_ISP * dv_mod); // Left
	th_rcs_b[6]	= CreateThruster (ASC_RCS_ENGINEPOS[6] + LM_ASC_OFFSET, _V( 0,-1, 0), ASC_RCS_MAXTHRUST, ph_rcs[1], ASC_RCS_ISP * dv_mod); // Down
	th_rcs_b[7]	= CreateThruster (ASC_RCS_ENGINEPOS[7] + LM_ASC_OFFSET, _V( 0, 0,-1), ASC_RCS_MAXTHRUST, ph_rcs[1], ASC_RCS_ISP * dv_mod); // Back

	// Ascent engine
	th_ascent	= CreateThruster (ASC_ENGINEPOS + LM_ASC_OFFSET, _V( 0, 1, 0), ASC_MAXTHRUST * thrust_mod, ph_ascent, ASC_ISP * dv_mod);

	// Descent engine
	th_descent	= CreateThruster (DES_ENGINEPOS + LM_DES_OFFSET, _V( 0, 1, 0), DES_MAXTHRUST * thrust_mod, ph_descent, DES_ISP * dv_mod);	
	
	// Camera parameters
	SetCameraOffset (_V( 0.0, 2.1, 0.0));

	// Associate meshes for the visual
	mesh_Cockpit	= AddMesh (mh_cockpit, &LM_ASC_OFFSET);	// Virtual cockpit mesh
	mesh_Ascent		= AddMesh (mh_ascent, &LM_ASC_OFFSET);	// Ascent stage mesh
	mesh_Descent	= AddMesh (mh_descent, &LM_DES_OFFSET);	// Descent stage mesh

	SetMeshVisibilityMode (mesh_Cockpit, MESHVIS_VC);		// Set cockpit mesh to be visible in virtual cockpit view 
	SetMeshVisibilityMode (mesh_Descent, MESHVIS_ALWAYS);	// Set desent stage mesh to be always visible, By default a vessel's own meshes are only visible when in an external view. I want the the descent stage's legs to be visible from the cockpit. 
} // End "LM::clbkSetClassCaps"

// --------------------------------------------------------------
// Read status from scenario file
// --------------------------------------------------------------
void LM::clbkLoadStateEx (FILEHANDLE scn, void *status)
{
	int i = 0;
	char *cbuf; 

	while (oapiReadScenario_nextline (scn, cbuf)) 
	{
		// Check to see if descent stage been jettisoned  
		if (!_strnicmp (cbuf, "DES_ATTACHED", 12))	// *Boolean parser courtesy of dbeachy1*
		{
			int temp = 0;							// must initialize temp in case sscanf fails to parse the integer; default to 0 (false) if the parse fails.  
			sscanf (cbuf+12, "%i", &temp);			// writes 4 bytes to temp -- no need to typecast the pointer since it's just a void * anyway
			DescentStage_attached = (temp != 0);	// to be tidy, a bool's value should either be true (1) or false (0)
		}

		// Load animation states
        else if (!_strnicmp (cbuf, "HATCH", 5)) // find line labeled "HATCH" in scn file
		{
			sscanf (cbuf+5, "%i %lf", &HatchStatus, &hatch_proc);	// Read values stored there to HatchStatus and hatch_proc
		}	

		else if (!_strnicmp (cbuf, "GEAR", 4)) // find line labeled "GEAR" in scn file
		{
			sscanf (cbuf+4, "%i %lf", &GearStatus, &gear_proc);		// Read values stored there to GearStatus and gear_proc
		}

		else if (!_strnicmp (cbuf, "SBAND", 5)) // find line labeled "SBAND" in scn file
		{
			sscanf (cbuf+5, "%lf %lf", &sband_proc[0], &sband_proc[1]);	// Read values stored there to GearStatus and gear_proc
		}

		else if (!_strnicmp (cbuf, "RADAR", 5)) // find line labeled "GEAR" in scn file
		{
			sscanf (cbuf+5, "%lf %lf", &radar_proc[0], &radar_proc[1]);	// Read values stored there to GearStatus and gear_proc
		}
		
		// Load cockpit panel state
		else if (vc->ParseScenarioLine (cbuf)) continue;	

		// load LM sub-system state
		else if (lm->ParseScenarioLine (cbuf)) continue;						
	
		// Load UMMUs
		else if (Crew.LoadAllMembersFromOrbiterScenario (cbuf) == true) continue;

		// Load UCGO cargoes
		else if (Cargo.LoadCargoFromScenario (cbuf) == true) continue;

		// Load default parameters
		else ParseScenarioLineEx (cbuf, status);

	} // End "while (oapiReadScenario_nextline (scn, cbuf))"

} // End "LM::clbkLoadStateEx"

// --------------------------------------------------------------
// Write status to scenario file
// --------------------------------------------------------------
void LM::clbkSaveState (FILEHANDLE scn)
{
	int i = 0;
	char cbuf[256];
	
	// Write default vessel parameters to scenario file
	VESSEL3::clbkSaveState (scn);

	// Write custom parameters to scenario file
	sprintf (cbuf, "%i", DescentStage_attached);						
	oapiWriteScenario_string (scn, "DES_ATTACHED", cbuf);

	sprintf (cbuf, "%i %0.4f", HatchStatus, hatch_proc);	// EVA Hatch status and animation state
	oapiWriteScenario_string (scn, "HATCH", cbuf);

	sprintf (cbuf, "%i %0.4f", GearStatus, gear_proc);		// Landing Gear status and animation state
	oapiWriteScenario_string (scn, "GEAR", cbuf);

	// Save cockpit panel state
	vc->SaveState (scn);
	
	// Save sub-system state
	lm->SaveState (scn);

	// Save UMmu state
	Crew.SaveAllMembersInOrbiterScenarios (scn);

	// Save UCGO cargo state
	Cargo.SaveCargoToScenario(scn);

} // End "LM::clbkSaveState"

// --------------------------------------------------------------
// Finalise vessel creation
// --------------------------------------------------------------
void LM::clbkPostCreation (void)
{
	DefineAnimations ();
	DefineLighting ();
	DefineParticleStreams ();
	DefineSounds ();

	vc->InitVC (mesh_Cockpit, lm);
	//lm->InitLM (vc)

	// If descent stage has already been jettisoned we need to apply the changes that would normally happen durring stage seperation...
	if (DescentStage_attached == false)	SetAscentConfig ();			

} // End "LM::clbkPostCreation"

// --------------------------------------------------------------
// Pre-Step processes
// --------------------------------------------------------------
void LM::clbkPreStep (double simt, double simdt, double mjd)
{
	// Update Vessel Mass
	SetEmptyMass (LM_EmptyMass ());

	// if GetAttachmentStatus returns a value other than NULL...
	if (GetAttachmentStatus (ah_booster) != NULL) 
	{
		// Fold and stow antennas and landing gear
		GearStatus = CLOSED;
		gear_proc = 0;
		SetAnimation (anim_Gear, gear_proc);

		radar_proc[0] = 0.28;
		SetAnimation (anim_RadarPitch, radar_proc[0]); 	

		radar_proc[1] = 0.50;
		SetAnimation (anim_RadarYaw, radar_proc[1]); 	

		// Disable engines
		SetThrusterLevel (th_descent, 0);	
		SetThrusterLevel (th_ascent, 0);

		Booster_attached = true;	// LM is attached to a launch vehicle	 
	}

	// If decent stage is attached...
	if (DescentStage_attached) 
	{
		SetDefaultPropellantResource (ph_descent);			// Set descent stage fuel tank as default propellant resource
		SetISP (DES_ISP);										// Set descent stage engine ISP as default
	}
	else // if not... 
	{
		SetDefaultPropellantResource (ph_ascent);			// Set ascent stage fuel tank as default propellant resource
		SetISP (ASC_ISP);										// Set ascent stage engine ISP as default 

		SetPropellantMass (ph_descent, 0.0);						
		SetPropellantMass (ph_descent_H2O, 0.0);
		SetPropellantMass (ph_descent_O2, 0.0);

		// If physical parameters have not been updated, update them... 
		if ((AscentConfig == false) && (GroundContact () == false)) SetAscentConfig ();
	}

	// Sub-Class callbacks
	vc->PreStep (simt, simdt, mjd);
	lm->PreStep (simt, simdt, mjd);
} // End "LM::clbkPreStep"

// --------------------------------------------------------------
// Post-Step processes
// --------------------------------------------------------------
void LM::clbkPostStep (double simt, double simdt, double mjd)
{
	// Animate EVA Hatch
	if (HatchStatus > OPEN)	
	{
		double	delta = simdt / 3;							// delta = simdt divided by duration of animation in seconds

		if (HatchStatus == OPENING)
		{
			if (hatch_proc < 1) hatch_proc += delta;	// If process value is less than 1 add delta to process value
			else HatchStatus = OPEN;					// If process value is greater than 1 set status to OPEN

			if (hatch_proc < 0.2) PlayVesselWave (OrbiterSoundID, SOUND_HATCHOPEN); // Play sound effect
		}
		
		if (HatchStatus == CLOSING)
		{
			if (hatch_proc > 0) hatch_proc -= delta;	// If process value is Greater than 0 subtract delta from process value
			else HatchStatus = CLOSED;					// If process value is less than 0 set status to CLOSED

			if (hatch_proc < 0.3) PlayVesselWave (OrbiterSoundID, SOUND_HATCHCLOSE); // Play sound effect
		}
		
		SetAnimation (anim_Hatch, hatch_proc);			// Update animation state
	}
	
	//// Debuggery
	//sprintf (oapiDebugString (), "Hatch Status %0.0f, hatch_proc %0.3f", (float)HatchStatus, hatch_proc);
	// End "if (HatchStatus > OPEN)"

	// Animate Landing Gear
	if (GearStatus > OPEN)	
	{
		double	delta = simdt / 0.3;				// delta = simdt divided by duration of animation in seconds

		if (GearStatus == OPENING)
		{
			if (gear_proc < 1) gear_proc += delta;	// If process value is less than 1 add delta to process value
			else GearStatus = OPEN;					// If process value is greater than 1 GearStatus = OPEN
		}
		
		if (GearStatus == CLOSING)
		{
			if (gear_proc > 0) gear_proc -= delta;	// If process value is Greater than 0 subtract delta from process value
			else GearStatus = CLOSED;				// If process value is less than 0 GearStatus = CLOSED
		}

		SetAnimation (anim_Gear, gear_proc);		

		//// Debuggery	
		//sprintf (oapiDebugString (), "Gear Status %0.0f, gear_proc %0.3f", (float)GearStatus, gear_proc);
	} // End "if (GearStatus > OPEN)"

	// Sub-Class callbacks
	vc->PostStep (simt, simdt, mjd);
	lm->PostStep (simt, simdt, mjd);
	UmmuPostStep (simt, simdt, mjd);

} // End "LM::clbkPostStep"

// --------------------------------------------------------------
// Setup the virtual cockpit HUD
// --------------------------------------------------------------
bool LM::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{
	if (oapiGetFocusInterface () != this) return false;

	// draw the default HUD
	VESSEL3::clbkDrawHUD (mode, hps, skp);

	// Ummu display messages
	if (dHudMessageDelay > 0)
	{
		skp->Text (5, hps->H / 60 * 15, cUmmuHudDisplay, strlen (cUmmuHudDisplay));
		dHudMessageDelay -= oapiGetSimStep ();

		if (dHudMessageDelay < 0) dHudMessageDelay = 0;
	}

	return true;
}

// --------------------------------------------------------------
// Respond to RCS mode change
// --------------------------------------------------------------
void LM::clbkRCSMode (int mode)
{
} // End "LM::clbkRCSMode"

// --------------------------------------------------------------
// Respond to MFD mode change
// --------------------------------------------------------------
void LM::clbkMFDMode (int mfd, int mode)
{
	// Update MFD Button Labels
	switch (mfd) 
	{
	case 0: // Panel 4 MFD
		oapiTriggerRedrawArea (0, -1, AID_MFD0_LBUTTONS);
		oapiTriggerRedrawArea (0, -1, AID_MFD0_RBUTTONS);
		break;
	}
} // End "LM::clbkMFDMode"

// --------------------------------------------------------------
// Respond to autopilot mode change
// --------------------------------------------------------------
void LM::clbkNavMode (int mode, bool active)
{
} // End "LM::clbkNavMode" 

// --------------------------------------------------------------
// Respond to changes in dock status
// --------------------------------------------------------------
void LM::clbkDockEvent (int dock, OBJHANDLE mate)
{
	// if LM is attached to a launch vehicle and another vessel docks with it, detach the LM from the LV.
	if ((Booster_attached) && (mate != NULL)) DetachParent (this, 0.01);
	
} // End "LM::clbkDockEvent" 

// --------------------------------------------------------------
// Process keyboard inputs
// --------------------------------------------------------------
int  LM::clbkConsumeBufferedKey (DWORD key, bool down, char *kstate)
{
	// only process keydown events
	if (!down) return 0; 

	// Egress when [E] is pressed.
	if (key == OAPI_KEY_E && !KEYMOD_SHIFT(kstate) && !KEYMOD_CONTROL (kstate) && !KEYMOD_ALT(kstate)) // [K] is down, no [shift], no [ctrl], no [alt]
	{
		UmmuTransferCrewMember (iSelectedUmmuMember);	// Activate hatch
		return 1;										// return '1' to indicate that the input has been processed
	}

	// Deploy Ganding Gear when [G] is pressed.
	if (key == OAPI_KEY_G && !KEYMOD_SHIFT(kstate) && !KEYMOD_CONTROL (kstate) && !KEYMOD_ALT(kstate)) // [K] is down, no [shift], no [ctrl], no [alt]
	{
		ActivateGear ();
		return 1;
	}

	// Open hatch when [K] is pressed.
	if (key == OAPI_KEY_K && !KEYMOD_SHIFT(kstate) && !KEYMOD_CONTROL (kstate) && !KEYMOD_ALT(kstate)) // [K] is down, no [shift], no [ctrl], no [alt]
	{
		ActivateHatch ();	// Activate hatch
		return 1;			// return '1' to indicate that the input has been processed
	}

	// Shift crew position when [TAB] is pressed 
	if (key == OAPI_KEY_TAB && !KEYMOD_SHIFT(kstate) && !KEYMOD_CONTROL(kstate) && !KEYMOD_ALT(kstate)) // [TAB] is down, no [shift], no [ctrl], no [alt]
	{
		if (CrewPos != 0) CrewPos = 0;
		else CrewPos = 1;
		clbkLoadVC (0);
		return 1;										
	}

	 // Kill thrust when Numpad [*] key is pressed.
	if(key == OAPI_KEY_MULTIPLY  && down && !KEYMOD_SHIFT(kstate) && !KEYMOD_CONTROL (kstate) && !KEYMOD_ALT(kstate))
	{
		lm->KillThrust ();
		return 1;
	}

	// Jettison descent stage when [Ctrl] + [J] is pressed.
	if (key == OAPI_KEY_J && !KEYMOD_SHIFT(kstate) && KEYMOD_CONTROL (kstate) && !KEYMOD_ALT(kstate))
	{ 
		// If attached to a booster or launch vehicle
		if (GetAttachmentStatus (ah_booster) != NULL) DetachParent (this, 0.01);
		else StageSep();	
		return 1;			
	}
	return 0;	// if no keys are pressed the function returns '0' and nothing happens
} // End "LM::clbkConsumeBufferedKey"

// --------------------------------------------------------------
// Load generic "Glass" cockpit
// --------------------------------------------------------------
bool LM::clbkLoadGenericCockpit ()
{
	// Check CG offset 
	VECTOR3 ofs;
	if (AscentConfig) ofs = _V( 0, 0, 0);
	else ofs = LM_ASC_OFFSET; 

	// Set camera parameters
	SetCameraOffset (_V(-0.57, 0.68, 1.12) + ofs);
	SetCameraDefaultDirection (_V( 0, 0, 1));
	oapiCameraSetCockpitDir ( 0, 0);								

	return true;
}

// --------------------------------------------------------------
// Load virtual cockpit mode
// --------------------------------------------------------------
bool LM::clbkLoadVC (int id)
{
	// Check CG offset 
	VECTOR3 ofs;
	if (AscentConfig) ofs = _V( 0, 0, 0);
	else ofs = LM_ASC_OFFSET; 

	// Register Camera (view) properties
	if (CrewPos == 0)	// Left Seat	
	{
		switch (id)
		{
		case 0:			// Commander's position looking forward
			SetCameraOffset (_V(-0.57, 0.68, 1.12) + ofs);				// Set camera position (x,y,z)
			SetCameraDefaultDirection (_V( 0, 0, 1));					// Set camera direction (x,y,z)
			SetCameraRotationRange (RAD*120, RAD*120, RAD*60, RAD*60);	// Set camera range of motion (Left, Right, Up, Down)
			oapiVCSetNeighbours (-1, 1,-1,-1);							// Set adjacent cameras (Left, Right, Top, Bottom)
			break;
	
		case 2:			// Commander's position (looking up through COAS Reticle)
			SetCameraOffset (_V(-0.59, 0.68, 1.12) + ofs);				// Set camera position (x,y,z)
			SetCameraDefaultDirection (_V( 0, 1, 0));					// Set camera direction (x,y,z)
			SetCameraRotationRange (RAD*15, RAD*15, RAD*15, RAD*15);	// Set camera range of motion (Left, Right, Up, Down)
			oapiVCSetNeighbours (-1,-1,-1, 0);							// Set adjacent cameras (Left, Right, Top, Bottom)
			break;
		}
	} // end "switch (id)"

	else
	{
		SetCameraOffset (_V( 0.57, 0.68, 1.12) + ofs);				// Set camera position (x,y,z)
		SetCameraDefaultDirection (_V( 0, 0, 1));					// Set camera direction (x,y,z)
		SetCameraRotationRange (RAD*120, RAD*120, RAD*60, RAD*60);	// Set camera range of motion (Left, Right, Up, Down)
		oapiVCSetNeighbours ( 0,-1,-1,-1);							// Set adjacent cameras (Left, Right, Top, Bottom)
	}


	//// Register Camera (view) properties
	//switch (id)
	//{
	//case 0: // Commander's position
	//	SetCameraOffset (_V(-0.57, 0.68, 1.12) + ofs);				// Set camera position (x,y,z)
	//	SetCameraDefaultDirection (_V( 0, 0, 1));					// Set camera direction (x,y,z)
	//	SetCameraRotationRange (RAD*120, RAD*120, RAD*60, RAD*60);	// Set camera range of motion (Left, Right, Up, Down)
	//	oapiVCSetNeighbours (-1, 1, 2,-1);							// Set adjacent cameras (Left, Right, Top, Bottom)
	//	break;

	//case 1: // Pilot's position
	//	SetCameraOffset (_V( 0.57, 0.68, 1.12) + ofs);				// Set camera position (x,y,z)
	//	SetCameraDefaultDirection (_V( 0, 0, 1));					// Set camera direction (x,y,z)
	//	SetCameraRotationRange (RAD*120, RAD*120, RAD*60, RAD*60);	// Set camera range of motion (Left, Right, Up, Down)
	//	oapiVCSetNeighbours ( 0,-1,-1,-1);							// Set adjacent cameras (Left, Right, Top, Bottom)
	//	break;

	//case 2: // Commander's position (looking up through COAS Reticle)
	//	SetCameraOffset (_V(-0.59, 0.68, 1.12) + ofs);				// Set camera position (x,y,z)
	//	SetCameraDefaultDirection (_V( 0, 1, 0));					// Set camera direction (x,y,z)
	//	SetCameraRotationRange (RAD*15, RAD*15, RAD*15, RAD*15);	// Set camera range of motion (Left, Right, Up, Down)
	//	oapiVCSetNeighbours (-1, 1,-1, 0);							// Set adjacent cameras (Left, Right, Top, Bottom)
	//	break;
	//} // end "switch (id)"

	// Register MFD
	static VCMFDSPEC Cmd_MFD = {mesh_Cockpit, VC_GRP_MFD_Cmd};		// Surface on which to display MFD. (Mesh, Mesh Group)
	oapiVCRegisterMFD (0, &Cmd_MFD);								// Register MFD in orbiter's interface [registry # and VCMFDSPEC]. (registry numbers start at 0 and count up)

	// Register active areas
	vc->RegisterActiveAreas (ofs);

	return true;
} // End "LM::clbkLoadVC"

// --------------------------------------------------------------
// Respond to virtual cockpit mouse events
// --------------------------------------------------------------
bool LM::clbkVCMouseEvent (int id, int event, VECTOR3 &p)
{
	//if (UMmuCrewCount <= 0) return false;		// If no one is onboard disable cockpit inputs, otherwise...
	//else return vc->MouseEvent (id, event, p);	// pass inputs to cockpit control class.

	return vc->MouseEvent (id, event, p);	// pass inputs to cockpit control class.
} // End "LM::clbkLoadVC"

// --------------------------------------------------------------
// Respond to virtual cockpit area redraw requests
// --------------------------------------------------------------
bool LM::clbkVCRedrawEvent (int id, int event, SURFHANDLE surf)
{
	if (oapiGetFocusInterface () == this) // secure calls to prevent conflicts with other vessels or UMMUs
	{
		return vc->RedrawEvent (id, event, surf); // Pass inputs to cockpit class
	}
	else return false;
} // End "LM::clbkVCRedrawEvent"

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new LM (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (LM*)vessel;
}
