//// ==============================================================
////                    ORBITER MODULE: LM
////                  Part of the ORBITER SDK
////               Copyright (C) 2012 Greg Hlynka
////                    All rights reserved
////
//// LM.h
//// Interface for LM vessel class
////
//// Notes: This file exists
//// ==============================================================

#pragma once
#define STRICT
#define ORBITER_MODULE

// Orbiter SDK files
#include "orbitersdk.h"
#include "AAPO_Common.h"

// Mesh resource files
#include "mesh_LM_AscentStage.h"
#include "mesh_LM_DescentStage.h"
#include "mesh_LM_VC.h"

// Other resources and sub-classes
#include "LM_SwitchDefs.h"

// ==============================================================
// Combined vessel parameters
// ==============================================================
const VECTOR3	LM_ASC_OFFSET		= { 0.00, 1.44, 0.00};		// Offset of Ascent stage CoG from combined CoG
const VECTOR3	LM_DES_OFFSET		= { 0.00,-0.80, 0.00};		// Offset of Descent stage CoG from combined CoG
const double	LM_SIZE				= 5;						// Vessel's mean radius [m]
const VECTOR3	LM_CS				= { 19.0, 19.2, 18.8};		// x,y,z cross sections [m^2]
const VECTOR3	LM_PMI				= { 2.68, 2.10, 2.64};		// Principal moments of inertia (mass-normalised) [m^2]
const VECTOR3	LM_RD				= {	0.04, 0.05, 0.04};		// Rotation drag coefficients
const VECTOR3	LM_TDP[3]			= {{ 0.0,-3.1, 4.4}, {-3.8,-3.1,-2.2}, { 3.8,-3.1,-2.2}};	// touchdown points [m]

const VECTOR3	LM_UMMU_LADDERPOS	= { 0.00,-1.74, 4.6};		// Position of Crew ladder (UMMU airlock and spawn position while landed)

// ==============================================================
// Ascent stage vessel parameters
// ==============================================================
const double	ASC_SIZE				= 2.5;					// Mean radius [m] after staging
const VECTOR3	ASC_CS					= { 9.62, 12.6, 9.25};	// Cross sections [m^2] after staging
const VECTOR3	ASC_PMI					= {	1.31, 1.68, 1.25};	// Principal moments of inertia (mass-normalised) [m^2] after staging
const VECTOR3	ASC_RD					= {	0.03, 0.07, 0.09};	// Rotation drag coefficients after staging
const double	ASC_CW[4]				= { 0.47, 0.47, 0.47, 0.47};	// Wind resistance coefficients (forward, back, lateral, vertical)
const VECTOR3	ASC_TDP[3]				= {{ 0.0,-1.4, 1.7}, {-0.9,-1.4,-1.5}, { 0.9,-1.4,-1.5}};	// touchdown points [m] after staging

const double	ASC_EMPTYMASS			= 1922.78;				// Ascent stage empty mass [kg] (4239 lbs, page 119)
const double	ASC_FUELMASS			= 2353.0;				// Ascent stage fuel mass [kg] 
const double	ASC_RCS_FUELMASS		= 633 * KILOS;				// RCS fuel mass [kg] (633 lbs, LMFM pg 72)
const double	ASC_WATERMASS			= 19.55;				// max fuel mass [kg] (page 177)
const double	ASC_O2MASS				= 1.102;				// max fuel mass [kg] (2.43 lbs)

const double	ASC_ISP					= 3050.91;				// Ascent stage specific impulse [m/s] (311 sec, Astronautix.com)
const double	ASC_MAXTHRUST			= 15568.8;				// Ascent stage max thrust [n] (3,500 lbf, LMFM pg 58)
const VECTOR3	ASC_ENGINEPOS			= {	0.00,-0.64, 0.00};	// Ascent stage engine position [m]

const double	ASC_RCS_ISP				= 290 * G;					// RCS specific impulse [m/s] (290 sec, Astronautix.com)
const double	ASC_RCS_MAXTHRUST		= 100 * NEWTONS;			// RCS thruster max thrust [n] (100 lbf)
const VECTOR3	ASC_RCS_ENGINEPOS[8]	= {{-1.68, 0.00, 1.68}, {-1.562, 0.00, 1.562}, {-1.68, 0.00,-1.68}, {-1.562, 0.00,-1.562}, { 1.68, 0.00,-1.68}, { 1.562, 0.00,-1.562}, { 1.68, 0.00, 1.68}, { 1.562, 0.00, 1.562}}; // RCS engine positions [m]

// Docking port specs
const VECTOR3	ASC_DOCK_POS			= { 0.00, 1.51, 0.00};				// docking port location [m]
const VECTOR3	ASC_DOCK_DIR			= { 0, 1, 0};						// docking port approach direction
const VECTOR3	ASC_DOCK_ROT			= { sin(RAD*298), 0, cos(RAD*298)};	// docking port alignment direction

// UMMU & UCCGO Support
const int		ASC_UMMU_CREWSIZE	= 3;						// I added an additional slot in case someone wants to reenact Apollo 13 ;)
const VECTOR3	ASC_UMMU_HATCHPOS	= {-0.04,-0.54, 2.3};		// Position of EVA hatch
const VECTOR3	ASC_UMMU_CREWROT	= {-RAD*90, RAD*180, 0};	// Orientation of UMMU on EVA

// Local Light sources
const VECTOR3	ASC_STROBE_POS		= { 0.00, 1.84, 2.28};
const VECTOR3	ASC_STROBE_COL		= { 0.9, 1.0, 1.0}; 
const VECTOR3	ASC_BEACON_POS[5]	= {{-2.52,-0.40, 0.20}, { 1.91,-0.70, 0.22}, {-0.28, 0.48, 2.27}, { 0.28, 0.48, 2.27}, { 0.00, 0.86,-1.81}};	
const VECTOR3	ASC_BEACON_COL[5]	= {{ 1.0, 0.5, 0.5}, { 0.5, 1.0, 0.5}, { 1.0, 1.0, 0.5}, { 1.0, 1.0, 0.5}, { 1.0, 1.0, 1.0}};					// Red, Green, Yellow x2, and White


// ==============================================================
// Lunar Module class interface
// ==============================================================

class LM: public VESSEL3 
{
public:
	LM (OBJHANDLE hVessel, int flightmodel);
	~LM ();

	// Member classes
	friend class	LM_COCKPIT;
	LM_COCKPIT		*vc;					// Virtual cockpit interface

	friend class	LM_SYSTEMS;
	LM_SYSTEMS		*lm;					// Subsytem management interface

	// Custom vessel functions
	void	DefineAnimations (void);												// Define mesh animations
	void	DefineLighting (void);													// Define local light sources
	void	DefineParticleStreams (void);											// Define exhaust and particle streams
	void	DefineSounds (void);													// Define sound effects

	double	LM_EmptyMass (void);													// Current empty mass of vessel
	bool	ActivateHatch (void);													// EVA Hatch control logic
	bool	ActivateGear (void);													// Landing gear control logic
	void	StageSep (void);														// Jettison descent stage
	void	SetAscentConfig (void);

	// Overloaded callback functions
	void	clbkSetClassCaps (FILEHANDLE cfg);										// Set the capabilities of the vessel class
	void	clbkLoadStateEx (FILEHANDLE scn, void *status);							// Read status from scenario file
	//void	clbkSetStateEx (const void *status);									// Set state parameters during vessel creation.
	void	clbkSaveState (FILEHANDLE scn);											// Write status to scenario file
	void	clbkPostCreation (void);												// Finalise vessel creation
	void	clbkPreStep (double simt, double simdt, double mjd);					// Pre-Step processes
	void	clbkPostStep (double simt, double simdt, double mjd);					// Post-Step processes
	//bool	clbkPlaybackEvent (double simt, double event_t, const char *event_type, const char *event);
	//void	clbkVisualCreated (VISHANDLE vis, int refcount);						// Called after a vessel visual has been created by the renderer.
	//void	clbkVisualDestroyed (VISHANDLE vis, int refcount);						// Called before a vessel visual is destroyed. 
	bool	clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);	// HUD redraw notification.
	//void	clbkRenderHUD (int mode, const HUDPAINTSPEC *hps, SURFHANDLE hTex);		// HUD render function.
	void	clbkRCSMode (int mode);													// Reaction Control System mode change notification. 
	//void	clbkADCtrlMode (DWORD mode);											// Aerodynamic control surface mode change notification.
	//void	clbkHUDMode (int mode);													// HUD mode change notification.
	void	clbkMFDMode (int mfd, int mode);										// MFD mode change modification.
	void	clbkNavMode (int mode, bool active);									// Navigation mode change notification.
	void	clbkDockEvent (int dock, OBJHANDLE mate);								// Docking event notification.
	//void	clbkAnimate (double simt)												// Manual animation notification.
	//int	clbkConsumeDirectKey (char *kstate);									// Keyboard status notification. 
	int		clbkConsumeBufferedKey (DWORD key, bool down, char *kstate);			// Process keyboard inputs
	bool	clbkLoadGenericCockpit ();
	//bool	clbkLoadPanel (int id);
	//bool	clbkLoadPanel2D (int id, PANELHANDLE hPanel, DWORD viewW, DWORD viewH);
	//bool	clbkPanelMouseEvent (int id, int event, int mx, int my, void *context);
	//bool	clbkPanelRedrawEvent (int id, int event, SURFHANDLE surf, void *context);
	bool	clbkLoadVC (int id);													// Load virtual cockpit mode
	bool	clbkVCMouseEvent (int id, int event, VECTOR3 &p);						// Respond to virtual cockpit mouse events
	bool	clbkVCRedrawEvent (int id, int event, SURFHANDLE surf);					// Respond to virtual cockpit area redraw requests
	//int  clbkGeneric (int msgid, int prm, void *context);

	// Implimentation of OrbiterSound 4.0
	int		OrbiterSoundID;

	// Implimentation of UMmu 2.5
	UMMUCREWMANAGMENT Crew;															// UMmu class interface
	int iSelectedUmmuMember;														// Select crew member to eva
	char cUmmuHudDisplay[256];														// UMmu hud char variable
	double dHudMessageDelay;														// UMmu hud display delay
	char *SendHudMessage (void);													// UMmu hud display function

	void	UmmuSetClassCaps (FILEHANDLE cfg);										// Set the capabilities of the vessel class
	void	UmmuPostStep (double simt, double simdt, double mjd);					// Post-Step processes
	
	// The Add Mmu without scenery editor declaration see PDF doc
	char	cAddUMmuToVessel[255];
	void	AddUMmuToVessel (BOOL bStartAdding = false);

	bool	UmmuGetAirlockState ();
	bool	UmmuSelectCrewMember ();
	bool	UmmuTransferCrewMember (int slot);

	// UCGO 2.5 CLASS HANDLE FUNCTION AND VARIABLES
	UCGO	Cargo;			// Cargo class handle
	int		iSelectedCargo;	// for the selection of cargos -1 by default
	bool	UcgoGrappleCargo (int slot);
	bool	UcgoReleaseCargo (int slot);

	// Thrusters and Propellant handles
	PROPELLANT_HANDLE	ph_rcs[2], ph_sm_rcs[4], ph_ascent, ph_descent;				// Functional Propellant tanks
	PROPELLANT_HANDLE	ph_ascent_O2, ph_ascent_H2O, ph_descent_O2, ph_descent_H2O;	// Represenative propellant tanks (consumables)
	
	THRUSTER_HANDLE		th_rcs_a[8], th_rcs_b[8], th_ascent, th_descent;			// Functional thrusters
	THRUSTER_HANDLE		th_dust, th_vent;											// Represenative thrusters (particle streams)

	AIRFOILHANDLE		afh_vert, afh_horz;											// Aerodynamic forces

	ATTACHMENTHANDLE	ah_booster, ah_payload;
	DOCKHANDLE			dh_lm_dock;
	OBJHANDLE			oh_LV, oh_descent;

	MESHHANDLE			mh_ascent, mh_cockpit, mh_descent;							// Mesh handles

private:
	// Vessel status variables
	double		thrust_mod, dv_mod, consumables_mod;
	bool		Booster_attached, DescentStage_attached, AscentConfig;
	int			UMmuCrewCount;	// Number of souls onboard
	int			CrewPos;
	doorstate	HatchStatus, GearStatus;
	
	double		gear_proc, hatch_proc, radar_proc[2], sband_proc[2];

	// Visuals
	UINT				mesh_Descent, mesh_Ascent, mesh_Cockpit;
	UINT				anim_Hatch, anim_RadarPitch, anim_RadarYaw, anim_SBandPitch, anim_SBandYaw, anim_Gear;	
	MGROUP_TRANSFORM	*mgt_Leg[4], *mgt_Strut[4], *mgt_Downlock[4];
	BEACONLIGHTSPEC		bspec_strobe, bspec_pos[5];
};