// ==============================================================
//                    ORBITER MODULE: LM
//                  Part of the ORBITER SDK
//               Copyright (C) 2012 Greg Hlynka
//                    All rights reserved
//
// LM_Cockpit.h
// Interface for LM vessel class
//
// Notes: This file exists
// ==============================================================

#pragma once
#include "orbitersdk.h"
#include "LM.h"

// ==============================================================
// Definitions
// ==============================================================

// Propellant valve definitions
#define			VALVE_ASC_HeREG1	PropValve[0]	// Ascent engine pressurant valves
#define			VALVE_ASC_HeREG2	PropValve[1]
#define			VALVE_DES_HeREG1	PropValve[2]	// Descent engine pressurant valves
#define			VALVE_DES_HeREG2	PropValve[3]
#define			VALVE_RCS_PROPSOVA	PropValve[4]	// RCS Propellant shut-of valves
#define			VALVE_RCS_PROPSOVB	PropValve[5]
#define			VALVE_RCS_OXSOVA	PropValve[6]	// RCS Oxidizer shut-of valves
#define			VALVE_RCS_OXSOVB	PropValve[7]
#define			VALVE_RCS_ASCXFDA	PropValve[8]	// RCS cross-feed valves
#define			VALVE_RCS_ASCXFDB	PropValve[9]
#define			VALVE_RCS_XFEED		PropValve[10]
#define			VALVE_FUELDUMP		PropValve[11]	// Fuel dump valve
#define			VALVE_OXDUMP		PropValve[12]	// Oxidizer dump valve

#define			VALVE_CABINDUMP		PropValve[13]
#define			VALVE_CABINO2		PropValve[14]
#define			VALVE_EGRESS		PropValve[15]

// Cartridge Activated (Explosive) Device definitions
#define			CAD_DES_ISOVALVE	CADblown[0]	
#define			CAD_GEAR_DEPLOY		CADblown[1]	
#define			CAD_STAGE_JETT		CADblown[2]	
#define			CAD_RCS_PRESS		CADblown[3]	
#define			CAD_DES_VENT		CADblown[4]	
#define			CAD_DES_PRESS		CADblown[5]	
#define			CAD_ASC_PRESS1		CADblown[6]	
#define			CAD_ASC_PRESS2		CADblown[7]

// ==============================================================
// Constants
// ==============================================================
const int		LM_PVALVECOUNT		= 19;					// Propellant valve count
const int		LM_CADCOUNT			= 12;					// CAD count 
const double	LM_SUITVOLUME		= 0.3;					// Volume of space-suit atmosphere [cubic meters]
const double	LM_SUITPRESS		= 3.75 * PA;			// Normal suit pressure [Pascals] 3.8 PSI
const double	LM_CABINVOLUME		= 6.55;					// Cabin habitable volume [cubic meters]
const double	LM_CABINTEMP		= 294.26;				// Normal cabin temprature [kelvin] 70 F*
const double	LM_CABINPRESS		= 4.8 * PA;				// Normal cabin pressure [pascals] 4.8 PSI
const double	LM_CABINPRESSLOW	= 3.5 * PA;				// Low cabin pressure [pascals] 3.5 PSI
const double	LM_CABINPRESSHIGH	= 5.4 * PA;				// High cabin pressure [pascals] 5.4 PSI
const double	LM_CABINPRESSEVA	= 0.08 * PA;			// Max cabin pressure differential [pascals] to open EVA hatch 0.08 PSI
const double	LM_CABINDUMPTIME	= 180;					// Time in seconds it takes for cabin pressure to go from normal PSI to vaccume with dump valve open
const double	LM_CABINCYCLE		= 35 * 60;				// Time in seconds it takes to completely recycle cabin atmosphere (35 minutes at rate of 6 cubic feet a minute)
const double	LM_O2FLOW			= 2.9 / 300;			// Max flow rate [kg/s] of Cabin O2 regulator
const double	LM_O2LEAK_CABIN		= 0.09 / 3600;			// Cabin leak rate at 5 psi [kg/s] 0.2 lbs/hr
const double	LM_O2LEAK_HP		= 0.0023 / 3600;		// Leakage from high pressure components [kg/s] 0.005 lbs/hr
const double	LM_O2LEAK_LP		= 0.0023 / 3600;		// Leakage from low pressure components [kg/s] 0.005 lbs/hr
const double	LM_O2CONSUMPTION	= 0.0478 / 3600;		// Metabolic consumption of O2 [kg/s] 0.081-0.130 lbs per person per hour
const double	LM_CO2PRODUCTION	= 0.0417 / 3600;		// Metabolic production of CO2 [kg/s] 0.092 lbs per person per hour


// ==============================================================
// Class declaration
// ==============================================================
class LM_SYSTEMS 
{
friend class LM;
friend class LM_COCKPIT;

public:
	LM_SYSTEMS (LM *vessel, LM_COCKPIT *cockpit);
	~LM_SYSTEMS ();

	inline const LM *GetVessel() const { return v; }

	// Public Functions
	void	InitLM (void);										// Load initial state
	bool	ParseScenarioLine (char *line);						// Read status from scenario file
	void	SaveState (FILEHANDLE scn);							// Write status to scenario file
	void	PreStep (double simt, double simdt, double mjd);	// Pre-Step processes
	void	PostStep (double simt, double simdt, double mjd);	// Post-Step processes

private:
	LM			*v;		// vessel interface
	LM_COCKPIT	*vc;	// virtual cockpit interface	

	// LM Status variables
	OBJHANDLE	rBody;
	VECTOR3		rPos, gPos, CoG_offset;
	int			nSouls;
	bool		Descent;
	doorstate	PropValve[LM_PVALVECOUNT];
	double		valve_proc[LM_PVALVECOUNT];
	bool		CADblown[LM_CADCOUNT];

	// Life support
	double		AmbientPress, AmbientTemp, CabinPress, CabinTemp, SuitPress, SuitTemp, PressDiff;
	double		percentO2, percentCO2, percentN2, percentH2O;

	double		Asc_O2_Qty, Asc_H2O_Qty, Des_O2_Qty, Des_H2O_Qty;

	double		Display_CabinTemp, Display_CabinPress, Display_SuitTemp, Display_SuitPress,  Display_CO2_Level;
	
	// Propulsion system
	double		Display_Alt, Display_VS, Display_dV, Display_Thrust, Display_Throttle;							// main propulsion system status.
	
	double		Asc_Fuel_Qty, Asc_Fuel_Temp, Asc_Fuel_Press, Asc_O2_Temp, Asc_O2_Press;
	double		Des_Fuel_Qty, Des_Fuel_Temp, Des_Fuel_Press, Des_O2_Temp, Des_O2_Press;

	double		RCS_Qty[2], RCS_Temp[2], RCS_Press[2], Display_RCS_Qty[2], Display_RCS_Temp[2], Display_RCS_Press[2];
	double		Display_Fuel_Qty, Display_Fuel_Temp, Display_Fuel_Press, Display_O2_Temp, Display_O2_Press, Display_O2_Qty, Display_H2O_Qty;
	
	double		GlycolTemp, GlycolPress;
	double		Display_GlycolTemp, Display_GlycolPress;
	
	// Private functions
	void	PropellantPreStep (double simt, double simdt);		// Propellant system pre-step processes
	void	PropellantPostStep (double simt, double simdt);		// Propellant system post-step processes

	bool	AscentPropFlow (void);			// are ascent engine's fuel lines pressurized?
	bool	DescentPropFlow (void);			// what about the descent engine's?
	bool	RCSaPropFlow (void);			// or the RCS A-loop...
	bool	RCSbPropFlow (void);			// ...and B-loop
	bool	RCSaAscentFeed (void);			// Are the Ascent Stage-to-A-loop crossfeed valves open?
	bool	RCSbAscentFeed (void);			// What about the B-loop's.
	bool	RCScrossfeed (void);			// Are the A-to-B crossfeed valves open?

	// Thruster Management
	void	ThrusterPreStep (double simt, double simdt);	// Thruster pre-Step process
	void	ThrusterPostStep (double simt, double simdt);	// Thruster post-step processes

	bool	AscentEngineArmed, DescentEngineArmed;	// is there power going to the Ascent and Descent stage engine controls?	
	bool	RCSQuadArmed_A[4], RCSQuadArmed_B[4];	// what about the RCS quads?
	
	// Clocks
	void	UpdateMissionClock (double simt, double simdt, double mjd);
	void	UpdateEventTimer (double simt, double simdt, double mjd);
	void	UpdateValves (double simt, double simdt);


	double	MissionTime, EventTime;
	double	Display_Clock, Display_Timer, Display_Date;
	
	// Flight controls
	void	EngineStart (void);				// 
	void	KillThrust (void);				//

	bool	ThrottleControlIsAuto (void);	// are we letting the autopilot make throttle inputs?
	bool	GimbalEnabled (void);			// is engine gimbaling enabled?
	bool	CmdOvrdOn (void);				// are we overriding the normal throttle controls?
	bool	PilotHasControls (void);		// who's driving? (sets priority of dual controls)
	bool	RCSVernierOn (void);			// is incremental throttleing of RCS on or off?
	bool	OrientYAxis (void);				// where's our referance axis? (Z or Y?)

	// Enviromental systems
	void	ECSPreStep (double simt, double simdt);		// Enviromental pre-Step process
	void	ECSPostStep (double simt, double simdt);	// Envirimental post-step processes
	bool	CabinAtmBreathable ();

	// Electrical systems and lighting
	bool	TrackLightOn (void);			// Is the tracking light on?
	bool	PosLightsOn (void);				// what about the position lights
	bool	PosLightsFlashing (void);		// if they are on are they flashing?
};