// ==============================================================
//                    ORBITER MODULE: LM
//      Part of the Apollo Application Project for Orbiter
//               Copyright (C) 2012 Greg Hlynka
//                    All rights reserved
//
// LM_UMMU.cpp
// Implimentation of UMMU 2.5 for AAPO Lunar Module derived vessels
//
// Notes: 
// ==============================================================

#include "orbitersdk.h"
#include "LM.h"

// --------------------------------------------------------------
// Send hud message
// --------------------------------------------------------------
char *LM::SendHudMessage ()
{
	dHudMessageDelay = 15;
	return cUmmuHudDisplay;
}

// --------------------------------------------------------------
// Determine airlock state
// --------------------------------------------------------------
bool LM::UmmuGetAirlockState ()
{
	if (GetDockStatus (dh_lm_dock) != NULL) // If docked to another vessel...
	{
		return (HatchStatus == CLOSED); // **place holder**
		
		// is cabin pressurized?
		// is top hatch unlocked?
		// if so return true.
		// else return false.
	}
	else // otherwise...
	{
		return (HatchStatus == OPEN); // If EVA hatch is open return true
	}
}

// --------------------------------------------------------------
// Select crew member
// --------------------------------------------------------------
bool LM::UmmuSelectCrewMember ()
{
	// we test there is someone aboard
	if (Crew.GetCrewTotalNumber() <= 0)
	{
		strcpy (SendHudMessage (),"ERROR: No souls on board.");
		return false;
	}

	// we test that we select existing member
	if (iSelectedUmmuMember < Crew.GetCrewTotalNumber () - 1) iSelectedUmmuMember ++;
	else iSelectedUmmuMember = 0;

	char *name = Crew.GetCrewNameBySlotNumber (iSelectedUmmuMember);
	sprintf (SendHudMessage (),"Crew member \"%s\" selected, Press [E] for EVA or Transfer.", name);
	return true;
}

// --------------------------------------------------------------
// Transfer crew member
// --------------------------------------------------------------
bool LM::UmmuTransferCrewMember (int slot)
{
	// Perform transfer
	int ReturnCode = Crew.EvaCrewMember (Crew.GetCrewNameBySlotNumber (slot));
	
	VESSEL *docked;
	OBJHANDLE oh_docked = GetDockStatus (dh_lm_dock);
	if (oh_docked != NULL) docked = oapiGetVesselInterface (oh_docked);

	// Provide feedback to user
	switch (ReturnCode)
	{
	case TRANSFER_TO_DOCKED_SHIP_OK:			
		sprintf(SendHudMessage(),"%s transfered to %s.", Crew.GetLastEvaedCrewName(), docked->GetName());
		iSelectedUmmuMember = 0;
		return true;

	case EVA_OK:
		sprintf(SendHudMessage(),"%s has left the ship.", Crew.GetLastEvaedCrewName());
		oapiSetFocusObject (Crew.GetObjHandleOfLastEVACrew());
		iSelectedUmmuMember = 0;
		return true;

	case ERROR_NO_ONE_ON_BOARD:
		strcpy(SendHudMessage(),"ERROR: No souls on board.");
		return false;

	case ERROR_AIRLOCK_CLOSED:
		strcpy(SendHudMessage(),"ERROR: Airlock is closed, unable to EVA.");
		return false;

	case ERROR_DOCKED_SHIP_HAVE_AIRLOCK_CLOSED:
		strcpy(SendHudMessage(),"ERROR: Docked ship's airlock is closed, unable to transfer");
		return false;

	case ERROR_DOCKED_SHIP_IS_FULL:
		strcpy(SendHudMessage(),"ERROR: Docked ship is already full");
		return false;

	case ERROR_CREW_MEMBER_NOT_FOUND:
		strcpy(SendHudMessage(),"ERROR: Crew member not found");
		return false;

	case ERROR_DOCKEDSHIP_DONOT_USE_UMMU:
		strcpy(SendHudMessage(), "ERROR: Docked ship do not suppport UMmu 2.5");
		return false;

	case ERROR_MISC_ERROR_EVAFAILED:
		strcpy(SendHudMessage(),"ERROR: Misc error with UMmu 2.5");
		return false;

	default:
		return false;
	} // End "switch(ReturnCode)"
}

// --------------------------------------------------------------
// Grapple Cargo
// --------------------------------------------------------------
bool LM::UcgoGrappleCargo (int slot)
{
	int ReturnCode = Cargo.GrappleOneCargo (slot);
	switch(ReturnCode) // for return code list see function "GrappleOneCargo" in the header
	{
		case 1:
			sprintf (SendHudMessage(), "Cargo grappled");
			return true;	
		case 0:
			sprintf (SendHudMessage(), "ERROR: No cargo in range");
			return false;
		case -1:
			sprintf (SendHudMessage(), "ERROR: Cargo exceeds maximum mass");
			return false;
		case -2:
			sprintf (SendHudMessage(), "ERROR: Bad config, mesh not found or slot not declared");
			return false;
		case -3:
			sprintf (SendHudMessage(), "ERROR: Unable to grapple cargo, slot already occupied");
			return false;
		case -4:
			sprintf (SendHudMessage(), "ERROR: Unable to grapple cargo, doors closed ");
			return false;
		case -5:
			sprintf (SendHudMessage(), "ERROR: Unable to grapple cargo, Ship full");
			return false;
		default:
			sprintf (SendHudMessage(), "ERROR: Unable to grapple cargo");
			return false;
	}
}

// --------------------------------------------------------------
// Release Cargo
// --------------------------------------------------------------
bool LM::UcgoReleaseCargo (int slot)
{
	if (Cargo.ReleaseOneCargo (slot))
	{
		sprintf(SendHudMessage(), "Cargo released");
		return true;
	}
	else
	{
		sprintf(SendHudMessage(), "No cargo released");
		return false;
	}
}

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void LM::UmmuSetClassCaps (FILEHANDLE cfg)
{
	// Initialize UMMU (Do this before anything else)
	Crew.InitUmmu (GetHandle());	
	Crew.DefineAirLockShape (true,-1, 1,-1, 1,-1, 1);				
	Crew.SetMembersPosRotOnEVA (ASC_UMMU_HATCHPOS, ASC_UMMU_CREWROT);	
	Crew.SetMaxSeatAvailableInShip (ASC_UMMU_CREWSIZE);
	
	// Add 3 default members for when the ship is spawned by scenario editor (Crew names and ages taken from Apollo 9, first flight of the complete 'CSM + LM' Apollo spacecraft)
	Crew.AddCrewMember ("James", 40, 65, 82, "Capt");
	Crew.AddCrewMember ("Rusty", 34, 65, 82, "Doc");

	// Initialize of selected crew member variable
	iSelectedUmmuMember = 0;	

	// DeclareActionArea
	Crew.DeclareActionArea ( 0, _V(-5, 0, 0), 2.5, true, "action_activated.wav", "You are on the left of the ship");
	Crew.DeclareActionArea ( 1, _V(5, 0, 0), 2.5, true, "action_repaired.wav", "You are on the right of the ship");

	// The Add Mmu without scenery editor variable see PDF doc
	cAddUMmuToVessel[0] = 0;	

	// The HUD display method variables
	cUmmuHudDisplay[0] = 0;		// Initialisation of UMmu hud char variable
	dHudMessageDelay = 0;		// Initialisation of UMmu delay variable

	//UCGO 2.5 Initialisation
	Cargo.Init(GetHandle());
	iSelectedCargo = -1;								// for the selection of cargos (-1 means "default" see header)

	Cargo.DeclareCargoSlot ( 0, _V( 0.71, 0.80, 0.71), _V( 0, 45, 0));	// slot 0
	
	Cargo.SetReleaseSpeedInSpace (0.1);					// release speed of cargo in space in m/s.
	Cargo.SetMaxCargoMassAcceptable (5000.0);			// max cargo mass in kg that your vessel can carry.
	Cargo.SetGrappleDistance (50);						// grapple distance radius in meter from center of ship.
	Cargo.SetGlobalGroundReleasePos (_V( 3, 0, 0));		// global release position on ground.
	Cargo.SetSlotDoorState (false);						// Set global door state for all slots, true = open.
}

// --------------------------------------------------------------
// Post-Step processes
// --------------------------------------------------------------
void LM::UmmuPostStep (double simt, double simdt, double mjd)
{
	// Check CG offset 
	VECTOR3 ofs;
	if (AscentConfig) ofs = _V( 0, 0, 0);
	else ofs = LM_ASC_OFFSET;

	// Detect UMMU Transfers 
	int ReturnCode = Crew.ProcessUniversalMMu ();
	switch (ReturnCode)
	{
	case UMMU_TRANSFERED_TO_OUR_SHIP:
		sprintf (SendHudMessage(),"%s transfered to %s", Crew.GetLastEnteredCrewName(), GetName ());
		break;

	case UMMU_RETURNED_TO_OUR_SHIP:
		sprintf (SendHudMessage (),"%s has entered %s", Crew.GetLastEnteredCrewName (), GetName ());
		break;
	} // End "switch (ReturnCode)"

	// Detect action area activation
	int ActionArea = Crew.DetectActionAreaActivated ();
	switch (ActionArea)
	{
	case 0:
		// Do something cool
		break;

	case 1:
		// Do something cool
		break;
	} // End "switch (ReturnCode)"

	// Update vessel parameters
	VECTOR3 hatchpos = ASC_UMMU_HATCHPOS + ofs;

	if ((GroundContact ()) && (DescentStage_attached)) // If Descent Stage is attached and ground contact is true, position UMMU airlock at foot of EVA ladder.
	{
		Crew.SetMembersPosRotOnEVA (LM_UMMU_LADDERPOS, _V( 0, RAD*180, 0));
		Crew.DefineAirLockShape (UmmuGetAirlockState (), LM_UMMU_LADDERPOS.x - 1, LM_UMMU_LADDERPOS.x + 1, LM_UMMU_LADDERPOS.y - 1, LM_UMMU_LADDERPOS.y + 3, LM_UMMU_LADDERPOS.z - 2, LM_UMMU_LADDERPOS.z + 4);
	}
	else // Otherwise position it at fwd hatch
	{
		Crew.SetMembersPosRotOnEVA (hatchpos, ASC_UMMU_CREWROT);
		Crew.DefineAirLockShape (UmmuGetAirlockState (), hatchpos.x - 1, hatchpos.x + 1, hatchpos.y - 1, hatchpos.y + 2, hatchpos.z - 1, hatchpos.z + 2);
	}
	// The Add Mmu without scenery editor process see PDF doc
	AddUMmuToVessel();

	// Display user messages
	if (dHudMessageDelay > 0) dHudMessageDelay -= simdt;
	else dHudMessageDelay = 0;

	if (oapiGetFocusInterface () == this) // Confrim that vessel is current focus object to prevent conflicts
	{
		if ((oapiCameraInternal () == false) || (oapiGetHUDMode () == HUD_NONE)) // If camera mode is external or no hud is available display user messages using debug string
		{
			if (dHudMessageDelay != 0) sprintf(oapiDebugString(), "%s", cUmmuHudDisplay); 
			else sprintf (oapiDebugString(),""); 
		}
	} // End "if ((oapiGetFocusInterface () == this)"

	// At the very end of clbkPostStep or clbkPreStep
	Crew.WarnUserUMMUNotInstalled ("AAPO");
	Cargo.WarnUserUCGONotInstalled("AAPO");
}

//-------------------------------------------------------------------------
// THIS IS FOR ADDING CREW SEE PDF doc "Allow user to add crew to your ship
// without scenery editor"
bool UMmuCrewAddCallback (void *id, char *str, void *data)
{
	if ((strlen(str) < 2) || (strlen(str)>38)) return false;

	char *cPtr = (char*)data;
	
	if(*cPtr==2)
	{
		*cPtr=3;
		strcpy(cPtr+2,str);
	}
	
	else if(*cPtr==4)
	{
		*cPtr=5;
		strcpy (cPtr+42,str);
	}
	
	else if(*cPtr==6)
	{
		*cPtr=7;
		strcpy(cPtr+82,str);
	}
	
	return true;
}

void LM::AddUMmuToVessel(BOOL bStartAdding)
{
	if ((bStartAdding == false) && (cAddUMmuToVessel[0] == 0)) return;

	if (bStartAdding == true)
	{
		int salut = sizeof(cAddUMmuToVessel);
		memset (cAddUMmuToVessel, 0, sizeof(cAddUMmuToVessel));
		cAddUMmuToVessel[0] = 1;
	}
	else if (cAddUMmuToVessel[0] == 1)
	{
		cAddUMmuToVessel[0] = 2;
		oapiOpenInputBox ("Enter name for new crew member, or press [Esc] to cancel.", UMmuCrewAddCallback, 0, 30, (void*)cAddUMmuToVessel);
	}
	else if (cAddUMmuToVessel[0] == 3)
	{
		cAddUMmuToVessel[0] = 4;
		oapiOpenInputBox ("Enter crew member's age.", UMmuCrewAddCallback, 0, 30, (void*)cAddUMmuToVessel);
	}
	else if(cAddUMmuToVessel[0] == 5)
	{
		cAddUMmuToVessel[0] = 6;
		oapiOpenInputBox ("Enter crew member's role. (Capt, Sec, Vip, Sci, Doc, Tech, Crew, Pax)", UMmuCrewAddCallback, 0, 30, (void*)cAddUMmuToVessel);
	}
	else if(cAddUMmuToVessel[0] == 7)
	{
		cAddUMmuToVessel[0] = 0;
		int Age = max(5, min(100, atoi(&cAddUMmuToVessel[42])));
		
		if (Crew.AddCrewMember (&cAddUMmuToVessel[2], Age, 70, 68, &cAddUMmuToVessel[82]) == true)
		{
			sprintf (SendHudMessage (),"Crew \"%s\" aged %i added to vessel.",&cAddUMmuToVessel[2],Age);
		}
		else
		{
			strcpy(SendHudMessage(),"ERROR: Unable to add crew member.");
		}
	}
}
