// ==============================================================
//                 ORBITER MODULE: LUT
//                  Part of the ORBITER SDK
//          Copyright (C) 2002-2004 Martin Schweiger
//                   All rights reserved
//
// LUT.cpp
// Control module for LUT vessel class
//
// Notes:
// This is an example for a "minimal" vessel implementation which
// only overloads the clbkSetClassCaps method to define vessel
// capabilities and otherwise uses the default VESSEL class
// behaviour.
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include "AAPO_Common.h"

// ==============================================================
// Some vessel parameters
// ==============================================================
const double	LUT_SIZE			= 40;					// mean radius [m]
const VECTOR3	LUT_CS				= { 10.5, 15.0, 5.8};	// x,y,z cross sections [m^2]
const VECTOR3	LUT_PMI				= { 2.28, 2.31, 0.79};	// principal moments of inertia (mass-normalised) [m^2]
const VECTOR3	LUT_RD				= { 1.00, 1.00, 1.00};	//{0.05,0.1,0.05};  // rotation drag coefficients
const double	LUT_MASS			= 5.0e6;				// LUT mass [kg]
const double	LUT_FUELMASS		= 750.0;				// max fuel mass [kg]
const double	LUT_ISP				= 5e4;					// fuel-specific impulse [m/s]
const VECTOR3	LUT_TDP[3]			= {{ 0,-65, 20}, {-10,-65,-15}, { 10,-65,-15}}; // touchdown points [m]

const VECTOR3	LUT_CAMERAPOS		= {-17.5, 47.0, 0.0};
const VECTOR3	LUT_MLP_OFFSET		= { 0, 14, 0};
const VECTOR3	LUT_DOCK_POS		= { 0, 1.3,-1};      // docking port location [m]
const VECTOR3	LUT_DOCK_DIR		= { 0, 1, 0};         // docking port approach direction
const VECTOR3	LUT_DOCK_ROT		= { 0, 0,-1};        // docking port alignment direction

// ==============================================================
// Launch Tower class interface
// ==============================================================

class LUT: public VESSEL3 {
public:
	LUT (OBJHANDLE hVessel, int flightmodel);
	~LUT ();
	void clbkSetClassCaps (FILEHANDLE cfg);
	bool clbkLoadGenericCockpit (void);

private:
	MESHHANDLE mh_LC, mh_LUT;
};

LUT::LUT (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
}

LUT::~LUT ()
{
}

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void LUT::clbkSetClassCaps (FILEHANDLE cfg)
{
	// physical vessel parameters
	SetSize (LUT_SIZE);
	SetEmptyMass (LUT_MASS);
	SetPMI (LUT_PMI);
	SetCrossSections (LUT_CS);
	SetRotDrag (LUT_RD);
	SetTouchdownPoints (LUT_TDP[0], LUT_TDP[1], LUT_TDP[2]);

	// docking port definitions
	SetDockParams (LUT_DOCK_POS, LUT_DOCK_DIR, LUT_DOCK_ROT);

	// associate a mesh for the visual
	mh_LC		= oapiLoadMeshGlobal ("AAPO/LC39");	
	mh_LUT		= oapiLoadMeshGlobal ("AAPO/LaunchTower");

	//AddMesh (mh_LC, &_V( 0,-65.5, 0));		
	AddMesh (mh_LUT, &LUT_MLP_OFFSET);
	SetMeshVisibilityMode (0, MESHVIS_ALWAYS);
}

bool LUT::clbkLoadGenericCockpit (void)
{
	// camera parameters
	SetCameraOffset (LUT_CAMERAPOS + LUT_MLP_OFFSET);
	SetCameraDefaultDirection (_V( 1, 0, 0));

	return true;
}
// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new LUT (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (LUT*)vessel;
}
