// ==============================================================
//                ORBITER MODULE: LM Descent
//                  Part of the ORBITER SDK
//               Copyright (C) 2012 Greg Hlynka
//                    All rights reserved
//
// CSM_SLAStage.cpp
// Control module for LM Descent Stage vessel class
//
// Notes:
// ==============================================================

#include "SLA.h"

// ==============================================================
// Launch Adapter vessel class Interface
// ==============================================================

CSM_SLA::CSM_SLA (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
	DefineAnimations();
	DefineSounds ();

	SLA_proc = 0;
	SLA_Status = CLOSED;

	// Load Mesh
	mh_adapter = oapiLoadMeshGlobal ("AAPO/LaunchAdapter");	
}

CSM_SLA::~CSM_SLA ()
{
}

// ==============================================================
// Custom vessel functions
// ==============================================================

// --------------------------------------------------------------
// Define Animations
// --------------------------------------------------------------
void CSM_SLA::DefineAnimations(void)
{
	// Mesh groups
	static UINT meshgroups_PanelA[3] = {GRP_SLA_Exterior_PanelA, GRP_SLA_Greebles_PanelA, GRP_SLA_Interior_PanelA};
	static UINT meshgroups_PanelB[3] = {GRP_SLA_Exterior_PanelB, GRP_SLA_Greebles_PanelB, GRP_SLA_Interior_PanelB};
	static UINT meshgroups_PanelC[3] = {GRP_SLA_Exterior_PanelC, GRP_SLA_Greebles_PanelC, GRP_SLA_Interior_PanelC};
	static UINT meshgroups_PanelD[3] = {GRP_SLA_Exterior_PanelD, GRP_SLA_Greebles_PanelD, GRP_SLA_Interior_PanelD};

	static MGROUP_ROTATE	mgt_PanelA ( 0, meshgroups_PanelA, 3, _V( 0.00, 2.98,-2.27), _V(-1.0, 0.0, 0.0), (float) 49*RAD); 
	static MGROUP_ROTATE	mgt_PanelB ( 0, meshgroups_PanelB, 3, _V(-2.98, 0.00,-2.27), _V( 0.0,-1.0, 0.0), (float) 52*RAD);
	static MGROUP_ROTATE	mgt_PanelC ( 0, meshgroups_PanelC, 3, _V( 0.00,-2.98,-2.27), _V( 1.0, 0.0, 0.0), (float) 61*RAD);
	static MGROUP_ROTATE	mgt_PanelD ( 0, meshgroups_PanelD, 3, _V( 2.98, 0.00,-2.27), _V( 0.0, 1.0, 0.0), (float) 45*RAD);

	anim_SLA = CreateAnimation (0.0);
	AddAnimationComponent (anim_SLA, 0, 1, &mgt_PanelA);
	AddAnimationComponent (anim_SLA, 0, 1, &mgt_PanelB);
	AddAnimationComponent (anim_SLA, 0, 1, &mgt_PanelC);
	AddAnimationComponent (anim_SLA, 0, 1, &mgt_PanelD);

} // end "DefineAnimations(void)"

// --------------------------------------------------------------
// Define sound effects
// --------------------------------------------------------------
void CSM_SLA::DefineSounds (void)
{
	// OrbiterSound 4.0
	OrbiterSoundID = ConnectToOrbiterSoundDLL (GetHandle());			// Init sound
	SoundOptionOnOff (OrbiterSoundID, PLAYCOUNTDOWNWHENTAKEOFF, FALSE);	// Disable the stock countdown... it's annoying :)

	LoadSoundFiles (OrbiterSoundID);
} // End "CSM_SLA::DefineSounds"

// --------------------------------------------------------------
// Calculate current empty mass of vessel
// --------------------------------------------------------------
double CSM_SLA::CalcEmptyMass (void)
{
	double		output	= SLA_MASS;					// If SLA is attached add it's mass to total
	OBJHANDLE	pyld	= GetAttachmentStatus (ah_payload);	// Get handle of vessel attached to SLA's Lunar Module attachment point 

	if (pyld != NULL)	// If vessel is attached, add it's mass to total
	{
		VESSEL *lm = oapiGetVesselInterface (pyld);
		output += lm->GetMass ();
	}

	return output;	// Return total
} // End "CSM::CalcEmptyMass"

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set capabilities of vessel
// --------------------------------------------------------------
void CSM_SLA::clbkSetClassCaps (FILEHANDLE cfg)
{
	// physical vessel parameters
	SetSize (SLA_SIZE);
	SetCrossSections (SLA_CS);
	SetPMI (SLA_PMI);
	SetRotDrag (SLA_RD);
	SetEmptyMass (SLA_MASS);

	// Attachment point definitions
	ah_booster		= CreateAttachment (true, _V( 0.0, 0.0,-4.25), _V( 0, 0,-1), _V( 0, 1, 0), "X");	// for attachment to launch vehicle
	ah_payload		= CreateAttachment (false, _V( 0.0, 0.0,-2.28), _V( 0, 0, 1), _V( 0, 1, 0), "X");	// for attachment of lunar module to SLA
	ah_SM			= CreateAttachment (false, _V( 0.0, 0.0, 4.25), _V( 0, 0, 1), _V( 0, 1, 0), "X");	// for attachment of service module to SLA

	// associate a mesh for the visual
	AddMesh (mh_adapter);	
	SetMeshVisibilityMode (0, MESHVIS_ALWAYS);	

	// camera parameters
	SetCameraOffset (_V( 0.0, 3.45, 0.0));
}

// --------------------------------------------------------------
// Read status from scenario file
// --------------------------------------------------------------
void CSM_SLA::clbkLoadStateEx (FILEHANDLE scn, void *status)
{
	int i = 0;
	char *cbuf; 

	while (oapiReadScenario_nextline (scn, cbuf)) 
	{
		// Load animation states
		if (!_strnicmp (cbuf, "SLA", 3)) // find line labeled "SLA" in scn file
		{
			sscanf (cbuf+3, "%i %lf", &SLA_Status, &SLA_proc);	// Read values stored there to SLA_Status andSLA_proc
			SetAnimation (anim_SLA, SLA_proc);					// Apply process value to animation.
		}

		// Load default parameters
		else ParseScenarioLineEx (cbuf, status);	

	} // End "while (oapiReadScenario_nextline (scn, cbuf))"

} // End "clbkLoadStateEx(FILEHANDLE scn, void *status)"

// --------------------------------------------------------------
// Write status to scenario file
// --------------------------------------------------------------
void CSM_SLA::clbkSaveState (FILEHANDLE scn)
{
	int i = 0;
	char cbuf[256];
	
	// Write default vessel parameters to file
	VESSEL3::clbkSaveState (scn);

	// Write custom parameters to file
	sprintf (cbuf, "%i %0.4f", SLA_Status, SLA_proc);		// Landing Gear status and animation state
	oapiWriteScenario_string (scn, "SLA", cbuf);

} // End "clbkSaveState(FILEHANDLE scn)"

// --------------------------------------------------------------
// Post-Step processes and Animations
// --------------------------------------------------------------
void CSM_SLA::clbkPostStep (double simt, double simdt, double mjd)
{
	SetEmptyMass (CalcEmptyMass ());

	// Animation control logic
	if (SLA_proc < 0)			// If process value is less than 0...
	{
		SLA_Status = CLOSED;	// ...set status to "CLOSED"
		SLA_proc = 0;			// and process value to 0
	}

	else if (SLA_proc > 1)		// If process value is greater than 1...
	{
		SLA_Status = OPEN;		// ...set status to "OPEN"
		SLA_proc = 1;			// and process value to 1
	}

	if (SLA_Status > CLOSED)	
	{
		double	delta = simdt / 5;	// delta = simdt divided by duration of animation in seconds

		if (SLA_Status == OPENING)		// if Status equals "OPENING"...
		{
			SLA_proc += delta;			// ...add delta to process value
		}
		
		if (SLA_Status == CLOSING)		// if Status equals "CLOSING"...
		{
			SLA_proc -= delta;			// ...subtract it.
		}
	}
	SetAnimation (anim_SLA, SLA_proc);	// apply process value to animation

	if ((GetAttachmentStatus (ah_SM) == NULL) && (SLA_Status != OPEN)) SLA_Status = OPENING;	// If SM is not attached and panels are not open, open them

	//// Debuggery
	//sprintf(oapiDebugString(), "Landing Gear Status %0.0f,SLA_proc %0.3f", (float)SLA_Status,SLA_proc);

	// Clean up
	OBJHANDLE self = GetHandle ();					// Get vessel handle 

	if (GroundContact ()) oapiDeleteVessel(self);	// If vessel touches the ground delete it 

	double drag = GetDrag () / GetMass ();			// Get decceleration due to drag
	if (drag > 8 * G) oapiDeleteVessel (self);		// if decceleration is greater than 8 Gs burn up
}

// --------------------------------------------------------------
// Process keyboard inputs
// --------------------------------------------------------------
int  CSM_SLA::clbkConsumeBufferedKey (DWORD key, bool down, char *kstate)
{
	// Detach payload when [J] is pressed.
	if(key == OAPI_KEY_J  && down && !KEYMOD_SHIFT(kstate) && !KEYMOD_CONTROL (kstate) && !KEYMOD_ALT(kstate))
	{
		DetachChild (ah_payload, 0.05); 
		return 1;
	}

	// Seperate from booster when [Alt] + [S] is pressed.
	if (key == OAPI_KEY_S && down && !KEYMOD_SHIFT(kstate) && !KEYMOD_CONTROL (kstate) && KEYMOD_ALT(kstate))
	{ 
		int idx = 0;											// attachment index
		OBJHANDLE oh_LV	= GetAttachmentStatus (ah_booster);		// object handle of launch vehicle
		
		if (oh_LV != NULL)	// check if launch vehicle handle is valid, if it is...
		{
			VESSEL *lv = oapiGetVesselInterface (oh_LV);	// get VESSEL interface for launch vehicle
			int attcount = lv->AttachmentCount (false);		// get number of attachment points

			for (int i = 0; i < attcount; i++) if (lv->GetAttachmentStatus (lv->GetAttachmentHandle (false, i)) == GetHandle ()) idx = i; // find point that we are currently attached to
			
			lv->DetachChild (lv->GetAttachmentHandle (false, idx), 0.1);	// dettach us
		}
		return 1;
	}

	return 0; // if no keys are pressed the function returns '0' and nothing happens
}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new CSM_SLA (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (CSM_SLA*)vessel;
}