// ==============================================================
//              Orbiter Module: SM
//    part of the Apollo Aplications Prject for Orbiter (AAPO)
//               copyright (c) 2012 greg hlynka
//                    all rights reserved
//
// SM.cpp
// control module for Apollo Service Module vessel class
//
// notes:
// ==============================================================

#include "SM.h"

// Mesh resource files
#include "mesh_SM.h"

// ==============================================================
// Class Constructor and Destructor
// ==============================================================

SM::SM (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
	// Config file difficulty modifiers (1.0 = 100% of default value)
	thrust_mod		= 1.0;
	dv_mod			= 1.0;
	consumables_mod = 1.0;

	// Set initial values
	AntennaStatus	= OPEN;
	antenna_proc	= 1.0;

	LightStatus		= OPEN;
	antenna_proc	= 1.0;

	UmbilicalStatus	= OPEN;
	umbilical_proc	= 1.0;

	gimbalx_proc	= 0.5;
	gimbaly_proc	= 0.5;

	sep_bt			= 0.0;

	// Load Mesh
	mh_sbaycover	= oapiLoadMeshGlobal ("AAPO/SM_BayCover");		
	mh_service		= oapiLoadMeshGlobal ("AAPO/SM");
}

SM::~SM ()
{
}

// ==============================================================
// Custom vessel functions
// ==============================================================

// --------------------------------------------------------------
// Define Animations
// --------------------------------------------------------------
void SM::DefineAnimations (UINT mesh)
{
	// Umbilicle
	static UINT				meshgroup_umbilical = GRP_SM_Umbilical;
	static MGROUP_ROTATE	mgt_umbilical (mesh, &meshgroup_umbilical, 1, SM_UMBILICAL_PIVOT, _V( 1.0, 0.0, 0.0), (float) 35*RAD); 

	anim_umbilical = CreateAnimation (1.0);
	AddAnimationComponent (anim_umbilical, 0.0f, 1.0f, &mgt_umbilical);	
	SetAnimation (anim_umbilical, umbilical_proc);

	// Eva Light
	static UINT				meshgroup_light = GRP_SM_EVA_Light;
	static MGROUP_ROTATE	mgt_light (mesh, &meshgroup_light, 1, SM_LIGHT_PIVOT, _V( sin(SM_LIGHT_CANT), cos(SM_LIGHT_CANT), 0.0), (float)-101*RAD); 

	anim_light = CreateAnimation (1.0);
	AddAnimationComponent (anim_light, 0.0f, 1.0f, &mgt_light);	
	SetAnimation (anim_light, light_proc);

	// Engine Gimbal
	static UINT				meshgroup_engine[2] = { GRP_SM_SPS_Nozzle, GRP_SM_SPS_NozzleExtention};
	static MGROUP_ROTATE	mgt_gimbalx (mesh, meshgroup_engine, 2, SM_ENGINEPOS, _V( 1.0, 0.0, 0.0), (float) 12*RAD); 
	static MGROUP_ROTATE	mgt_gimbaly (mesh, meshgroup_engine, 2, SM_ENGINEPOS, _V( 0.0,-1.0, 0.0), (float) 12*RAD); 

	anim_gimbalx = CreateAnimation (0.5);
	AddAnimationComponent (anim_gimbalx, 0.0f, 1.0f, &mgt_gimbalx);

	anim_gimbaly = CreateAnimation (0.5);
	AddAnimationComponent (anim_gimbaly, 0.0f, 1.0f, &mgt_gimbaly);

	// High Gain Antenna 
	static UINT				meshgroup_HGAstrut = GRP_SM_HGA_Strut;
	static MGROUP_ROTATE	mgt_antenna (mesh, &meshgroup_HGAstrut, 1, SM_HGA_MOUNT, _V( sin(SM_HGA_CANT), cos(SM_HGA_CANT), 0.0), (float) 80*RAD); 

	anim_antenna = CreateAnimation (1.0);
	ach_antenna = AddAnimationComponent (anim_antenna, 0.0f, 1.0f, &mgt_antenna);
	SetAnimation (anim_antenna, antenna_proc);

	static UINT				meshgroup_HGApivot = GRP_SM_HGA_Pivot;
	static MGROUP_ROTATE	mgt_HGAyaw (mesh, &meshgroup_HGApivot, 1, SM_HGA_PIVOT, _V( cos(SM_HGA_CANT), sin(SM_HGA_CANT), 0.0), (float) 360*RAD); 
	
	anim_HGAyaw = CreateAnimation (0.0);
	ach_HGAyaw = AddAnimationComponent (anim_HGAyaw, 0.0f, 1.0f, &mgt_HGAyaw, ach_antenna);

	static UINT				meshgroup_HGAntenna[5] = {GRP_SM_HGA_Antennas, GRP_SM_HGA_Body, GRP_SM_HGA_Reflectors};
	static MGROUP_ROTATE	mgt_HGApitch (mesh, meshgroup_HGAntenna, 3, SM_HGA_PIVOT, _V( cos(SM_HGA_CANT), sin(SM_HGA_CANT), 0.0), (float) 90*RAD); 
	
	anim_HGApitch = CreateAnimation (1.0);
	ach_HGApitch = AddAnimationComponent (anim_HGApitch, 0.0f, 1.0f, &mgt_HGApitch, ach_antenna);
	
} // end "DefineAnimations(void)"

// --------------------------------------------------------------
// Define exhaust and particle streams
// --------------------------------------------------------------
void SM::DefineParticleStreams (void)
{
	// Load exhaust textures
	SURFHANDLE tex_main = oapiRegisterExhaustTexture ("Exhaust2");
	SURFHANDLE tex_rcs = oapiRegisterExhaustTexture ("Exhaust_atrcs");	// Load RCS exhaust texture

	// Add main engine exhaust
	AddExhaust (th_sps, 4.0*thrust_mod, 1, 3.0, tex_main);				// Service module propulsion system exhaust [length*thrust_mod, width*thrust_mod, offset]
	
	// Add rcs exhaust
	for (int i = 0; i < 16; i++)
	{
		AddExhaust (th_sm_rcs[i], 3.0, 0.2, 0.3, tex_rcs);
	}

} // End "SM::DefineParticleStreams"

// --------------------------------------------------------------
// Define sound effects
// --------------------------------------------------------------
void SM::DefineSounds (void)
{

} // End "SM::DefineSounds"

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set capabilities of vessel
// --------------------------------------------------------------
void SM::clbkSetClassCaps (FILEHANDLE cfg)
{
	int i = 0;

	// Physical parameters
	SetSize (SM_SIZE);
	SetEmptyMass (SM_EMPTYMASS);
	SetPMI (SM_PMI);
	SetCrossSections (SM_CS);
	SetRotDrag (SM_RD);
	SetSurfaceFrictionCoeff (1, 1);

	// Aerodynamic properties
	SetCW (SM_CW[0], SM_CW[1], SM_CW[2], SM_CW[3]);

	// docking port definitions
	SetDockParams (SM_DOCK_POS, SM_DOCK_DIR, SM_DOCK_ROT);

	// Attachment point definitions
	ah_payload = CreateAttachment (false, SM_DOCK_POS, SM_DOCK_DIR, SM_DOCK_ROT, "X");		// for attachment of a payload when using SM as independant upper stage 
	ah_booster = CreateAttachment (true, SM_ATTACH_POS, SM_ATTACH_DIR, SM_ATTACH_ROT, "X");	// for attachment of SM to launch vehicle or SLA

	// propellant resources
	ph_sm_O2 = CreatePropellantResource (SM_O2MASS);
	ph_sm_H2 = CreatePropellantResource (SM_H2MASS);
	ph_sps	= CreatePropellantResource (SM_FUELMASS);
	for (int i = 0; i < 4; i ++) ph_sm_rcs[i] = CreatePropellantResource (SM_RCS_FUELMASS);

	// Main engine
	th_sps	= CreateThruster (SM_ENGINEPOS, _V( 0, 0, 1), SM_MAXTHRUST, ph_sps, SM_ISP);	

	// Service Module RCS engines
	double	alpha	= RAD * 10;
	double	beta	= RAD * 9.25;
	VECTOR3	shift	= _V( 0.0, 0.0, 0.137);

	// Service module Quadrant A (12 o'clock)
	th_sm_rcs[ 0]	= CreateThruster (SM_RCS_ENGINEPOS[0] - shift, RotateVector (_V( 0,-sin(alpha), cos(alpha)), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[0], SM_RCS_ISP * dv_mod);	// Fwd
	th_sm_rcs[ 1]	= CreateThruster (SM_RCS_ENGINEPOS[0] + shift, RotateVector (_V( 0,-sin(alpha),-cos(alpha)), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[0], SM_RCS_ISP * dv_mod);	// Back
	th_sm_rcs[ 2]	= CreateThruster (SM_RCS_ENGINEPOS[0] - shift, RotateVector (_V( cos(alpha),-sin(alpha), 0), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[0], SM_RCS_ISP * dv_mod);	// Right
	th_sm_rcs[ 3]	= CreateThruster (SM_RCS_ENGINEPOS[0] + shift, RotateVector (_V(-cos(alpha),-sin(alpha), 0), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[0], SM_RCS_ISP * dv_mod);	// Left
	
	// Service module  Quadrant B (3 o'clock)
	th_sm_rcs[ 4]	= CreateThruster (SM_RCS_ENGINEPOS[1] - shift, RotateVector (_V( sin(alpha), 0, cos(alpha)), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[1], SM_RCS_ISP * dv_mod);	// Fwd
	th_sm_rcs[ 5]	= CreateThruster (SM_RCS_ENGINEPOS[1] + shift, RotateVector (_V( sin(alpha), 0,-cos(alpha)), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[1], SM_RCS_ISP * dv_mod);	// Back
	th_sm_rcs[ 6]	= CreateThruster (SM_RCS_ENGINEPOS[1] - shift, RotateVector (_V( sin(alpha), cos(alpha), 0), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[1], SM_RCS_ISP * dv_mod);	// Up
	th_sm_rcs[ 7]	= CreateThruster (SM_RCS_ENGINEPOS[1] + shift, RotateVector (_V( sin(alpha),-cos(alpha), 0), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[1], SM_RCS_ISP * dv_mod);	// Down

	// Service module  Quadrant C (6 o'clock)
	th_sm_rcs[ 8]	= CreateThruster (SM_RCS_ENGINEPOS[2] - shift, RotateVector (_V( 0, sin(alpha), cos(alpha)), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[2], SM_RCS_ISP * dv_mod);	// Fwd
	th_sm_rcs[ 9]	= CreateThruster (SM_RCS_ENGINEPOS[2] + shift, RotateVector (_V( 0, sin(alpha),-cos(alpha)), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[2], SM_RCS_ISP * dv_mod);	// Back
	th_sm_rcs[10]	= CreateThruster (SM_RCS_ENGINEPOS[2] + shift, RotateVector (_V( cos(alpha), sin(alpha), 0), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[2], SM_RCS_ISP * dv_mod);	// Right
	th_sm_rcs[11]	= CreateThruster (SM_RCS_ENGINEPOS[2] - shift, RotateVector (_V(-cos(alpha), sin(alpha), 0), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[2], SM_RCS_ISP * dv_mod);	// Left

	// Service module  Quadrant D (9 o'clock)
	th_sm_rcs[12]	= CreateThruster (SM_RCS_ENGINEPOS[3] - shift, RotateVector (_V(-sin(alpha), 0, cos(alpha)), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[3], SM_RCS_ISP * dv_mod);	// Fwd
	th_sm_rcs[13]	= CreateThruster (SM_RCS_ENGINEPOS[3] + shift, RotateVector (_V(-sin(alpha), 0,-cos(alpha)), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[3], SM_RCS_ISP * dv_mod);	// Back
	th_sm_rcs[14]	= CreateThruster (SM_RCS_ENGINEPOS[3] + shift, RotateVector (_V(-sin(alpha), cos(alpha), 0), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[3], SM_RCS_ISP * dv_mod);	// Up
	th_sm_rcs[15]	= CreateThruster (SM_RCS_ENGINEPOS[3] - shift, RotateVector (_V(-sin(alpha),-cos(alpha), 0), beta, _V(0,0,1)), SM_RCS_MAXTHRUST, ph_sm_rcs[3], SM_RCS_ISP * dv_mod);	// Down
	
	// Associate mesh for the visual
	mesh_SM = AddMesh (mh_service);	
	SetMeshVisibilityMode (0, MESHVIS_ALWAYS);	

	// camera parameters
	SetCameraOffset (_V( 0.32, 2.16, 0.86));
}

// --------------------------------------------------------------
// Read status from scenario file
// --------------------------------------------------------------
void SM::clbkLoadStateEx (FILEHANDLE scn, void *status)
{
	int i = 0;
	char *cbuf; 

	while (oapiReadScenario_nextline (scn, cbuf)) 
	{

		// Is SM bay open?
		if (!_strnicmp( cbuf, "BAYOPEN", 7))
		{
			int temp = 0;					 
			sscanf (cbuf + 7, "%i", &temp);
			BayOpen = (temp != 0);
			
			if (!BayOpen) AddMesh (mh_sbaycover, &SM_SIMBAY_OFFSET);	// Add SIM Bay cover mesh
		}

		// Load animation states
		else if (!_strnicmp (cbuf, "HGA", 4)) // find line labeled "GEAR" in scn file
		{
			sscanf (cbuf+4, "%i %lf", &AntennaStatus, &antenna_proc);	// Read values stored there to GearStatus and gear_proc
			SetAnimation (anim_antenna, antenna_proc);					// Apply process value to animation.
		}

		// Load default parameters
		else ParseScenarioLineEx (cbuf, status);	

	} // End "while (oapiReadScenario_nextline (scn, cbuf))"

} // End "clbkLoadStateEx(FILEHANDLE scn, void *status)"

// --------------------------------------------------------------
// Write status to scenario file
// --------------------------------------------------------------
void SM::clbkSaveState (FILEHANDLE scn)
{
	int i = 0;
	char cbuf[256];
	
	// Write default vessel parameters to file
	VESSEL3::clbkSaveState (scn);

	// Bay Open?
	sprintf (cbuf, "%i", BayOpen);						
	oapiWriteScenario_string (scn, "BAYOPEN", cbuf);

	// Antenna status and animation state
	sprintf (cbuf, "%i %0.4f", AntennaStatus, antenna_proc);		
	oapiWriteScenario_string (scn, "HGA", cbuf);

} // End "clbkSaveState(FILEHANDLE scn)"

// --------------------------------------------------------------
// Finalise vessel creation
// --------------------------------------------------------------
void SM::clbkPostCreation ()
{
	DefineAnimations (mesh_SM);
	DefineParticleStreams ();

	SetDefaultPropellantResource (ph_sps);		// Set main fuel tank as default propellant resource
	SetISP (GetThrusterIsp0 (th_sps));			// Set main engine ISP as default
	
	// Assign thruster control groups
	THRUSTER_HANDLE	th_group[4];

	// RCS control groups (Rotation)
	th_group[0]	= th_sm_rcs[1];
	th_group[1] = th_sm_rcs[8];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHUP);

	th_group[0]	= th_sm_rcs[0];
	th_group[1] = th_sm_rcs[9];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_PITCHDOWN);

	th_group[0]	= th_sm_rcs[7];
	th_group[1] = th_sm_rcs[14];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKLEFT);

	th_group[0]	= th_sm_rcs[2];
	th_group[1] = th_sm_rcs[11];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_BANKRIGHT);

	th_group[0]	= th_sm_rcs[4];
	th_group[1] = th_sm_rcs[13];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0]	= th_sm_rcs[5];
	th_group[1] = th_sm_rcs[12];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);

	// RCS control groups (Translation)
	th_group[0] = th_sm_rcs[6];
	th_group[1] = th_sm_rcs[14];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_UP);

	th_group[0] = th_sm_rcs[7];
	th_group[1] = th_sm_rcs[15];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_DOWN);

	th_group[0] = th_sm_rcs[2];
	th_group[1] = th_sm_rcs[10];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_RIGHT);

	th_group[0] = th_sm_rcs[3];
	th_group[1] = th_sm_rcs[11];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_LEFT);
		
	th_group[0] = th_sm_rcs[0];
	th_group[1] = th_sm_rcs[4];
	th_group[2] = th_sm_rcs[8];
	th_group[3] = th_sm_rcs[12];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_FORWARD);

	th_group[0] = th_sm_rcs[1];
	th_group[1] = th_sm_rcs[5];
	th_group[2] = th_sm_rcs[9];
	th_group[3] = th_sm_rcs[13];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_BACK);

	// Main engine
	CreateThrusterGroup (&th_sps, 1, THGROUP_MAIN);

} // End "SM::clbkPostCreation ()"

// --------------------------------------------------------------
// Post-Step processes and Animations
// --------------------------------------------------------------
void SM::clbkPostStep (double simt, double simdt, double mjd)
{
	// Update Bay status
	if (GetMeshCount () > 1) BayOpen = false;
	else  BayOpen = true;

	// Set timer for seperation burn
	if (GetThrusterGroupLevel (THGROUP_ATT_BACK) > 0) sep_bt += simdt;
	else sep_bt = 0;

	if (sep_bt > 25)
	{
		for (int i = 0; i < 16; i++) SetThrusterLevel (th_sm_rcs[i], 0);
	}

	// Engine Gimbal and CG balancing
	VECTOR3 cgdir, cgpos, enginepos;
	GetSuperstructureCG (cgpos);														// The position of combined CG of the stage and any docked modules in vessel frame
	GetThrusterRef (th_sps, enginepos);													// The position of main engine in vessel frame
	cgdir = _V(cgpos.x - enginepos.x, cgpos.y - enginepos.y, cgpos.z - enginepos.z);	// Get the vector between them...  
	VECTOR3 thrustdir = -cgdir;
	normalise (thrustdir);		

	// Debuggery
	//sprintf (oapiDebugString (), "X %f, Y %f, Z %f", cgdir.x, cgdir.y, cgdir.z);
	//sprintf (oapiDebugString (), "X %f, Y %f, Z %f", thrustdir.x, thrustdir.y, thrustdir.z);

	// Process attitude inputs
	double pitchinput, yawinput;
		
	pitchinput = GetThrusterGroupLevel (THGROUP_ATT_PITCHUP) - GetThrusterGroupLevel (THGROUP_ATT_PITCHDOWN);
	thrustdir.y += pitchinput * sin (RAD * 3);

	yawinput = GetThrusterGroupLevel (THGROUP_ATT_YAWRIGHT) - GetThrusterGroupLevel (THGROUP_ATT_YAWLEFT);
	thrustdir.x += yawinput * sin (RAD * 3);

	//sprintf (oapiDebugString (), "Pitch %f, Yaw %f", pitchinput, yawinput);

	// Restrict X and Y to valid range
	if (thrustdir.y > sin (SM_GIMBAL_RANGE)) thrustdir.y = sin (SM_GIMBAL_RANGE);  
	else if (thrustdir.y < sin (-SM_GIMBAL_RANGE)) thrustdir.y = sin (-SM_GIMBAL_RANGE);  

	if (thrustdir.x > sin (SM_GIMBAL_RANGE)) thrustdir.x = sin (SM_GIMBAL_RANGE);  
	else if (thrustdir.x < sin (-SM_GIMBAL_RANGE)) thrustdir.x = sin (-SM_GIMBAL_RANGE);  

	// Animate engine gimbal
	double pitchgimbal = 0.5 + (asin (thrustdir.y) / (2 * SM_GIMBAL_RANGE));
	double yawgimbal = 0.5 + (asin (thrustdir.x) / (2 * SM_GIMBAL_RANGE));

	//sprintf (oapiDebugString (), "X %f, Y %f", yawgimbal, pitchgimbal);

	if (gimbalx_proc != pitchgimbal)
	{
		double delta = simdt / SM_GIMBAL_RATE;
		double diff = gimbalx_proc - pitchgimbal;
	
		if (fabs (diff) < delta) gimbalx_proc = pitchgimbal;
		else if (diff < 0) gimbalx_proc += delta;
		else gimbalx_proc -= delta;
	}

	if (gimbaly_proc != yawgimbal)
	{
		double delta = simdt / SM_GIMBAL_RATE;
		double diff = gimbaly_proc - yawgimbal;
		
		if (fabs (diff) < delta) gimbaly_proc = yawgimbal;
		else if (diff < 0) gimbaly_proc += delta;
		else gimbaly_proc -= delta;
	}
	SetAnimation (anim_gimbalx, gimbalx_proc);
	SetAnimation (anim_gimbaly, gimbaly_proc);

	// Update thrust vector
	thrustdir.x = sin (SM_GIMBAL_RANGE) - (2 * SM_GIMBAL_RANGE * gimbaly_proc); 
	thrustdir.y = sin (SM_GIMBAL_RANGE) - (2 * SM_GIMBAL_RANGE * gimbalx_proc); 
	thrustdir.z = sqrt (1 - pow (thrustdir.x, 2) - pow (thrustdir.y, 2));
	SetThrusterDir (th_sps, thrustdir);

	//sprintf (oapiDebugString (), "X %f, Y %f, Z %f", thrustdir.x, thrustdir.y, thrustdir.z);
	
	// Animate High Gain Antenna
	if (AntennaStatus > OPEN)	
	{
		double	delta = simdt / 0.5;				// delta = simdt divided by duration of animation in seconds

		if (AntennaStatus == OPENING)
		{
			if (antenna_proc < 1) antenna_proc += delta;	// If process value is less than 1 add delta to process value
			else AntennaStatus = OPEN;					// If process value is greater than 1 GearStatus = OPEN
		}
		
		if (AntennaStatus == CLOSING)
		{
			if (antenna_proc > 0) antenna_proc -= delta;	// If process value is Greater than 0 subtract delta from process value
			else AntennaStatus = CLOSED;				// If process value is less than 0 GearStatus = CLOSED
		}

		SetAnimation (anim_antenna, antenna_proc);
		SetAnimation (anim_light, antenna_proc);
		SetAnimation (anim_umbilical, antenna_proc);
	} // End "if (GearStatus > OPEN)"

	if (GetAttachmentStatus (ah_booster) != NULL)
	{
		SetAnimation (anim_light, 0);
		SetAnimation (anim_antenna, 0);
	}

	// Clean up
	OBJHANDLE self = GetHandle ();					// Get vessel handle
	double deccel = GetDrag () / GetMass ();		// Decceleration due to drag
	if (deccel > 6 * G) oapiDeleteVessel (self);	// if deccel is greater than 6 Gs burn up

	// Debuggery
	// sprintf (oapiDebugString (), "Decceleration [Gs] %0.2f", deccel / G);
}

// --------------------------------------------------------------
// Process keyboard inputs
// --------------------------------------------------------------
int  SM::clbkConsumeBufferedKey (DWORD key, bool down, char *kstate)
{
	if (!down) return 0;

	// Deploy HGA when [K] is pressed.
	if (key == OAPI_KEY_K && !KEYMOD_SHIFT(kstate) && !KEYMOD_CONTROL (kstate) && !KEYMOD_ALT(kstate)) // [K] is down, no [shift], no [ctrl], no [alt]
	{
		if (AntennaStatus != OPEN) AntennaStatus = OPENING;
		else if (AntennaStatus != CLOSED) AntennaStatus = CLOSING;
		return 1;			// return '1' to indicate that the input has been processed
	}

	 // Kill thrust when Numpad [*] key is pressed.
	if(key == OAPI_KEY_MULTIPLY  && down && !KEYMOD_SHIFT(kstate) && !KEYMOD_CONTROL (kstate) && !KEYMOD_ALT(kstate))
	{
		for (int i = 0; i < 16; i++) SetThrusterLevel (th_sm_rcs[i], 0);
		for (int i = 0; i < 8; i++) DeactivateNavmode (i); 
		return 1;
	}

	return 0; // if no keys are pressed the function returns '0' and nothing happens
}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new SM (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (SM*)vessel;
}