// ==============================================================
//              Orbiter Module: SM
//    part of the Apollo Aplications Prject for Orbiter (AAPO)
//               copyright (c) 2012 greg hlynka
//                    all rights reserved
//
// SM.h
// control module for Apollo Service Module vessel class
//
// notes:
// ==============================================================

#pragma once
#include "CSM.h"

// ==============================================================
// Some vessel constants
// ==============================================================
const VECTOR3	SM_SIMBAY_OFFSET	= { 1.065, 1.434, 0.752};	// Position of SIM-Bay relative to service module's CoG

const double	SM_SIZE				= 4.68;						// Service module mean radius [m]
const VECTOR3	SM_CS				= { 25.28, 25.16, 14.81};	// Service module x,y,z cross sections [m^2]
const VECTOR3	SM_PMI				= { 12.00, 12.00, 7.00};	// Service module principal moments of inertia (mass-normalised) [m^2]
const VECTOR3	SM_RD				= { 0.70, 0.70, 0.30};		// Service module rotation drag coefficients
const VECTOR3	SM_COM				= {-0.12,-0.18, 0.62};		// Service module center of mass 
const double	SM_CW[4]			= { 0.3, 0.3, 1.4, 1.4};	// Wind resistance coefficients (forward, aft, lateral, vertical)

const double	SM_EMPTYMASS		= 6110.00;					// Service module empty mass [kg] (Encyclopedia Astronautica) 
const double	SM_O2MASS			= 147.87;					// Service module O2 reserve [kg] (326 lbs per tank)
const double	SM_H2MASS			= 13.15;					// Service module H2 reserve [kg] (29 lbs per tank)
const double	SM_FUELMASS			= 18535.15;					// Service module propulsion system fuel mass [kg] (15,723 lbs of fuel + 25,140 lbs of oxidizer, Apollo CSM News Reference pg 206)
const double	SM_RCS_FUELMASS		= 154.22;					// Service module RCS fuel mass [kg] (114 lbs of fuel + 226 lbs of oxidizer per quad, Apollo CSM News Reference pg 196)
const int		SM_TANKCOUNT		= 7;						// Number of service module  propellant recources

const double	SM_ISP				= 3080.34;					// Service module propulsion system specific impulse [m/s] (314 s, Wikipedia)
const double	SM_MAXTHRUST		= 91188.54;					// Service module propulsion system max thrust [n] (20,500 lbf, Apollo CSM News Reference pg 205)
const VECTOR3	SM_ENGINEPOS		= { 0.00, 0.00,-1.410};		// Service module main engine position [m]
const double	SM_GIMBAL_RANGE		= RAD * 6;					// Service module main engine gimbal range [r] 6 degrees
const double	SM_GIMBAL_RATE		= 1;						// Service module main engine gimbal range [r] 3 degrees a second
const double	SM_RCS_ISP			= 2844.90;					// Service module RCS specific impulse [m/s] (290 s, Encyclopedia Astronautica)
const double	SM_RCS_MAXTHRUST	= 444.82;					// Service module RCS max thrust [n] (100 lbf, Apollo CSM News Reference pg 194)
const VECTOR3	SM_RCS_ENGINEPOS[4]	= {{-0.325, 1.996, 1.810}, {-1.996,-0.325, 1.810}, { 0.325,-1.996, 1.810}, { 1.996, 0.325, 1.810}};	// Service module RCS quad positions [m] (A, B, C, D)

const VECTOR3	SM_DOCK_POS			= { 0.00, 0.00, 3.05};		// docking port location [m]
const VECTOR3	SM_DOCK_DIR			= { 0, 0, 1};				// docking port approach direction
const VECTOR3	SM_DOCK_ROT			= { 0, 1, 0};				// docking port alignment direction

const VECTOR3	SM_ATTACH_POS		= { 0.00, 0.00,-1.26};		// aft attachment point location [m]
const VECTOR3	SM_ATTACH_DIR		= { 0, 0,-1};				// aft attachment point direction
const VECTOR3	SM_ATTACH_ROT		= { 0, 1, 0};				// aft attachment point direction

const VECTOR3	SM_HGA_MOUNT		= {-1.0714,-1.3837,-1.3600};	// High-gain Antenna mounting point 
const VECTOR3	SM_HGA_PIVOT		= {-1.6836,-2.1744,-1.3600};	// High-gain Antenna pivot point, for pitch and yaw animations.
const double	SM_HGA_CANT			=  RAD *-52;					// High-gain Antenna Rotation angle

const VECTOR3	SM_LIGHT_PIVOT		= { 1.5, 1.29, 3.1};		// 
const double	SM_LIGHT_CANT		=  RAD *-45;				// 

const VECTOR3	SM_UMBILICAL_PIVOT	= { 0,-1.9685, 3.330};

// ==============================================================
// Descent stage class Interface
// ==============================================================

class SM: public VESSEL3 {
public:
	SM (OBJHANDLE hVessel, int flightmodel);
	~SM ();

	// Custom vessel functions
	void	DefineAnimations (UINT mesh);		// Define mesh animations
	void	DefineParticleStreams (void);		// Define exhaust and particle streams
	void	DefineSounds (void);				// Define sound effects

	// Overloaded callback functions
	void	clbkSetClassCaps (FILEHANDLE cfg);								// Set the capabilities of the vessel class
	void	clbkLoadStateEx (FILEHANDLE scn, void *status);					// Read status from scenario file
	void	clbkSaveState (FILEHANDLE scn);									// Write status to scenario file
	void	clbkPostCreation (void);
	void	clbkPostStep (double simt, double simdt, double mjd);			// Manage Animations and Post-Step processes
	int		clbkConsumeBufferedKey (DWORD key, bool down, char *kstate);	// Process keyboard inputs

	// Implimentation of OrbiterSound 4.0
	int		OrbiterSoundID;

	ATTACHMENTHANDLE	ah_booster, ah_payload;

	// Thruster and Propellant handles
	PROPELLANT_HANDLE	ph_sm_rcs[4], ph_sps;								// Functional Propellant tanks
	PROPELLANT_HANDLE	ph_sm_O2, ph_sm_H2;									// Represenative propellant tanks (consumables)
	
	THRUSTER_HANDLE		th_sm_rcs[16], th_sps;								// Functional thrusters
	THRUSTER_HANDLE		th_vent;											// Represenative thrusters (particle streams)

private:
	double	thrust_mod, dv_mod, consumables_mod;
	double	sep_bt;

	// Vessel status variables
	enum doorstate	AntennaStatus, LightStatus, UmbilicalStatus;
	double			antenna_proc, light_proc, umbilical_proc, gimbalx_proc, gimbaly_proc;
	bool			BayOpen;
	

	// Animations
	UINT						anim_light, anim_umbilical, anim_gimbalx, anim_gimbaly, anim_antenna, anim_HGApitch, anim_HGAyaw;
	ANIMATIONCOMPONENT_HANDLE	ach_antenna, ach_HGApitch, ach_HGAyaw;

	// Meshes
	MESHHANDLE	mh_sbaycover, mh_service;
	UINT		mesh_SM;
};
